/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/*
 * $Id: dmpl.c 1255 2006-11-01 23:08:47Z jinfante $
 *
 * Diagnostic utility for LightPulse Fibre Channel PCI Adapter
 */


/* The following line is required by Doxygen */
/*! \file lpfc_dfc.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <stdint.h>

#define MPULSE 1
#include "mplib.h"
#include "mpapi.h"
#include "mpl_ioctl.h"

typedef struct fcPathId {
   MPL_PATH_LIST_ENTRY m_id;
   MPL_PATH_STATE m_state;
   uint16_t m_flag;
   uint8_t  m_tratio;        /* this paths component of the traffic ratio */
   uint8_t  m_tratio_cnt;    /* current traffic ratio counter for this path */
   void   * m_devptr;        /* used to point to internal driver device info */
   void  *(*m_fail_routine)(void *, void *); /* for future use */
   void  *(*m_restore_routine)(void *, void *); /* for future use */
   void   * m_hbeat_tmo;     /* Holds heartbeat tmo clk */
} fcPathIdentifier;

typedef struct fcRouteId {
   MPL_ROUTE_LIST_ENTRY    m_id;    /* Device ID of this route */
   MPL_PATH_CONTROL_HEADER m_hdr;
   uint16_t             m_lastpath;   /* last path used, for load balancing */
   uint16_t             m_state;      /* route state */
   uint8_t              m_fstate;     /* failover state */
   uint8_t              m_fflag;      /* failover flag */
   uint8_t              m_maxvstate;  /* fcVcheck state */
   uint8_t              m_vflag;      /* fcVcheck flags */
   fcPathIdentifier   * m_reservePath;
   fcPathIdentifier     m_paths[MPL_MAX_PATHS];
} fcRouteIdentifier;

#define DFC_MAXNODES    512  /* Max nodes dfc can display */

/* Global function declarations */

typedef struct cmdinfo {
   int                  c_cmd;
   char                 *c_string;
   int                  (*c_routine)(struct cmdinfo *cp, void *p);
} CMDINFO_t;

void dfc_mac_sprintf(uint8_t *IEEEp, char *buf);

/* Local function declarations */
static   int promptContinueOrAbort( void);


/* after making changes to this list of active functions, please */
/*  re-alphabetize, thank you (write this section out to a file, then do a */
/*  shell sort /unsorted > /sorted, then read back in) */
#ifdef MPULSE
int     maddpath( CMDINFO_t *cp, void *p);
int     maddroute( CMDINFO_t *cp, void *p);
#endif
#ifdef MPULSE
int     mdelpath( CMDINFO_t *cp, void *p);
int     mdelroute( CMDINFO_t *cp, void *p);
int     mexportlist( CMDINFO_t *cp, void *p);
int     mexportpath( CMDINFO_t *cp, void *p);
int     mgetpathctl( CMDINFO_t *cp, void *p);
int     mgetpathinfo( CMDINFO_t *cp, void *p);
int     mgetpathstat( CMDINFO_t *cp, void *p);
int     mimportpath( CMDINFO_t *cp, void *p);
int     mpathlist( CMDINFO_t *cp, void *p);
int     mportlist( CMDINFO_t *cp, void *p);
int     mroutelist( CMDINFO_t *cp, void *p);
int     msetpathctl( CMDINFO_t *cp, void *p);
int     msetpathstat( CMDINFO_t *cp, void *p);
int     mgetcfg( CMDINFO_t *cp, void *p);
int     msetcfg( CMDINFO_t *cp, void *p);
#endif
int     help( CMDINFO_t *cp, void *p);
int     inval( CMDINFO_t *cp, void *p);
int     quit( CMDINFO_t *cp, void *p);
int     set( CMDINFO_t *cp, void *p);
int     shell( CMDINFO_t *cp, void *p);

#define LPFC_DFC_EXIT             0x1000
#define LPFC_DFC_DBG              0x2000
#define LPFC_DFC_SHELL            0x3000
#define LPFC_DFC_INVAL            0x4000
#define LPFC_DFC_SET              0x5000
#define LPFC_DFC_HELP             0x6000

/* after making changes to this list of active functions, please */
/*  re-alphabetize;  (write this section out to a file, then do a */
/*  shell sort /unsorted > /sorted, then read back in) */
CMDINFO_t dfcCmds[] = 
{
   { LPFC_DFC_INVAL,              "inval",        inval,       },
   { LPFC_DFC_SHELL,              "!",            shell,       },
   { LPFC_DFC_EXIT,               "bye",          quit,	       },
   { LPFC_DFC_HELP,               "help",         help,        },
#ifdef MPULSE
   { ELX_MPL_ADD_PATH_LIST,       "maddpath",     maddpath,    },
   { ELX_MPL_CREATE_ROUTE,        "maddroute",    maddroute,   },
   { ELX_MPL_REMOVE_PATH_LIST,    "mdelpath",     mdelpath,    },
   { ELX_MPL_DELETE_ROUTE,  	  "mdelroute",    mdelroute,   },
   { ELX_MPL_GET_EXPORT_PATH_LIST,"mexportlist",  mexportlist, },
   { ELX_MPL_EXPORT_PATH_LIST,    "mexportpath",  mexportpath, },
   { ELX_MPL_GET_PATH_CONTROL,    "mgetpathctl",  mgetpathctl, },
   { ELX_MPL_GET_PATH_INFO,       "mgetpathinfo", mgetpathinfo,},
   { ELX_MPL_GET_PATH_STATUS,     "mgetpathstat", mgetpathstat,},
   { ELX_MPL_IMPORT_PATH_LIST, 	  "mimportpath",  mimportpath, },
   { ELX_MPL_PATHLIST,         	  "mpathlist",    mpathlist,   },
   { ELX_MPL_PORTLIST	,         "mportlist",    mportlist,   },
   { ELX_MPL_ROUTELIST,           "mroutelist",   mroutelist,  },
   { ELX_MPL_SET_PATH_CONTROL,    "msetpathctl",  msetpathctl, },
   { ELX_MPL_SET_PATH_STATUS,     "msetpathstat", msetpathstat,},
   { ELX_MPL_GETCFG,              "mgetcfg",      mgetcfg,},
   { ELX_MPL_SETCFG,   	          "msetcfg",      msetcfg,},
#endif
   { LPFC_DFC_EXIT,               "quit",         quit,	       },
   { LPFC_DFC_SET,	          "set",          set,	       },
};

CMDINFO_t * getcmd( char *buf, int size);
#define DELIMS      " \t,\n"
#define MAX_CMD_ARGS    7
char    *margv[MAX_CMD_ARGS];
int dfccmdcnt=sizeof(dfcCmds)/sizeof(CMDINFO_t);

#define MAX_ELX_BRDS 10
static union {
   struct outdata {
      char      dataout[256*1024];
      char      slim[8192];
   } d;
   char  version[MAX_ELX_BRDS];
}dfc_un;

int binfo_fc_flag;
uint8_t *binfo_fc_slim2_virt;
uint8_t *dfc_bigbuffer;
uint8_t *dfc_slimbuffer;

int margc;
int cur_iocb;
int cur_ring;
int cur_board;
int gotcfg  = 0;
int num_lines = 16;

char *dfc_err_cannot_issue_cmd = "Cannot issue command\n";

#ifdef MPULSE
fcRouteIdentifier dfc_mpcfg[(DFC_MAXNODES / 2)];
elxDiDeviceIdentifier dfcCurDevId[8];
MPL_PATH_LIST_ENTRY dfc_expathlist[DFC_MAXNODES];
MPL_PATH_LIST_ENTRY dfc_impathlist[DFC_MAXNODES];
extern elxDiDeviceIdentifier mplNullIdentifier;
int     dfc_mpcfg_cnt = 0;
int     dfc_Routes_cnt = 0;
int     dfc_impathlist_cnt = 0;
int     dfc_expathlist_cnt = 0;
#endif /* MPULSE */

#ifdef MPULSE
char *mptype[7] = {
      "UNKNOWN",
      "COLD STANDBY",
      "TRAFFIC RATIO",
      "DYNAMIC LUNQ",
      "DYNAMIC TGTQ",
      "DYNAMIC HBAQ",
      "DYNAMIC TIME",
};

char *mpstate[10] = {
      "UNKNOWN",
      "AVAILABLE",
      "UNKNOWN",
      "FAILED",
      "UNKNOWN",
      "OFFLINE",
      "UNKNOWN",
      "UNKNOWN",
      "UNKNOWN",
      "DISABLED",
};
#endif /* MPULSE */

extern uint32_t EventMissedCount;
extern uint32_t dfc_state;
extern uint32_t dfc_refresh_timer[MAX_ELX_BRDS];

extern void mbinval(CMDINFO_t *cp, void *p);

typedef void (*elx_sighandler_t)(int);

/*! main 
 *
 * \pre
 * \post
 * \param  argc     
 * \param  argv     
 * \param  envp    
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 *    main function code 
 *
 */

int
main(int argc, char **argv, char **envp)
{
   CMDINFO_t *cp;
   char   cmd_line[132] = { 0 };
   uint32_t  numbrds, i;

#ifdef _aix
   if(!__KERNEL_32()) 
       {
       if (strcmp(argv[0],"/usr/lpp/diagnostics/dfc64")){
            system("/usr/lpp/diagnostics/dfc64");
            exit();
           }
       }
#endif

   signal(SIGQUIT, (elx_sighandler_t)quit);
   signal(SIGINT,  (elx_sighandler_t)quit);
   signal(SIGTERM, (elx_sighandler_t)quit);
   signal(SIGQUIT, (elx_sighandler_t)quit);

   dfc_bigbuffer = (uint8_t *)dfc_un.d.dataout;
   dfc_slimbuffer = (uint8_t *)dfc_un.d.slim;
   bzero((uint8_t * )dfc_bigbuffer, sizeof(dfc_un.version));
   printf("LightPulse Engineering MultiPulse Debug Utility\n");
   printf("Copyright 2003-2006, Emulex Corporation.\n");
   numbrds = mplInitialize((char *)&dfc_un.version[0], 16);
   if(numbrds != MPL_STATUS_SUCCESS) {
      printf("\nFound No Adapters\n");
      exit(1);
   }
   printf("\nDo not run this utility unless instructed by Emulex Technical Support\n");

   printf("\nMPULSE version %s\n", &dfc_un.version[0]);

   /* Initialize slim and ring global variables for default board and ring */
   cur_iocb = 0;
   cur_ring = 0;
   cur_board = numbrds;

   /* Prompt user for input and call appropriate callout routine
    * to do work.
    */
   for (;;) {
      printf("cmd> ");
      cp = getcmd(cmd_line, sizeof(cmd_line));
      if((i = (*(cp->c_routine))(cp, 0)) != 0)
         exit(0);
   }
}

/*! getcmd 
 *
 * \pre
 * \post
 * \param  buf     
 * \param  size
 *
 * \return CMDINFO_t     
 *
 * \b Description:
 *    Read a command line from the user and tries to match
 *    a supported command.
 */
CMDINFO_t *
getcmd(
   char *buf,
   int   size)
{
   char savebuf[132];
   int  i = 0;

   bcopy(buf, savebuf, 132);
   while (! fgets(buf, 132, stdin));

   /* Replace the <CR> by a null terminator */
   while ((buf[i] != '\n') && (i < 131))
      i ++;
   buf[i] = '\0';

   /* Check for run previous command */
   if (strcmp(buf, "p") == 0) {
      bcopy(savebuf, buf, 132); /* run previous command */
   } else {
      /* setup margc and margv for command callout routine */
      if ((margv[0] = strtok(buf, DELIMS)) == 0)
         return(&dfcCmds[0]);  /* Invalid command */

      for (i = 1; i < MAX_CMD_ARGS; i++) {
         if ((margv[i] = strtok(NULL, DELIMS)) == 0) {
            break;
         }
      }
      margc = i;
   }

   if (margv[0] == 0)
      return(&dfcCmds[0]);

   /* Look for a match on command string */
   for (i = 0; i < dfccmdcnt; i++) {
      if (strcmp(margv[0], dfcCmds[i].c_string) == 0)
         return(&dfcCmds[i]);  /* match */
   }
   return(&dfcCmds[0]);  /* Invalid command */
}


/*! getint 
 *
 * \pre
 * \post
 * \param  s                                            
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 *    Prompts the user for input, then returns the integer value
 */
int getint(
   char    *s)
{
   char buf[132];
   char *bp;

   printf("%s: ", s);
   fgets(buf, 132, stdin);
   bp = strtok(buf, DELIMS);
   return(atoi(buf));
}


/*! getxint 
 *
 * \pre
 * \post
 * \param  s         
 *
 * \return uint32_t Return Code  0 = success
 *
 * \b Description:
 *    Prompts the user for input, then returns the hex integer value
 */
uint32_t getxint(
   char    *s)
{
   char buf[132];
   char *bp;

   printf("%s (hex): ", s);
   fgets(buf, 132, stdin);
   bp = strtok(buf, DELIMS);
   return(strtoul(buf, NULL, 16));
}

/*! inval
 *
 * \pre
 * \post
 * \param  cp        
 * \param  p        
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 *    Called whenever an invalid command is encountered
 */
int
inval(
   CMDINFO_t *cp,
   void *p)
{
   if ((margv[0] == 0) || (*margv[0] == 0))
      return(0);
   printf("Invalid command: %s\n", margv[0]);
   return(0);
}

/*! quit 
 *
 * \pre
 * \post
 * \param  cp
 * \param  p
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 *    Cleanup and exit application 
 */
int
quit(
   CMDINFO_t *cp,
   void *p)
{
   mplClose();
   return(1);
}


/*! shell 
 *
 * \pre
 * \post
 * \param  cp
 * \param  p
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 *    Execute shell command
 */
int
shell(
   CMDINFO_t *cp,
   void *p)
{
   char user_cmd[180];
   int   i;

   sprintf(user_cmd, "%s", margv[1]);

   for (i=2; i<margc; i++)
   {
      sprintf(user_cmd, "%s %s", user_cmd, margv[i]);
   }

   system(user_cmd);

   return (0);
}

/*! set  
 *
 * \pre
 * \post
 * \param  cp
 * \param  p
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 *    Display or set the current board / ring /iocb parameters
 *    for this application.
 */
int
set(
   CMDINFO_t *cp,
   void *p)
{
   char *cmd;
   int  value;
   int  value1;

   if (margc == 1) {
      printf("board %d  ring %d  iocb %d", cur_board, cur_ring, cur_iocb);
      printf(" lines %d", num_lines);
      printf("\n");
      return(0);
   }

   if ((margc != 3) && (margc != 4)) {
      printf("Usage: set [board | ring | iocb | wwn | lines] value\n");
      return(0);
   }

   cmd = margv[1];
   value = atoi(margv[2]);
   value1=0;
   if (margc == 4)
      value1 = atoi(margv[3]);

   if (strcmp(cmd, "board") == 0) {
      cur_board = value;
      /* rring(cp, (void *)0); */
   } else if (strcmp(cmd, "ring") == 0) {
      cur_ring = value;
      /* rring(cp, (void *)0); */
   } else if (strcmp(cmd, "iocb") == 0) {
      cur_iocb = value;
   } else if (strcmp(cmd, "lines") == 0) {
      num_lines = value;
   } else {
      printf("Usage: set [board | ring | iocb | wwn | lines] value\n");
      return(0);
   }

   return(0);
}


/*! initbrds
 *
 * \pre
 * \post
 * \param  cp
 * \param  p
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 *    Initialize board -- To be Implemented      
 */
int
initbrds(
   CMDINFO_t *cp,
   void *p)
{
return(0);
}


/*! help 
 *
 * \pre
 * \post
 * \param  cp
 * \param  p
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 * Print list of available commands
 */
int
help(
   CMDINFO_t *cp,
   void *p)
{
   int  i, j;

   j=0;
   for (i = 0; i < dfccmdcnt; i++) {
      if ((strncmp(dfcCmds[i].c_string, "inval", 5 ) == 0) || (strncmp(dfcCmds[i].c_string, "!", 1 ) == 0))
         continue;
      printf("%10s ", dfcCmds[i].c_string);
      j++;
      if ((j % 4) == 0)
         printf("\n");
   }
   printf("%10s\n", "p");
   return(0);
}


#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mportlist(
   CMDINFO_t *cp,
   void *p)
{
   int rc, i;
   char buf1[32];
   MPL_GET_PORT_LIST *pl;
   MPL_PORT_LIST_ENTRY *pe;

   pl = (PMPL_GET_PORT_LIST)dfc_bigbuffer;
   pl->Count = 256;
   if((rc = mplGetPortList(pl)) != MPL_STATUS_SUCCESS) {
      printf("mplGetPortList failed: %d\n", rc);
      return(0);
   }
   printf("Port List by WWPN    cnt %d\n", pl->Count);
   pe = (MPL_PORT_LIST_ENTRY *)&pl->Ports[0];
   for (i = 0; i < (int)pl->Count; i++) {
      dfc_mac_sprintf((uint8_t *)&pe->DevId.un.m_wwn[0], buf1);
      printf("%d: %s  state %d\n", i, buf1, pe->State);
      if ((i % num_lines) == (num_lines-1)) {
         if( promptContinueOrAbort())
            break;  /* abort */
      }
      pe++;
   }
   return(0);
}
#endif /* MPULSE */

static char *print_routeid(struct ElxDiDevId *devid)
{
	static char buffer[256];

	if (devid->m_type == ELX_DI_DEVID_SCSI_NEXUS) {
		snprintf(buffer, sizeof(buffer),
			 "lpfc%dt%dd%d",
			 devid->un.m_ScsiId.m_adapter,
			 devid->un.m_ScsiId.m_target,
			 devid->un.m_ScsiId.m_lun);
	} else if (devid->m_type == ELX_DI_DEVID_NEXUS_LONG) {
		snprintf(buffer, sizeof(buffer),
			 "%02x%02x%02x%02x%02x%02x%02x%02x:"
			 "%02x%02x%02x%02x%02x%02x%02x%02x:%u",
			 devid->un.m_NexusLong.m_adapter_wwpn[0],
			 devid->un.m_NexusLong.m_adapter_wwpn[1],
			 devid->un.m_NexusLong.m_adapter_wwpn[2],
			 devid->un.m_NexusLong.m_adapter_wwpn[3],
			 devid->un.m_NexusLong.m_adapter_wwpn[4],
			 devid->un.m_NexusLong.m_adapter_wwpn[5],
			 devid->un.m_NexusLong.m_adapter_wwpn[6],
			 devid->un.m_NexusLong.m_adapter_wwpn[7],
			 devid->un.m_NexusLong.m_target_wwpn[0],
			 devid->un.m_NexusLong.m_target_wwpn[1],
			 devid->un.m_NexusLong.m_target_wwpn[2],
			 devid->un.m_NexusLong.m_target_wwpn[3],
			 devid->un.m_NexusLong.m_target_wwpn[4],
			 devid->un.m_NexusLong.m_target_wwpn[5],
			 devid->un.m_NexusLong.m_target_wwpn[6],
			 devid->un.m_NexusLong.m_target_wwpn[7],
			 devid->un.m_NexusLong.m_lun[0]);
	} else if (devid->m_type == ELX_DI_DEVID_INQSN) {
		char *cp = buffer;
		int  i;

		for (i = 0; i < devid->m_len && cp < buffer + sizeof(buffer) - 1; ++i) {
			cp += snprintf(cp, buffer + sizeof(buffer) - cp,
				       "%02x",
				       devid->un.m_InquiryData[i]);
		}
	} else {
		snprintf(buffer, sizeof(buffer),
			 "bad devid type %d", devid->m_type);
	}
	return buffer;
}

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mroutelist(
   CMDINFO_t *cp,
   void *p)
{
   int rc, i, k;
   MPL_GET_ROUTE_LIST *rl;
   MPL_ROUTE_LIST_ENTRY *re;

   rl = (PMPL_GET_ROUTE_LIST)dfc_bigbuffer;
   rl->Count = 0;  /* dfc_Routes[] */
   rl->Count = 256;  /* dfc_Routes[] */
   if((rc = mplGetRouteList(rl)) != MPL_STATUS_SUCCESS) {
	   printf("mplGetRouteList failed: %d, count = %d\n", rc, rl->Count);
      return(0);
   }
   dfc_Routes_cnt = rl->Count;

   printf("Route List    cnt %d\n", rl->Count);
   re = (MPL_ROUTE_LIST_ENTRY *)&rl->Routes[0];
   for (i = 0; i < (int)rl->Count; i++) {

      bcopy((uint8_t *)re, (uint8_t *)&dfc_mpcfg[i].m_id,
         sizeof(MPL_ROUTE_LIST_ENTRY));

      printf("Route idx:%d  len %d: ", i, re->RouteId.m_len);
      printf("%s  ", print_routeid(&re->RouteId));

      if(re->ValidationId.m_len) {
         printf("ValID len %d: ", re->ValidationId.m_len);
         for(k=0;k<re->ValidationId.m_len;k++) {
            printf("%02x", (unsigned char)re->ValidationId.un.m_InquiryData[k]);
         }
      }
      printf(" fb %d fnd %d\n", re->AutoFailBack, re->Found);

      if ((i % num_lines) == (num_lines-1)) {
         if( promptContinueOrAbort())
            break;  /* abort */
      }
      re++;
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mpathlist(
   CMDINFO_t *cp,
   void *p)
{
   int rc, i, k, index;
   MPL_GET_PATH_LIST   *pl;
   MPL_PATH_LIST_ENTRY *pe;

   pl = (PMPL_GET_PATH_LIST)dfc_bigbuffer;
   switch(margc) {
   case 1:
      index = 0;
      bcopy( (uint8_t *)&mplNullIdentifier, (uint8_t *)&pl->RouteId,
         sizeof(elxDiDeviceIdentifier));
      pl->Count = DFC_MAXNODES;
      break;
   case 2:
      index  = strtoul(margv[1], 0, 0);

      if(index >= dfc_Routes_cnt) {
         printf("mpathlist route_index not valid, run mroutelist\n");
         return(0);
      }
      bcopy( (uint8_t *)&dfc_mpcfg[index].m_id.RouteId, (uint8_t *)&pl->RouteId,
         sizeof(elxDiDeviceIdentifier));
      pl->Count = (DFC_MAXNODES / 2); /* min of 2 paths per route */
      break;
   default:
      printf("Usage: mpathlist [route_index]\n");
      return(0);
   }


   if((rc = mplGetPathList(pl)) != MPL_STATUS_SUCCESS) {
      printf("mplGetPathList failed: %d\n", rc);
      if(rc != MPL_STATUS_BUFFER_TOO_SMALL)
         return(0);
   }
   if(margc == 1) {
      printf("Imported Path List:   cnt %d\n", pl->Count);
      if(pl->Count > DFC_MAXNODES)
         pl->Count = DFC_MAXNODES;
   }
   else {
      printf("Path List for Route: ");
      printf("%s", print_routeid(&pl->RouteId));
      printf("    cnt %d\n", pl->Count);
   }
   pe = (MPL_PATH_LIST_ENTRY *)&pl->Paths[0];
   dfc_mpcfg[index].m_hdr.PathCount = 0;
   for (i = 0; i < (int)pl->Count; i++) {

      if(margc == 1) {
         if(i == 0) {
            bcopy((uint8_t *)pe, (uint8_t *)&dfc_impathlist[0],
               (pl->Count * sizeof(MPL_PATH_LIST_ENTRY)));
            dfc_impathlist_cnt = pl->Count;
         }
      }
      else {
         dfc_mpcfg[index].m_hdr.PathCount++;
         bcopy((uint8_t *)pe, (uint8_t *)&dfc_mpcfg[index].m_paths[i].m_id,
            sizeof(MPL_PATH_LIST_ENTRY));
      }

      printf("Path idx:%d  %s dft %d fmask 0x%x bnd %d %s %s ",
	     i, print_routeid(&pe->DevId),
	     pe->Default, pe->FailoverBitMask, pe->Bound,
             (pe->Found ? "found" : "not_found"),
             (pe->StaticConfig ? "static" : "dynamic"));
      if(pe->ValidationId.m_len) {
         printf("ValID len %d: ", pe->ValidationId.m_len);
         for(k=0;k<pe->ValidationId.m_len;k++) {
            printf("%02x", (unsigned char)pe->ValidationId.un.m_InquiryData[k]);
         }
      }
      printf("\n");
      if ((i % num_lines) == (num_lines-1)) {
         if( promptContinueOrAbort())
            break;  /* abort */
      }
      pe++;
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mexportlist(
   CMDINFO_t *cp,
   void *p)
{
   int rc, i, k, index;
   MPL_GET_PATH_LIST   *pl;
   MPL_PATH_LIST_ENTRY *pe;

   pl = (PMPL_GET_PATH_LIST)dfc_bigbuffer;
   switch(margc) {
   case 1:
      index = 0;
      bcopy( (uint8_t *)&mplNullIdentifier, (uint8_t *)&pl->RouteId,
         sizeof(elxDiDeviceIdentifier));
      pl->Count = DFC_MAXNODES;
      break;
   default:
      printf("Usage: mexportlist\n");
      return(0);
   }


   if((rc = mplGetExportedPaths(pl)) != MPL_STATUS_SUCCESS) {
      printf("mplGetPathList failed: %d\n", rc);
      if(rc != MPL_STATUS_BUFFER_TOO_SMALL)
         return(0);
   }
   printf("Exported Path List:   cnt %d\n", pl->Count);
   pe = (MPL_PATH_LIST_ENTRY *)&pl->Paths[0];
   dfc_mpcfg[index].m_hdr.PathCount = 0;
   for (i = 0; i < (int)pl->Count; i++) {

      if(i == 0) {
         bcopy((uint8_t *)pe, (uint8_t *)&dfc_expathlist[0],
            (pl->Count * sizeof(MPL_PATH_LIST_ENTRY)));
         dfc_expathlist_cnt = pl->Count;
      }

      printf("Path idx:%d  %s fmask 0x%x  %s ",
	     i, print_routeid(&pe->DevId), pe->FailoverBitMask,
             (pe->Found ? "   found" : "notfound"));
      if(pe->ValidationId.m_len) {
         printf("ValID len %d: ", pe->ValidationId.m_len);
         for(k=0;k<pe->ValidationId.m_len;k++) {
            printf("%02x", (unsigned char)pe->ValidationId.un.m_InquiryData[k]);
         }
      }
      printf("\n");
      if ((i % num_lines) == (num_lines-1)) {
         if( promptContinueOrAbort())
            break;  /* abort */
      }
      pe++;
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
maddroute(
   CMDINFO_t *cp,
   void *p)
{
   elxDiDeviceIdentifier *rvp;
   struct dfc_pctl { /* This duplicates MPL_GET_PATH_LIST */
      elxDiDeviceIdentifier routeid;
      int cnt;
      int pad;
      MPL_PATH_LIST_ENTRY pl[MPL_MAX_PATHS];
   } path;
   union {
      MPL_PATH_CONTROL_HEADER pc;
      MPL_TRAFFIC_RATIO_CONTROL tr;
      MPL_FAILOVER_CONTROL fc;
   } un;
   int j, idx, pcflag, rc;

   if (margc < 3) {
      printf("Usage: maddroute pcflag [pathlist_index0 ... pathlist_indexN]\n");
      return(0);
   }
   if(dfc_impathlist_cnt == 0) {
      printf("Usage: maddroute: Must run mpathlist first\n");
      return(0);
   }

   bzero((uint8_t *)&path, sizeof(struct dfc_pctl));

   /* First build Path Control information */
   bzero((uint8_t *)&un.pc, MPL_MAX_PATH_CONTROL_SIZE);

   pcflag = strtoul(margv[1], 0, 0);
   un.pc.Type = (pcflag & 0xf);
   if(un.pc.Type == 0) {
      un.pc.Type = MPL_PC_TYPE_FAILOVER; /* Default */ 
   }
   un.pc.AutoFailBack = 0;
   un.pc.InquiryHeartbeat = 0;
   if(pcflag & MPL_PC_CFG_AUTO_FAILBACK) {
      un.pc.AutoFailBack = 1;
   }
   if(pcflag & MPL_PC_CFG_INQUIRY_HEARTBEAT) {
      un.pc.InquiryHeartbeat = 1;
   }
   if(pcflag & MPL_PC_CFG_FAIL_MASK) {
      if(pcflag & MPL_PC_CFG_FAIL_ABORT_TASK_SET) {
         un.pc.FailoverAbortTaskSet = 1;
      }
      else if(pcflag & MPL_PC_CFG_FAIL_LUN_RESET) {
         un.pc.FailoverLunReset = 1;
      }
      else if(pcflag & MPL_PC_CFG_FAIL_TARGET_RESET) {
         un.pc.FailoverTgtReset = 1;
      }
   }
   else {
      un.pc.FailoverAbortTaskSet = 1;  /* Default */
   }

   un.pc.PathCount = (margc - 2);
   path.cnt = un.pc.PathCount;

   printf("Adding route for paths: ");
   for(j=0;j<un.pc.PathCount;j++) {
      idx = strtoul(margv[j+2], 0, 0);
      if((idx >= dfc_impathlist_cnt) || (j >= MPL_MAX_PATHS)) {
         printf("Usage: maddroute: pathlist_index invalid\n");
         return(0);
      }

      bcopy((uint8_t *)&dfc_impathlist[idx], (uint8_t *)&path.pl[j],
         sizeof(MPL_PATH_LIST_ENTRY));

      path.pl[j].FcpTarget = 1;
      path.pl[j].FailoverBitMask = MPL_PATH_DISCONNECTED;

      switch(un.pc.Type) {
      case MPL_PC_TYPE_FAILOVER:
         if(j == 0) {
            path.pl[j].Default = 1;
            un.pc.Length =  sizeof(MPL_FAILOVER_CONTROL);
            bcopy((uint8_t *)&dfc_impathlist[idx].DevId,
               (uint8_t *)&un.fc.PrimaryPathDevId,
               sizeof(elxDiDeviceIdentifier));
            rvp = &dfc_impathlist[idx].ValidationId;
            bcopy((uint8_t *)rvp, (uint8_t *)&un.fc.PrimaryPathValidationId,
               sizeof(elxDiDeviceIdentifier));
            rvp = &dfc_impathlist[idx].DevId;
            bcopy((uint8_t *)rvp, (uint8_t *)&path.routeid,
               sizeof(elxDiDeviceIdentifier));
            un.fc.PrimaryPathDevId.m_type = 0;
         }
         else {
            path.pl[j].Default = 0;
         }
         break;
      case MPL_PC_TYPE_TRAFFIC_RATIO:
         un.tr.Count = (j+1);
         un.tr.PathRatios[j].Ratio = 2;
         bcopy( (uint8_t *)&mplNullIdentifier,
            (uint8_t *)&un.tr.PathRatios[j].ValidationId, /* Not used for UNIX */
            sizeof(elxDiDeviceIdentifier));
         bcopy( (uint8_t *)&mplNullIdentifier,
            (uint8_t *)&un.tr.PathRatios[j].DeviceId, /* Not used for UNIX */
            sizeof(elxDiDeviceIdentifier));
         if(j == 0) {
            path.pl[j].Default = 1;
            un.pc.Length =  sizeof(MPL_TRAFFIC_RATIO_CONTROL);
            rvp = &dfc_impathlist[idx].DevId;
            bcopy((uint8_t *)rvp, (uint8_t *)&path.routeid,
               sizeof(elxDiDeviceIdentifier));
         }
         else {
            path.pl[j].Default = 0;
         }
         break;
      default:
         if(j == 0) {
            path.pl[j].Default = 1;
            un.pc.Length =  sizeof(MPL_PATH_CONTROL_HEADER);
            rvp = &dfc_impathlist[idx].DevId;
            bcopy((uint8_t *)rvp, (uint8_t *)&path.routeid,
               sizeof(elxDiDeviceIdentifier));
         }
         else {
            path.pl[j].Default = 0;
         }
         break;
      }

      printf("%d ",idx);
   }
   printf("\n");

   if((rc = mplCreateRoute((MPL_GET_PATH_LIST *)&path, &un.pc)) != MPL_STATUS_SUCCESS) {
      printf("mplCreateRoute failed: %d\n", rc);
      return(0);
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
maddpath(
   CMDINFO_t *cp,
   void *p)
{
   struct dfc_pctl { /* This duplicates MPL_GET_PATH_LIST */
      elxDiDeviceIdentifier routeid;
      int cnt;
      int pad;
      MPL_PATH_LIST_ENTRY pl[MPL_MAX_PATHS];
   } path;
   int j, idx, rc;

   if (margc < 3) {
      printf("Usage: maddpath route_index [pathlist_index0 ... pathlist_indexN]\n");
      return(0);
   }
   if(dfc_impathlist_cnt == 0) {
      printf("Usage: maddpath: Must run mpathlist first\n");
      return(0);
   }

   idx  = strtoul(margv[1], 0, 0);
   if(idx >= dfc_Routes_cnt) {
      printf("maddpath route_index not valid, run mroutelist\n");
      return(0);
   }
   bzero((uint8_t *)&path, sizeof(struct dfc_pctl));
   bcopy( (uint8_t *)&dfc_mpcfg[idx].m_id.RouteId, (uint8_t *)&path.routeid,
         sizeof(elxDiDeviceIdentifier));

   path.cnt = (margc - 2);

   printf("Adding to Route: ");
   printf("%s\n", print_routeid(&path.routeid));
   printf("  Paths: ");

   for(j=0;j<path.cnt;j++) {
      idx = strtoul(margv[j+2], 0, 0);
      if((idx >= dfc_impathlist_cnt) || (j >= MPL_MAX_PATHS)) {
         printf("Usage: maddpath: pathlist_index invalid\n");
         return(0);
      }

      bcopy((uint8_t *)&dfc_impathlist[idx], (uint8_t *)&path.pl[j],
         sizeof(MPL_PATH_LIST_ENTRY));

      path.pl[j].FcpTarget = 1;
      path.pl[j].FailoverBitMask = MPL_PATH_DISCONNECTED;
      path.pl[j].Default = 0;

      printf("Index %d:%s ",
	     idx, print_routeid(&path.pl[j].DevId));
   }
   printf("\n");

   if((rc = mplAddPathListToRoute((MPL_GET_PATH_LIST *)&path)) != MPL_STATUS_SUCCESS) {
      printf("mplAddPathListToRoute failed: %d\n", rc);
      return(0);
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mexportpath(
   CMDINFO_t *cp,
   void *p)
{
   MPL_GET_PATH_LIST * pl;
   int j, idx, rc;

   if (margc < 2) {
      printf("Usage: mexportpath [pathlist_index0 ... pathlist_indexN]\n");
      return(0);
   }
   if(dfc_impathlist_cnt == 0) {
      printf("Usage: mexportpath: Must run mpathlist first\n");
      return(0);
   }
   pl = (MPL_GET_PATH_LIST *)dfc_bigbuffer;

   pl->Count = (margc - 1);
   printf("Exporting ");
   printf("Paths: ");

   for(j=0;j<pl->Count;j++) {
      idx = strtoul(margv[j+1], 0, 0);
      if(idx >= dfc_impathlist_cnt) {
         printf("Usage: mexportpath: pathlist_index invalid\n");
         return(0);
      }
      bcopy((uint8_t *)&dfc_impathlist[idx], (uint8_t *)&pl->Paths[j],
         sizeof(MPL_PATH_LIST_ENTRY));

      printf("Index %d:%s ", idx, print_routeid(&pl->Paths[j].DevId));
   }
   printf("\n");

   if((rc = mplExportPaths(pl)) != MPL_STATUS_SUCCESS) {
      printf("mplExportPaths failed: %d\n", rc);
      return(0);
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mimportpath(
   CMDINFO_t *cp,
   void *p)
{
   MPL_GET_PATH_LIST * pl;
   int j, idx, rc;

   if (margc < 2) {
      printf("Usage: mimportpath [pathlist_index0 ... pathlist_indexN]\n");
      return(0);
   }
   if(dfc_expathlist_cnt == 0) {
      printf("Usage: mimportpath: Must run mexportlist first\n");
      return(0);
   }
   pl = (MPL_GET_PATH_LIST *)dfc_bigbuffer;

   pl->Count = (margc - 1);
   printf("Importing ");
   printf("Paths: ");

   for(j=0;j<pl->Count;j++) {
      idx = strtoul(margv[j+1], 0, 0);
      if(idx >= dfc_expathlist_cnt) {
         printf("Usage: mimportpath: pathlist_index invalid\n");
         return(0);
      }
      bcopy((uint8_t *)&dfc_expathlist[idx], (uint8_t *)&pl->Paths[j],
         sizeof(MPL_PATH_LIST_ENTRY));

      printf("Index %d:%s ", idx, print_routeid(&pl->Paths[j].DevId));
   }
   printf("\n");

   if((rc = mplImportPaths(pl)) != MPL_STATUS_SUCCESS) {
      printf("mplImportPaths failed: %d\n", rc);
      return(0);
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mdelpath(
   CMDINFO_t *cp,
   void *p)
{
   struct dfc_pctl { /* This duplicates MPL_GET_PATH_LIST */
      elxDiDeviceIdentifier routeid;
      int cnt;
      int pad;
      MPL_PATH_LIST_ENTRY pl[MPL_MAX_PATHS];
   } path;
   int j, idx, index, rc;

   if (margc < 3) {
      printf("Usage: mdelpath route_index [pathlist_index0 ... pathlist_indexN]\n");
      return(0);
   }

   index  = strtoul(margv[1], 0, 0);
   if(index >= dfc_Routes_cnt) {
      printf("mdelpath route_index not valid, run mroutelist\n");
      return(0);
   }
   bzero((uint8_t *)&path, sizeof(struct dfc_pctl));
   bcopy( (uint8_t *)&dfc_mpcfg[index].m_id.RouteId, (uint8_t *)&path.routeid,
         sizeof(elxDiDeviceIdentifier));


   if((path.cnt = (margc - 2)) >= MPL_MAX_PATHS) {
      printf("mdelpath too many paths\n");
      return(0);
   }

   printf("Deleting from Route: ");
   printf("%s\n", print_routeid(&path.routeid));
   printf("  Paths: ");

   for(j=0;j<path.cnt;j++) {
      idx = strtoul(margv[j+2], 0, 0);
      if((idx >= dfc_mpcfg[index].m_hdr.PathCount) || (idx >= MPL_MAX_PATHS)) {
         printf("Usage: mdelpath: pathlist_index invalid\n");
         return(0);
      }

      bcopy((uint8_t *)&dfc_mpcfg[index].m_paths[idx].m_id, (uint8_t *)&path.pl[j],
         sizeof(MPL_PATH_LIST_ENTRY));

      path.pl[j].FcpTarget = 1;
      path.pl[j].FailoverBitMask = MPL_PATH_DISCONNECTED;
      if(idx == 0)
         path.pl[j].Default = 1;
      else
         path.pl[j].Default = 0;

      printf("Index %d:%s ", idx, print_routeid(&path.pl[j].DevId));
   }
   printf("\n");

   if((rc = mplRemovePathListFromRoute((MPL_GET_PATH_LIST *)&path)) != MPL_STATUS_SUCCESS) {
      printf("mplRemovePathListFromRoute failed: %d\n", rc);
      return(0);
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mdelroute(
   CMDINFO_t *cp,
   void *p)
{
   int rc, index;
   elxDiDeviceIdentifier *dv;

   if (margc != 2) {
      printf("Usage: mdelroute route_index\n");
      return(0);
   }
   index  = strtoul(margv[1], 0, 0);

   if(index >= dfc_Routes_cnt) {
      printf("mdelroute route_index not valid, run mroutelist\n");
      return(0);
   }

   dv = (elxDiDeviceIdentifier *)dfc_bigbuffer;
   bcopy( (uint8_t *)&dfc_mpcfg[index].m_id.RouteId, (uint8_t *)dv,
      sizeof(elxDiDeviceIdentifier));

   printf("Deleting Route: ");
   printf("%s\n", print_routeid(dv));

   if((rc = mplDeleteRoute(dv)) != MPL_STATUS_SUCCESS) {
      printf("mplDeleteRoute failed: %d\n", rc);
      return(0);
   }
   return(0);
}
#endif /* MPULSE */


#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mgetpathctl(
   CMDINFO_t *cp,
   void *p)
{
   int rc, k, index;
   PMPL_PATH_CONTROL_HEADER pcp;

   if (margc != 2) {
      printf("Usage: mgetpathctl route_index\n");
      return(0);
   }
   index  = strtoul(margv[1], 0, 0);

   if(index >= dfc_Routes_cnt) {
      printf("mgetpathctl route_index not valid, run mroutelist\n");
      return(0);
   }

   pcp = (PMPL_PATH_CONTROL_HEADER)dfc_bigbuffer;
   if((rc = mplGetPathControl(&dfc_mpcfg[index].m_id.RouteId, pcp,
      MPL_MAX_PATH_CONTROL_SIZE)) != MPL_STATUS_SUCCESS) {
      printf("mplGetPathControl failed: %d\n", rc);
      return(0);
   }
   printf("Path Control for Route: ");
   printf("%s\n", print_routeid(&dfc_mpcfg[index].m_id.RouteId));
   printf("Type %d PathCnt %d autofb:%d InqHbt:%d ", pcp->Type,
      pcp->PathCount, pcp->AutoFailBack, pcp->InquiryHeartbeat);
   printf(" found:%d Static:%d prim:%d\n", pcp->Found,
      pcp->StaticConfig, pcp->StaticConfigPrimary);

   if(pcp->Type == MPL_PC_TYPE_FAILOVER) {
      PMPL_FAILOVER_CONTROL pp;

      pp = (PMPL_FAILOVER_CONTROL)pcp;
      printf("Primary Path  %s  SN Validation ",
	     print_routeid(&pp->PrimaryPathDevId));
      for(k=0;k<pp->PrimaryPathValidationId.m_len;k++) {
         printf("%02x", (unsigned char)(pp->PrimaryPathValidationId.un.m_InquiryData[k]));
      }
      printf("\n");
   }
   if(pcp->Type == MPL_PC_TYPE_TRAFFIC_RATIO) {
      PMPL_TRAFFIC_RATIO_CONTROL pp;

      pp = (PMPL_TRAFFIC_RATIO_CONTROL)pcp;
      printf("Traffic Ratios");
      for(k=0;k<pcp->PathCount;k++) {
         printf(": %d ",pp->PathRatios[k].Ratio);
      }
      printf("\n");
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
msetpathctl(
   CMDINFO_t *cp,
   void *p)
{
   int rc, i, idx;
   elxDiDeviceIdentifier *rvp;
   MPL_PATH_CONTROL_HEADER   *pc;
   MPL_FAILOVER_CONTROL      *fpc;
   MPL_TRAFFIC_RATIO_CONTROL *trpc;
   MPL_GET_PATH_LIST         *pl;
   char pcdata[MPL_MAX_PATH_CONTROL_SIZE];
   int pcflag;

   if (margc < 3) {
      printf("Usage: msetpathctl route_index pcflag pctype_info ...\n");
      return(0);
   }
   idx  = strtoul(margv[1], 0, 0);
   if(idx >= dfc_Routes_cnt) {
      printf("msetpathctl route_index not valid, run mroutelist\n");
      return(0);
   }
   rvp = &dfc_mpcfg[idx].m_id.RouteId;

   bzero((uint8_t *)&pcdata[0], MPL_MAX_PATH_CONTROL_SIZE);
   pc = (MPL_PATH_CONTROL_HEADER *)&pcdata[0];
   pcflag = strtoul(margv[2], 0, 0);
   pc->Type = (pcflag & 0xf);
   if(pc->Type == 0) {
      pc->Type = MPL_PC_TYPE_FAILOVER; /* Default */ 
   }
   switch(pc->Type) {
   case MPL_PC_TYPE_FAILOVER:

      if (margc > 4) {
         printf("Usage: msetpathctl route_index pcflag [primary_path_index]\n");
         return(0);
      }
      if (margc == 4)
         i = strtoul(margv[3], 0, 0);
      else
         i = 0;  /* Default Primary Path index */
      pl = (PMPL_GET_PATH_LIST)dfc_bigbuffer;

      bcopy( (uint8_t *)&dfc_mpcfg[idx].m_id.RouteId, (uint8_t *)&pl->RouteId,
         sizeof(elxDiDeviceIdentifier));
      pl->Count = (DFC_MAXNODES / 2); /* min of 2 paths per route */
      if((rc = mplGetPathList(pl)) != MPL_STATUS_SUCCESS) {
         printf("mplGetPathList failed: %d\n", rc);
         return(0);
      }

      fpc = (MPL_FAILOVER_CONTROL *)pc;
      bcopy((uint8_t *)&pl->Paths[i].DevId,
         (uint8_t *)&fpc->PrimaryPathDevId, sizeof(elxDiDeviceIdentifier));
      bcopy((uint8_t *)&pl->Paths[i].ValidationId,
         (uint8_t *)&fpc->PrimaryPathValidationId, sizeof(elxDiDeviceIdentifier));
      break;
   case MPL_PC_TYPE_TRAFFIC_RATIO:
      trpc = (MPL_TRAFFIC_RATIO_CONTROL *)pc;
      trpc->Count = 0;
      pl = (PMPL_GET_PATH_LIST)dfc_bigbuffer;
      bcopy( (uint8_t *)&dfc_mpcfg[idx].m_id.RouteId, (uint8_t *)&pl->RouteId,
         sizeof(elxDiDeviceIdentifier));
      pl->Count = (DFC_MAXNODES / 2); /* min of 2 paths per route */
      if((rc = mplGetPathList(pl)) != MPL_STATUS_SUCCESS) {
         printf("mplGetPathList failed: %d\n", rc);
         return(0);
      }
      for(i=3;i<margc;i++) {
         rc = strtoul(margv[i], 0, 0);
         trpc->PathRatios[i-3].Ratio = rc;
	 bcopy((uint8_t *)&pl->Paths[i-3].DevId,
	       (uint8_t *)&trpc->PathRatios[i-3].DeviceId,
	       sizeof(elxDiDeviceIdentifier));
	 bcopy((uint8_t *)&pl->Paths[i-3].ValidationId,
	       (uint8_t *)&trpc->PathRatios[i-3].ValidationId,
	       sizeof(elxDiDeviceIdentifier));
         trpc->Count++;
         if(i == (MPL_MAX_PATHS + 2))
            break;
      }
      break;
   }
   pc->AutoFailBack = 0;
   pc->InquiryHeartbeat = 0;
   if(pcflag & MPL_PC_CFG_AUTO_FAILBACK) {
      pc->AutoFailBack = 1;
   }
   if(pcflag & MPL_PC_CFG_INQUIRY_HEARTBEAT) {
      pc->InquiryHeartbeat = 1;
   }
   if(pcflag & MPL_PC_CFG_FAIL_MASK) {
      if(pcflag & MPL_PC_CFG_FAIL_ABORT_TASK_SET) {
         pc->FailoverAbortTaskSet = 1;
      }
      else if(pcflag & MPL_PC_CFG_FAIL_LUN_RESET) {
         pc->FailoverLunReset = 1;
      }
      else if(pcflag & MPL_PC_CFG_FAIL_TARGET_RESET) {
         pc->FailoverTgtReset = 1;
      }
   }
   else {
      pc->FailoverAbortTaskSet = 1;  /* Default */
   }

   printf("Changing path control for route\n");
   if((rc = mplSetPathControl(rvp, pc)) != MPL_STATUS_SUCCESS) {
      printf("mplSetPathControl failed: %d\n", rc);
      return(0);
   }
   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mgetpathstat(
   CMDINFO_t *cp,
   void *p)
{
   int rc, k, ridx, pidx;
   PMPL_PATH_STATUS psp;
   MPL_PATH_LIST_ENTRY *pe;

   if (margc != 3) {
      printf("Usage: mgetpathstat route_index path_index\n");
      return(0);
   }
   ridx  = strtoul(margv[1], 0, 0);
   pidx  = strtoul(margv[2], 0, 0);

   if(ridx >= dfc_Routes_cnt) {
      printf("mgetpathstat route_index not valid, run mroutelist\n");
      return(0);
   }
   pe = &dfc_mpcfg[ridx].m_paths[pidx].m_id;

   if((pe->ValidationId.m_len == 0) || (pe->DevId.m_len == 0)) {
      printf("mgetpathstat path_index not valid, run mpathlist\n");
      return(0);
   }

   psp = (PMPL_PATH_STATUS)dfc_bigbuffer;
   if((rc = mplGetPathStatus(&pe->DevId, &pe->ValidationId, psp)) != MPL_STATUS_SUCCESS) {
      printf("mplGetPathStatus failed: %d\n", rc);
      return(0);
   }
   printf("Path Status for Device: %s   ", print_routeid(&pe->DevId));

   for(k=0;k<pe->ValidationId.m_len;k++) {
      printf("%02x", (unsigned char)(pe->ValidationId.un.m_InquiryData[k]));
   }
   printf("\n");
   printf("State %d Active %d Qdepth %d TratioCnt %d\n", psp->State,
      psp->ActiveCount, psp->QueueDepth, psp->RatioCount);
   printf("Cmd %d Cmpl %d QFull %d Error %d\n",
      psp->TotalCmdCount, psp->TotalCmplCount, psp->QueueFullCount,
      psp->ErrorCount);

   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
mgetpathinfo(
   CMDINFO_t *cp,
   void *p)
{
   int rc, k, ridx, pidx;
   PMPL_PATH_INFO pip;
   MPL_PATH_LIST_ENTRY *pe;

   if (margc != 3) {
      printf("Usage: mgetpathinfo route_index path_index\n");
      return(0);
   }
   ridx  = strtoul(margv[1], 0, 0);
   pidx  = strtoul(margv[2], 0, 0);

   if(ridx >= dfc_Routes_cnt) {
      printf("mgetpathstat route_index not valid, run mroutelist\n");
      return(0);
   }
   pe = &dfc_mpcfg[ridx].m_paths[pidx].m_id;

   if((pe->ValidationId.m_len == 0) || (pe->DevId.m_len == 0)) {
      printf("mgetpathinfo path_index not valid, run mpathlist\n");
      return(0);
   }

   pip = (PMPL_PATH_INFO)dfc_bigbuffer;
   bzero((void *)pip, sizeof(MPL_PATH_INFO));

   if((rc = mplGetPathInfo(&pe->DevId, &pe->ValidationId, pip)) != MPL_STATUS_SUCCESS) {
      printf("mplGetPathInfo failed: %d\n", rc);
      return(0);
   }
   printf("Path Info for Device: %s   ", print_routeid(&pe->DevId));

   for(k=0;k<pe->ValidationId.m_len;k++) {
      printf("%02x", (unsigned char)(pe->ValidationId.un.m_InquiryData[k]));
   }
   printf("\n");
   printf("Port %d Bus %d Target %d Lun %d\n", pip->Port,
      pip->Bus, pip->Target, pip->Lun);
   printf("Route len %d: ", pip->RouteId.m_len);
   printf("%s\n", print_routeid(&pip->RouteId));
   printf("Vendor: %s  Product: %s  Rev: %s\n", pip->Vendor, pip->Product,
   pip->Rev);

   return(0);
}
#endif /* MPULSE */

#ifdef MPULSE
/*!
 * Doxygen routine header
 */

int
msetpathstat(
   CMDINFO_t *cp,
   void *p)
{
   int rc, ridx, pidx;
   PMPL_PATH_STATUS psp;
   MPL_PATH_LIST_ENTRY *pe;

   if (margc != 4) {
      printf("Usage: msetpathstat route_index path_index [online offline zero]\n");
      return(0);
   }
   ridx  = strtoul(margv[1], 0, 0);
   pidx  = strtoul(margv[2], 0, 0);

   if(ridx >= dfc_Routes_cnt) {
      printf("mgetpathstat route_index not valid, run mroutelist\n");
      return(0);
   }
   pe = &dfc_mpcfg[ridx].m_paths[pidx].m_id;

   if((pe->ValidationId.m_len == 0) || (pe->DevId.m_len == 0)) {
      printf("mgetpathstat path_index not valid, run mpathlist\n");
      return(0);
   }

   psp = (PMPL_PATH_STATUS)dfc_bigbuffer;
   bzero((void *)psp, sizeof(MPL_PATH_STATUS));

   if (strcmp(margv[3], "offline") == 0) {
      psp->Mask = MPL_PATH_MASK_STATE;
      psp->State = MPL_PS_STATE_OFFLINE;
   } else if (strcmp(margv[3], "online") == 0) {
      psp->Mask = MPL_PATH_MASK_STATE;
      psp->State = MPL_PS_STATE_ONLINE;
   } else if (strcmp(margv[3], "zero") == 0) {
      psp->Mask = MPL_PATH_MASK_SETABLE_ALL;
      psp->ErrorCount = 0;
      psp->QueueFullCount = 0;
      psp->TotalCmdCount = 0;
      psp->TotalCmplCount = 0;
   } else {
      printf("Usage: msetpathstat route_index path_index [online offline zero]\n");
      return(0);
   }

   if((rc = mplSetPathStatus(&pe->DevId, &pe->ValidationId, psp)) !=
      MPL_STATUS_SUCCESS) {
      printf("mplSetPathStatus failed: %d\n", rc);
      return(0);
   }
   return(0);
}


MPL_CFGPARAM cfgparam[MPL_MAX_CFG_PARAM];

/*! mgetcfg
 *
 * \pre
 * \post
 * \param  cp
 * \param  p
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 *    Displays configurtion parameters for the adapter 
 */
int
mgetcfg(
   CMDINFO_t *cp,
   void *p)
{
   int  i, num;
   int  h;

   switch (margc)  {
   case 1:
      h = 0;
      break;
   case 2:
      h = 1;
      break;
   default:
      printf("Usage: mgetcfg [help]\n");
      return(0);
   }

   if ((num=mplGetCfgParam(cfgparam)) == 0) {
      fprintf(stderr,dfc_err_cannot_issue_cmd, errno);
      return(0);
   }
   gotcfg = num;
   /* Display results */
   printf("Low        High       Default    Current    Name / flag / chgstate\n");
   for (i = 0; i < num; i++) {
      printf("0x%08x 0x%08x 0x%08x 0x%08x %s %d %d\n",
      cfgparam[i].a_low, cfgparam[i].a_hi, cfgparam[i].a_default, cfgparam[i].a_current,
      cfgparam[i].a_string, cfgparam[i].a_flag, cfgparam[i].a_changestate);
      if(h)
         printf("%s\n", cfgparam[i].a_help);

      if ((i % num_lines) == (num_lines-1)) {
         if( promptContinueOrAbort())
            break;  /* abort */
         printf("Low        High       Default    Current    Name / flag / chgstate\n");
      }
   }

   return(0);
}

/*! msetcfg
 *
 * \pre
 * \post
 * \param  cp
 * \param  p
 *
 * \return int      Return Code  0 = success
 *
 * \b Description:
 *    Sets configurtion parameters for the adapter 
 */
int
msetcfg(
   CMDINFO_t *cp,
   void *p)
{
   uint32_t  data;
   uint32_t  offset;
   int  i;
   char *ptr;

   if (margc != 3) {
      printf("Usage: setcfg cfgstring value\n");
      return(0);
   }

   if (gotcfg <=  0) {
      printf("Must run getcfg first\n");
      return(0);
   }

   ptr = (char *)margv[1];
   data = strtoul(margv[2], 0, 0);

   for (i = 0; i < gotcfg; i++) {
      if (strlen(ptr) == strlen((char *)cfgparam[i].a_string))
         if (strcmp(ptr, (char *)cfgparam[i].a_string) == 0)
            break;
   }
   offset = i;

   if (mplSetCfgParam(offset,  data)) {
      fprintf(stderr,dfc_err_cannot_issue_cmd, errno);
      return(0);
   }
   return(0);
}

#endif /* MPULSE */



/*!
 * Doxygen routine header
 */
/* This routine converts a 48 bit IEEE address
 * into "dot notation" for printing. 
 * Note, buf must point to at least 18 bytes of free storage.
 */
void dfc_mac_sprintf(
   register uint8_t *IEEEp,
   register char   *buf)
{
   int i;
   static char  digits[] = "0123456789abcdef";

   for (i = 0; i < 8; i++) {
      *buf++ = digits[*IEEEp >> 4];
      *buf++ = digits[*IEEEp++ & 0xf];
      *buf++ = ':';
   }
   *--buf = 0;
   return;
}


/*!
 * Doxygen routine header
 */
/* This routine displays the user prompt 
 * Hit <CR> to continue (x to abort):
 * and waits for user input. */

static int
promptContinueOrAbort( void)
{
   char c;

   printf("\n Hit <CR> to continue (x to abort): ");
   c = getchar();
   printf("\n");
   if (c == 'x') {
#ifndef _LH6000
      while ((c = getchar()) != '\n')
         ;
#endif
      return(1); /* abort */
   }
   return(0);    /* continue */
}

