/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* mpl_lib.c: MultiPulse for Linux API library.
 * $Id: mpl_lib.c 1461 2008-03-13 19:56:11Z clajeunesse $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdint.h>
#include <errno.h>
#include <sys/types.h>
#include <regex.h>
#include <dirent.h>
#include <unistd.h>

#define SIZEOF_WWPN   8	  /* Number of bytes in a FC WWPN */
#define PATH_ID_SIZE 100  /* 32 byte inquiry data in xx:yy:zz:... form */
#include "mplib.h"
#include "mpapi.h"
#include "mpdevid.h"

#define NUM_ENTRIES(x) (sizeof(x) / sizeof(x[0]))
#define min(a, b) ((a) < (b) ? (a) : (b))
#define MPL_CONFIG_NODE "/proc/scsi/lpfcmpl/config"
#define MPL_CONTROL_NODE "/proc/scsi/lpfcmpl/control"
#define MPL_VERSION_NODE "/proc/scsi/lpfcmpl/version"
#define SYS_SCSI_HOSTS "/sys/class/scsi_host"

static char mpl_library_version[] = "3.0.30";

elxDiDeviceIdentifier mplNullIdentifier = { 0 };
static int initialized;
typedef int (*match_list_test)(char *line_buffer, regmatch_t *match_list);

static regex_t hex_wwn_re;
static regex_t colon_separated_wwn_re;
static regex_t version_re;
static regex_t port_name_re;
static regex_t port_state_re;
static regex_t device_re;
static regex_t path_re;
static regex_t host_re;
static regex_t cfg_param1_re;
static regex_t cfg_param2_re;
static regex_t cfg_param3_1_re;
static regex_t cfg_param3_2_re;
static regex_t proc_scsi_host_re;

static int mpl_current_host = 0;

/* In order to provide for evolving /proc/scsi/lpfcmpl/config syntax,
 * particularly for the device and path representations, we indulge in
 * a painful bit of indirection to keep the parsing routines from
 * getting too ugly.
 *
 * First, we define an enumeration (e.g. device_match_index, below)
 * that describes the indicies into an array.  The array contains the
 * regex field index (parenthesized expression count) for the match
 * list entry that represents the fields of interest.
 *
 * Secondly, we define regex strings and a corresponding array of
 * indicies (e.g. device_match_indicies, below) to parse a
 * particular version of the .../config file.
 *
 * The array associated with the version we expect to read is
 * indirectly referenced by an int ** (e.g. device_match_indices).
 *
 * Finally the whole mess is masked by a set of macros that select the
 * appropriate field index after the regex has matched the input line
 * (e.g. DEVICE_MATCH_HOST, and friends)
 *
 */

enum device_match_index {
	DEVICE_MATCH_HOST_ENTRY,
	DEVICE_MATCH_TARGET_ENTRY,
	DEVICE_MATCH_LUN_ENTRY,
	DEVICE_SCSI_TYPE_ENTRY,
	DEVICE_MATCH_VENDOR_ENTRY,
	DEVICE_MATCH_PRODUCT_ENTRY,
	DEVICE_MATCH_REV_ENTRY,
	DEVICE_MATCH_SERIAL_ENTRY,
	DEVICE_MATCH_ORIGIN_ENTRY,
	DEVICE_MATCH_STATE_ENTRY,
	DEVICE_MATCH_AUTOFAIL_ENTRY,
	DEVICE_MATCH_STATIC_ENTRY,
	DEVICE_MATCH_COUNT_ENTRY,
};

static char device_re_v2[] = {
	"device[ :=]*"
	"([[:digit:]:U]+)[/:]"	/* adapter */
	"([[:digit:]:U]+)[/:]"	/* target */
	"([[:digit:]:U]+)"	/* lun */
	"[[:space:]]+"		/* delimiter */
	"([^,]*),"		/* scsi_type */
	"[[:space:]]+"
	"([^,]*),"		/* vendor */
	"[[:space:]]+"
	"([^,]*),"		/* product */
	"[[:space:]]+"
	"([^,]*),"		/* rev */
	"[[:space:]]+"
	"([[:xdigit:]:]*),"	/* serial number */
	"[[:space:]]+"
	"(configured|unconfigured|unavailable)"
	"(,"			/* Optional only provided */
	"[[:space:]]+"		/* for configured devices */
	"(online|failing|failed|notfound|unknown),"
	"[[:space:]]+"
	"(autofail|noautofail|ratio|lunq|tgtq|hbaq|time|unknown),"
	"[[:space:]]+"
	"(static_p|static|dynamic),"
	")?"
};

static int device_match_indicies_v2[] = {
	[DEVICE_MATCH_HOST_ENTRY] = 1,
	[DEVICE_MATCH_TARGET_ENTRY] = 2,
	[DEVICE_MATCH_LUN_ENTRY] = 3,
	[DEVICE_SCSI_TYPE_ENTRY] = 4,
	[DEVICE_MATCH_VENDOR_ENTRY] = 5,
	[DEVICE_MATCH_PRODUCT_ENTRY] = 6,
	[DEVICE_MATCH_REV_ENTRY] = 7,
	[DEVICE_MATCH_SERIAL_ENTRY] = 8,
	[DEVICE_MATCH_ORIGIN_ENTRY] = 9,
	[DEVICE_MATCH_STATE_ENTRY] = 11,
	[DEVICE_MATCH_AUTOFAIL_ENTRY] = 12,
	[DEVICE_MATCH_STATIC_ENTRY] = 13,
	[DEVICE_MATCH_COUNT_ENTRY] = 13,
};

static int *device_match_indicies = device_match_indicies_v2;

#define DEVICE_MATCH_HOST    device_match_indicies[DEVICE_MATCH_HOST_ENTRY]
#define DEVICE_MATCH_TARGET  device_match_indicies[DEVICE_MATCH_TARGET_ENTRY]
#define DEVICE_MATCH_LUN     device_match_indicies[DEVICE_MATCH_LUN_ENTRY]
#define DEVICE_SCSI_TYPE     device_match_indicies[DEVICE_SCSI_TYPE_ENTRY]
#define DEVICE_MATCH_VENDOR  device_match_indicies[DEVICE_MATCH_VENDOR_ENTRY]
#define DEVICE_MATCH_PRODUCT device_match_indicies[DEVICE_MATCH_PRODUCT_ENTRY]
#define DEVICE_MATCH_REV     device_match_indicies[DEVICE_MATCH_REV_ENTRY]
#define DEVICE_MATCH_SERIAL  device_match_indicies[DEVICE_MATCH_SERIAL_ENTRY]
#define DEVICE_MATCH_ORIGIN  device_match_indicies[DEVICE_MATCH_ORIGIN_ENTRY] 
#define DEVICE_MATCH_STATE   device_match_indicies[DEVICE_MATCH_STATE_ENTRY]
#define DEVICE_MATCH_AUTOFAIL device_match_indicies[DEVICE_MATCH_AUTOFAIL_ENTRY]
#define DEVICE_MATCH_STATIC  device_match_indicies[DEVICE_MATCH_STATIC_ENTRY]
#define DEVICE_MATCH_COUNT   device_match_indicies[DEVICE_MATCH_COUNT_ENTRY]

enum path_match_index {
	PATH_MATCH_HOST_ENTRY,
	PATH_MATCH_TARGET_ENTRY,
	PATH_MATCH_LUN_ENTRY,
	PATH_MATCH_ORIGIN_ENTRY,
	PATH_MATCH_STATE_ENTRY,
	PATH_MATCH_PRIMARY_ENTRY,
	PATH_MATCH_RATIO_ENTRY,
	PATH_MATCH_WWPN_ENTRY,
	PATH_MATCH_COUNT_ENTRY,
};

static char path_re_v2[] = {
	"path[ :=]*"
	"([[:digit:]:U]+)[/:]"	/* adapter */
	"([[:digit:]:U]+)[/:]"	/* target */
	"([[:digit:]:U]+),"	/* lun */
	"[[:space:]]+"
	"(configured|unconfigured|unavailable)"
	"(,"			/* Optional.  Only provided */
	"[[:space:]]+"		/* for configured paths */
	"(online|standby|failed|disabled|notfound|unknown),"
	"[[:space:]]+"
	"(primary|secondary|traffic|unknown|([[:digit:]]+))),"
	"[[:space:]]+"
	"(dynamic|static)"
	")?,"
	"[[:space:]]+"
	"(active_cntrlr|inactive_cntrlr|cntrlr_unknown),"
	"[[:space:]]+"
	"WWPN[[:space:]]+"
	"(([[:xdigit:]]{1,2}:?){8})"
};

static int path_match_indicies_v2[] = {
	[PATH_MATCH_HOST_ENTRY] = 1,
	[PATH_MATCH_TARGET_ENTRY] = 2,
	[PATH_MATCH_LUN_ENTRY] = 3,
	[PATH_MATCH_ORIGIN_ENTRY] = 4,
	[PATH_MATCH_STATE_ENTRY] = 6,
	[PATH_MATCH_PRIMARY_ENTRY] = 7,
	[PATH_MATCH_RATIO_ENTRY] = 9,
	[PATH_MATCH_WWPN_ENTRY] = 11,
	[PATH_MATCH_COUNT_ENTRY] = 13,
};

#define PATH_MATCH_HOST    path_match_indicies[PATH_MATCH_HOST_ENTRY]
#define PATH_MATCH_TARGET  path_match_indicies[PATH_MATCH_TARGET_ENTRY]
#define PATH_MATCH_LUN     path_match_indicies[PATH_MATCH_LUN_ENTRY]
#define PATH_MATCH_ORIGIN  path_match_indicies[PATH_MATCH_ORIGIN_ENTRY]
#define PATH_MATCH_STATE   path_match_indicies[PATH_MATCH_STATE_ENTRY]
#define PATH_MATCH_PRIMARY path_match_indicies[PATH_MATCH_PRIMARY_ENTRY]
#define PATH_MATCH_RATIO   path_match_indicies[PATH_MATCH_RATIO_ENTRY]
#define PATH_MATCH_WWPN    path_match_indicies[PATH_MATCH_WWPN_ENTRY]
#define PATH_MATCH_COUNT   path_match_indicies[PATH_MATCH_COUNT_ENTRY]

static int *path_match_indicies = path_match_indicies_v2;

static int
mpl_re_compile(regex_t *rep, char *regexp)
{
	char buffer[64];
	int  rc;

	rc = regcomp(rep, regexp, REG_EXTENDED | REG_ICASE);
	if (rc) {
		regerror(rc, rep, buffer, sizeof(buffer));
		fprintf(stderr, "mpllib: regex compile error: %s: %s\n",
			buffer, regexp);
	}
	return rc;
}

static int
init_regexp(void)
{
	int rc;

#define WWN_MATCH_DIGITS  2
	rc = mpl_re_compile(&hex_wwn_re, "(0x)?([[:xdigit:]]{1,16})");
	if (rc)
		return rc;
#define WWN_MATCH_COLONS  1
	rc = mpl_re_compile(&colon_separated_wwn_re,
			    "([[:xdigit:]]{1,2}):"
			    "([[:xdigit:]]{1,2}):"
			    "([[:xdigit:]]{1,2}):"
			    "([[:xdigit:]]{1,2}):"
			    "([[:xdigit:]]{1,2}):"
			    "([[:xdigit:]]{1,2}):"
			    "([[:xdigit:]]{1,2}):"
			    "([[:xdigit:]]{1,2})");
	if (rc)
		goto free_hex;

#define PORT_NAME_MATCH_WWN   1
	rc = mpl_re_compile(&port_name_re, "^Portname: *"
			    "([[:xdigit:]:]+)"		/* wwn */
			    "[[:space:]]");
	if (rc)
		goto free_colon;

#define PORT_STATE_MATCH_STATE 1
	rc = mpl_re_compile(&port_state_re, "^Link "
			    "(Down|Up - Discovery|Up - Ready|Up)");
	if (rc)
		goto free_port_name;

	device_match_indicies = device_match_indicies_v2;
	path_match_indicies = path_match_indicies_v2;
	rc = mpl_re_compile(&device_re, device_re_v2);
	if (rc)
		goto free_port_state;
	rc = mpl_re_compile(&path_re, path_re_v2);
	if (rc)
		goto free_device;

#define HOST_MATCH_WWPN 1
	rc = mpl_re_compile(&host_re, "^host\\([[:digit:]]\\): .*"
			    "WWPN[[:space:]]+"
			    "(([[:xdigit:]]{1,2}:?){8})");
	if (rc)
		goto free_path;

#define CFG_MATCH_LOG_VERBOSE    1
#define CFG_MATCH_HBEAT_INTERVAL 2
#define CFG_MATCH_HBEAT_FAIL     3
#define CFG_MATCH_HBEAT_RETRY    4
	rc = mpl_re_compile(&cfg_param1_re, "^module: .*verbose\\(x([[:xdigit:]]+)\\).*"
		"interval\\(([[:digit:]]+)\\).*"
		"fail\\(([[:digit:]]+)\\).*"
		"retry\\(([[:digit:]]+)\\)");
	if (rc)
		goto free_host;

#define CFG_MATCH_AUTO_CONFIGURE    1
#define CFG_MATCH_AUTO_BALANCE      2
	rc = mpl_re_compile(&cfg_param2_re, "^   AUTO: .*configure\\(([[:digit:]]+)\\).*"
		"balance\\(([[:digit:]]+)\\)");
	if (rc)
		goto free_param1;

#define CFG_MATCH_ENABLE_TGT_QDEPTH 1
#define CFG_MATCH_HANDLE_QFULL      2
	rc = mpl_re_compile(&cfg_param3_2_re, "^ QDEPTH: .*enable_tgt\\(([[:digit:]]+)\\).*"
		"handle_qfull\\(([[:digit:]]+)\\)");
	if (rc)
		goto free_param2;

	rc = mpl_re_compile(&cfg_param3_1_re, "^ QDEPTH: .*enable_tgt\\(([[:digit:]]+)\\).*");
	if (rc)
		goto free_param3_2;

#define PROC_SCSI_HOST    1
#define PROC_SCSI_TARGET  2
#define PROC_SCSI_LUN     3
	rc = mpl_re_compile(&proc_scsi_host_re,
			    "^Host: scsi([[:digit:]]+) Channel: 00 "
			    "Id: ([[:digit:]]+) Lun: ([[:digit:]]+)");
	if (rc)
		goto free_queue_depth;

	return rc;

free_queue_depth:
	regfree(&cfg_param3_1_re);
free_param3_2:
	regfree(&cfg_param3_2_re);
free_param2:
	regfree(&cfg_param2_re);
free_param1:
	regfree(&cfg_param1_re);
free_host:
	regfree(&host_re);
free_path:
	regfree(&path_re);
free_device:
	regfree(&device_re);
free_port_state:
	regfree(&port_state_re);
free_port_name:
	regfree(&port_name_re);
free_colon:
	regfree(&colon_separated_wwn_re);
free_hex:
	regfree(&hex_wwn_re);
	return rc;
}

#define MATCH_COUNT_MAX ((DEVICE_MATCH_COUNT > PATH_MATCH_COUNT	 \
			  ? DEVICE_MATCH_COUNT			 \
			  : PATH_MATCH_COUNT) + 1)

static int fetch_match(char *buffer, size_t size, char *input,
		       regmatch_t *match)
{
	int n;

	if (match->rm_so == -1) {
		return 0;
	}
	memset(buffer, 0, size);
	n = min(match->rm_eo - match->rm_so, size - 1);
	memcpy(buffer, input + match->rm_so, n);
	buffer[n] = 0;
	return 1;
}

static void get_current_host(char *line_buffer)
{
	if (!memcmp(line_buffer, "host(", 5)) {
		mpl_current_host = atoi(&line_buffer[5]);
	}
	return;
}

static int fetch_match_integer(char *input, regmatch_t *match, int match_index)
{
	char parameter[32];

	fetch_match(parameter, sizeof(parameter), input,
		    match + match_index);
	return strtol(parameter, NULL, 0);
}

static int
parse_hex_wwn(unsigned char wwn[8], char *wwn_buffer)
{
	regmatch_t match_list[3];
	int  i, rc;

	memset(wwn, 0, 8);
	rc = regexec(&hex_wwn_re, wwn_buffer, NUM_ENTRIES(match_list),
		     match_list, 0);
	if (!rc) {
		char wwn_digits[32];

		if (fetch_match(wwn_digits, sizeof(wwn_digits), wwn_buffer,
				match_list + WWN_MATCH_DIGITS)) {
			if (i < 8) {
				unsigned long long wwn_binary;
				wwn_binary = strtoull(wwn_digits, NULL, 16);
				memcpy(wwn, &wwn_binary, 8);
			}
		}
	}
	return rc;

}


static int
parse_colon_separated_wwn(unsigned char wwn[8], char *wwn_buffer)
{
	regmatch_t match_list[10];
	int  i, rc;

	memset(wwn, 0, 8);
	rc = regexec(&colon_separated_wwn_re, wwn_buffer,
		     NUM_ENTRIES(match_list), match_list, 0);
	if (!rc) {
		for (i = 0; i < 8; ++i) {
			char octet[32];

			if (fetch_match(octet, sizeof(octet), wwn_buffer,
					match_list + i + WWN_MATCH_COLONS)) {
				wwn[i] = strtoul(octet, NULL, 16);
			}
		}
	}
	return rc;
}

static int
parse_wwn(unsigned char wwn[8], char *wwn_buffer)
{
	if (!parse_colon_separated_wwn(wwn, wwn_buffer))
		return 0;
	return parse_hex_wwn(wwn, wwn_buffer);
}

static inline unsigned
xtobin(unsigned char c)
{
	return c >= '0' && c <= '9' ? c - '0' : tolower(c) - 'a' + 10;
}

static void
crack_inquiry_encoding(unsigned char *dest, char *src, size_t size)
{
	while (size && *src) {
		if (isxdigit(src[0]) && isxdigit(src[1])) {
			*dest++ = (xtobin(src[0]) << 4) + xtobin(src[1]);
			--size;
			src += 2;
		}
		if (*src == ':') {
			++src;
		}
	}
	while (size) {
		*dest++ = 0;
		--size;
	}
}

static int
format_inqsn(char *buf, size_t size, unsigned char *serial,
	     size_t serial_size)
{
	char *cp = buf;
	size_t  n;
	size_t  i;

	for (i = 0; size && i < serial_size; ++i) {
		if (cp > buf && size > 1) {
			*cp++ = ':';
			--size;
		}
		n = snprintf(cp, size, "%02x", serial[i]);
		if (n > size) {
			n = size;
		}
		size -= n;
		cp += n; 
	}
	return cp - buf;
}


struct port_desc {
	unsigned char port_wwpn[8];
};
static struct port_desc *port_cache;
static int  port_count;

static MPL_STATUS build_port_cache()
{
	FILE *fp;
	char line_buffer[256];
	char port_name[32];
	regmatch_t match_list[5];
	MPL_STATUS status = MPL_STATUS_SUCCESS;
	int  rc, i;

	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		status = MPL_STATUS_NO_DEVICE;
		goto out;
	}
				/* Count host entries */
	i = 0;
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		rc = regexec(&host_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc && fetch_match(port_name, sizeof(port_name),
				       line_buffer,
				       match_list + HOST_MATCH_WWPN)) {
			++i;
		}
	}
	rewind(fp);
	port_count = i;
	port_cache = malloc(port_count * sizeof(*port_cache));
	i = 0;
	while (i < port_count && fgets(line_buffer, sizeof(line_buffer), fp)) {
		rc = regexec(&host_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc && fetch_match(port_name, sizeof(port_name),
				       line_buffer,
				       match_list + HOST_MATCH_WWPN)) {
			parse_wwn(port_cache[i].port_wwpn, port_name);
			++i;
		}
	}
	fclose(fp);
out:
	return status;
}

static void
make_scsi_id(struct ElxDiDevId  *device, char *line_buffer,
	     regmatch_t *match_list)
{
	char ascii[sizeof(device->un.m_InquiryData) * 3];

	memset(device, 0, sizeof(*device));
	fetch_match(ascii, sizeof(ascii), line_buffer,
		    match_list + DEVICE_MATCH_SERIAL);
	crack_inquiry_encoding(device->un.m_InquiryData, ascii,
			       sizeof(device->un.m_InquiryData));
	device->m_len  = (strlen(ascii) + 1) / 3;
	device->m_dflags = ELX_DI_DEVID_FLAGS_BINARY;
	device->m_type = ELX_DI_DEVID_INQSN;
}

static int
is_scsi_id_eq(struct ElxDiDevId  *device, char *line_buffer,
	      regmatch_t *match_list)
{
	char ascii[sizeof(device->un.m_InquiryData) * 3];
	unsigned char inquiry[sizeof(device->un.m_InquiryData)];

	fetch_match(ascii, sizeof(ascii), line_buffer,
		    match_list + DEVICE_MATCH_SERIAL);
	crack_inquiry_encoding(inquiry, ascii, sizeof(inquiry));
	return !memcmp(device->un.m_InquiryData, inquiry, device->m_len);
}

static void
make_long_route_id(struct ElxDiDevId  *device, char *line_buffer,
		   regmatch_t *match_list)
{
	char ascii[sizeof(device->un.m_InquiryData) * 3];

	memset(device, 0, sizeof(*device));
	fetch_match(ascii, sizeof(ascii), line_buffer,
		    match_list + DEVICE_MATCH_SERIAL);
	crack_inquiry_encoding(device->un.m_InquiryData, ascii,
			       sizeof(device->un.m_InquiryData));
	device->m_len  = (strlen(ascii) + 1) / 3;
	device->m_dflags = ELX_DI_DEVID_FLAGS_BINARY;
	device->m_type = ELX_DI_DEVID_INQSN;
}

static void
make_route_id(struct ElxDiDevId  *device, char *line_buffer,
	      regmatch_t *match_list)
{
	make_long_route_id(device, line_buffer, match_list);
}

static void
make_scsi_path_entry(struct ElxDiDevId  *path, char *line_buffer,
		     regmatch_t *match_list)
{
	char number[16];

	memset(path, 0, sizeof(*path));
	if (fetch_match(number, sizeof(number), line_buffer,
			match_list + PATH_MATCH_HOST)) {
		path->un.m_ScsiId.m_adapter = strtol(number, NULL, 0);
	}
	if (fetch_match(number, sizeof(number), line_buffer,
			match_list + PATH_MATCH_TARGET)) {
		path->un.m_ScsiId.m_target = strtol(number, NULL, 0);
	}
	if (fetch_match(number, sizeof(number), line_buffer,
			match_list + PATH_MATCH_LUN)) {
		path->un.m_ScsiId.m_lun = strtol(number, NULL, 0);
	}
	path->m_len  = (4 * sizeof(uint16_t));  /* straight from 4.1 ioctl */
	path->m_dflags = ELX_DI_DEVID_FLAGS_BINARY;
	path->m_type = ELX_DI_DEVID_SCSI_NEXUS;
}

static void
make_long_path_entry(struct ElxDiDevId  *path,
		     unsigned char target_wwpn[8],
		     char *line_buffer, regmatch_t *match_list)
{
	char number[16];
	long host;

	memset(path, 0, sizeof(*path));
	if (fetch_match(number, sizeof(number), line_buffer,
			match_list + PATH_MATCH_HOST)) {
		if (number[0] == 'U')
			host = mpl_current_host;
		else
			host = strtol(number, NULL, 0);
	}
	memcpy(path->un.m_NexusLong.m_adapter_wwpn, port_cache[host].port_wwpn,
	       sizeof(path->un.m_NexusLong.m_adapter_wwpn));
	memcpy(path->un.m_NexusLong.m_target_wwpn, target_wwpn,
	       sizeof(path->un.m_NexusLong.m_target_wwpn));
	if (fetch_match(number, sizeof(number), line_buffer,
			match_list + PATH_MATCH_LUN)) {
		unsigned int lun;

		lun = strtol(number, NULL, 0);
		memset(path->un.m_NexusLong.m_lun, 0,
		       sizeof(path->un.m_NexusLong.m_lun));
		path->un.m_NexusLong.m_lun[0] = lun;
	}
	path->m_len  = 24;
	path->m_dflags = ELX_DI_DEVID_FLAGS_BINARY;
	path->m_type = ELX_DI_DEVID_NEXUS_LONG;
}

static void
make_path_entry(struct ElxDiDevId  *path, char *line_buffer,
		regmatch_t *match_list)
{
	char target_name[64];
	unsigned char target_wwpn[8];

	if (PATH_MATCH_WWPN && fetch_match(target_name, sizeof(target_name),
					   line_buffer,
					   match_list + PATH_MATCH_WWPN)) {
		parse_wwn(target_wwpn, target_name);
		make_long_path_entry(path, target_wwpn, line_buffer,
				     match_list);
	} else {
		make_scsi_path_entry(path, line_buffer, match_list);
	}
}

static int
is_long_path_eq(struct ElxDiDevId  *path, char *line_buffer,
		regmatch_t *match_list)
{
	char target_name[64];
	char number[16];
	unsigned char target_wwpn[8];
	long l;

	if (!fetch_match(number, sizeof(number), line_buffer,
			 match_list + PATH_MATCH_HOST)) {
		return 0;
	}
	l = strtol(number, NULL, 0);
	if (memcmp(path->un.m_NexusLong.m_adapter_wwpn,
		   port_cache[l].port_wwpn,
		   sizeof(path->un.m_NexusLong.m_adapter_wwpn))) {
		return 0;
	}
	if (!fetch_match(target_name, sizeof(target_name), line_buffer,
			 match_list + PATH_MATCH_WWPN)) {
		return 0;
	}
	parse_wwn(target_wwpn, target_name);
	if (memcmp(path->un.m_NexusLong.m_target_wwpn, target_wwpn,
		   sizeof(path->un.m_NexusLong.m_target_wwpn))) {
		return 0;
	}
	if (!fetch_match(number, sizeof(number), line_buffer,
			 match_list + PATH_MATCH_LUN)) {
		return 0;
	}
	l = strtol(number, NULL, 0);

	return path->un.m_NexusLong.m_lun[0] == l;
}

static int
is_scsi_path_eq(struct ElxDiDevId  *path, char *line_buffer,
		regmatch_t *match_list)
{
	char number[16];

	if (!fetch_match(number, sizeof(number), line_buffer,
			 match_list + PATH_MATCH_HOST) ||
	    path->un.m_ScsiId.m_adapter != strtol(number, NULL, 0))
		return 0;
	if (!fetch_match(number, sizeof(number), line_buffer,
			 match_list + PATH_MATCH_TARGET) ||
	    path->un.m_ScsiId.m_target != strtol(number, NULL, 0))
		return 0;
	if (!fetch_match(number, sizeof(number), line_buffer,
			 match_list + PATH_MATCH_LUN) ||
	    path->un.m_ScsiId.m_lun != strtol(number, NULL, 0))
		return 0;
	return 1;
}

static int
is_path_eq(struct ElxDiDevId  *path, char *line_buffer,
	     regmatch_t *match_list)
{
	if (path->m_type == ELX_DI_DEVID_SCSI_NEXUS)
		return is_scsi_path_eq(path, line_buffer, match_list);
	if (path->m_type == ELX_DI_DEVID_NEXUS_LONG)
		return is_long_path_eq(path, line_buffer, match_list);
	return 0;
}

static int
is_device_eq(struct ElxDiDevId  *device, char *line_buffer,
	     regmatch_t *match_list)
{
	char ascii[sizeof(device->un.m_InquiryData) * 3];
	unsigned char inquiry[sizeof(device->un.m_InquiryData)];

	if (device->m_type == ELX_DI_DEVID_INQSN) {
		fetch_match(ascii, sizeof(ascii), line_buffer,
			    match_list + DEVICE_MATCH_SERIAL);
		crack_inquiry_encoding(inquiry, ascii, sizeof(inquiry));
		return (device->m_len  == (strlen(ascii) + 1) / 3 &&
			!memcmp(device->un.m_InquiryData, inquiry,
				device->m_len));
	}
	return 0;
}

static int
failover_bit_mask(char *line_buffer, regmatch_t *match_list)
{
	char parameter[32];

	fetch_match(parameter, sizeof(parameter), line_buffer,
		    match_list + PATH_MATCH_STATE);
	if (strcmp(parameter, "online") == 0) {
		return 0;
	}
	if (strcmp(parameter, "standby") == 0) {
		return MPL_PATH_INACTIVE;
	}
	if (strcmp(parameter, "failed") == 0) {
		return MPL_PATH_LINK_DOWN;
	}
	if (strcmp(parameter, "disabled") == 0) {
		return MPL_PATH_MAINT_MODE;
	}
	return MPL_PATH_INVALID;
}

static void
make_path_list_details(struct MplPathListEntry *path, char *line_buffer,
		       regmatch_t *match_list)
{
	char parameter[32];

	path->FailoverBitMask = failover_bit_mask(line_buffer, match_list);
	fetch_match(parameter, sizeof(parameter), line_buffer,
		    match_list + PATH_MATCH_PRIMARY);
	path->Default = (strcmp(parameter, "primary") == 0);
	path->FcpTarget = 1;
	path->Pad = 0;
	fetch_match(parameter, sizeof(parameter), line_buffer,
		    match_list + PATH_MATCH_STATE);
	path->Found = (strcmp(parameter, "notfound") != 0);
	fetch_match(parameter, sizeof(parameter), line_buffer,
		    match_list + PATH_MATCH_RATIO);
	path->StaticConfig = (strcmp(parameter, "static") == 0);
}

static int
get_traffic_ratio(char *line_buffer, regmatch_t *match_list)
{
	return fetch_match_integer(line_buffer, match_list, PATH_MATCH_RATIO);
}

static int
device_treatment_type(char *line_buffer, regmatch_t *match_list)
{
	static char *names[] = {
		[MPL_PC_TYPE_FAILOVER] = "autofail",
		[MPL_PC_TYPE_TRAFFIC_RATIO] = "ratio",
		[MPL_PC_TYPE_DYNAMIC_LUNQ] = "lunq",
		[MPL_PC_TYPE_DYNAMIC_TGTQ] = "tgtq",
		[MPL_PC_TYPE_DYNAMIC_HBAQ] = "hbaq",
		[MPL_PC_TYPE_DYNAMIC_TIME] = "time",
	};
	char parameter[32];
	int  i;

	fetch_match(parameter, sizeof(parameter), line_buffer, 
		    match_list + DEVICE_MATCH_AUTOFAIL);
	for (i = 0; i < NUM_ENTRIES(names); ++i) {
		if (names[i] && !strcmp(parameter, names[i])) {
			return i;
		}
	}
	return MPL_PC_TYPE_FAILOVER;
}

static int
is_configured(char *line_buffer, regmatch_t *match_list)
{
	char parameter[32];

	fetch_match(parameter, sizeof(parameter), line_buffer, match_list);
	return !strcmp(parameter, "configured");
}

static int
is_device_configured(char *line_buffer, regmatch_t *match_list)
{
	return is_configured(line_buffer, match_list + DEVICE_MATCH_ORIGIN);
}

static int
is_vendor_empty(char *line_buffer, regmatch_t *match_list)
{
	char vendor[16];
	
	fetch_match(vendor, sizeof(vendor), line_buffer,
		    match_list + DEVICE_MATCH_VENDOR);
	return !vendor[0];
}

static int
is_device_disk(char *line_buffer, regmatch_t *match_list)
{
	static char *type_names[] = {
		"disk",
		"tape",
		"printer",
		"worm",
		"cd/dvd",
		"simple-disk",
	};
	char parameter[32];
	int  i;
	

	if (is_vendor_empty(line_buffer, match_list)) {
		return 0;
	}

	fetch_match(parameter, sizeof(parameter), line_buffer, 
		    match_list + DEVICE_SCSI_TYPE);
	for (i = 0; i < NUM_ENTRIES(type_names); ++i) {
		if (!strcmp(parameter, type_names[i])) {
			return 1;
		}
	}
	return 0;
}

static int
is_device_autofail(char *line_buffer, regmatch_t *match_list)
{
	char parameter[32];

	fetch_match(parameter, sizeof(parameter), line_buffer, 
		    match_list + DEVICE_MATCH_AUTOFAIL);
	return !strcmp(parameter, "autofail");
}

static int
is_path_configured(char *line_buffer, regmatch_t *match_list)
{
	return is_configured(line_buffer, match_list + PATH_MATCH_ORIGIN);
}

static int
is_path_unconfigured(char *line_buffer, regmatch_t *match_list)
{
	char parameter[32];

	fetch_match(parameter, sizeof(parameter), line_buffer, 
		    match_list + PATH_MATCH_ORIGIN);
	return !strcmp(parameter, "unconfigured");
}

static int
is_path_exported(char *line_buffer, regmatch_t *match_list)
{
	char parameter[32];

	fetch_match(parameter, sizeof(parameter), line_buffer, 
		    match_list + PATH_MATCH_ORIGIN);
	return !strcmp(parameter, "unavailable");
}

static int
is_path_primary(char *line_buffer, regmatch_t *match_list)
{
	char parameter[32];

	fetch_match(parameter, sizeof(parameter), line_buffer, 
		    match_list + PATH_MATCH_PRIMARY);
	return !strcmp(parameter, "primary");
}


static char *
format_path_id(char *buffer, size_t size, struct ElxDiDevId *path)
{
	if (path->m_type == ELX_DI_DEVID_SCSI_NEXUS) {
		snprintf(buffer, size, "%d/%d/%d",
			 path->un.m_ScsiId.m_adapter,
			 path->un.m_ScsiId.m_target,
			 path->un.m_ScsiId.m_lun);
	} else if (path->m_type == ELX_DI_DEVID_NEXUS_LONG) {
		snprintf(buffer, size,
			 "%02x%02x%02x%02x%02x%02x%02x%02x:"
			 "%02x%02x%02x%02x%02x%02x%02x%02x:"
			 "%d",
			 path->un.m_NexusLong.m_adapter_wwpn[0],
			 path->un.m_NexusLong.m_adapter_wwpn[1],
			 path->un.m_NexusLong.m_adapter_wwpn[2],
			 path->un.m_NexusLong.m_adapter_wwpn[3],
			 path->un.m_NexusLong.m_adapter_wwpn[4],
			 path->un.m_NexusLong.m_adapter_wwpn[5],
			 path->un.m_NexusLong.m_adapter_wwpn[6],
			 path->un.m_NexusLong.m_adapter_wwpn[7],
			 path->un.m_NexusLong.m_target_wwpn[0],
			 path->un.m_NexusLong.m_target_wwpn[1],
			 path->un.m_NexusLong.m_target_wwpn[2],
			 path->un.m_NexusLong.m_target_wwpn[3],
			 path->un.m_NexusLong.m_target_wwpn[4],
			 path->un.m_NexusLong.m_target_wwpn[5],
			 path->un.m_NexusLong.m_target_wwpn[6],
			 path->un.m_NexusLong.m_target_wwpn[7],
			 path->un.m_NexusLong.m_lun[0]);
	} else if (path->m_type == ELX_DI_DEVID_INQSN) {
		format_inqsn(buffer, size, path->un.m_InquiryData, path->m_len);
	} else {
		snprintf(buffer, size, "invalid device type %d", path->m_type);
	}
	
	return buffer;
}

/*
 * Utility functions for removing adding upper layer devices.  When we
 * add a path to a route, we want to tear down the associated SCIS
 * mid-layerd device; so the mid-layer device only sees the single,
 * presentation, path to the device.
 *
 * When we remove a path from a device, we want to make that path
 * visible to the mid-layer.
 */

static int numeric_compare(const void *d1, const void *d2){

	char *s1 = (*((struct dirent **) d1))->d_name;
	char *s2 = (*((struct dirent **) d2))->d_name;
	int  n1, n2;

				/* sort names in directory so numeric names
				 * appear first w/ "9" < "10".
				 */
	if (isdigit(*s1) && isdigit(*s2)) {
		n1 = strtol(s1, NULL, 0);
		n2 = strtol(s2, NULL, 0);
		return n1 - n2;
	}
	if (isdigit(*s1)) {
		return -1;
	}
	if (isdigit(*s2)) {
		return -1;
	}
	return strcmp(s1, s2);
}

static int ignore_dotted_files(const struct dirent *dent)
{
	return dent->d_name[0] != '.';
}


static int
get_first_adapter_number(void) 
{
	static int first_adapter = 0xffff;
	struct dirent **namelist;
	char fname[64];
	char * ptr;
	int  i, num;
	int  n;


	n  = scandir(SYS_SCSI_HOSTS, &namelist, ignore_dotted_files,
		     numeric_compare);
	if (n < 0) {
		first_adapter = 0;
		return first_adapter;
	}

	for (i = 0; i < n; ++i) {
				/* skip non lpfc scsi devices */
		snprintf(fname, sizeof(fname), "%s/%s/lpfc_drvr_version",
			 SYS_SCSI_HOSTS, namelist[i]->d_name);
		if (!access(fname, R_OK)) {
			ptr = namelist[i]->d_name + 4; /* Skip past "host" */
			num =  atoi(ptr);
			if (num < first_adapter)
				first_adapter = num;
			if (first_adapter < 0)
				first_adapter = 0;
		}
		free(namelist[i]);
	}
	free(namelist);

	if (first_adapter == 0xffff)
		first_adapter = 0;

	return first_adapter;
}

#if 0
static int
proc_scsi_host_match(char *input, regmatch_t *match_list,
		     int host, int target, int lun)
{
	return fetch_match_integer(input, match_list, PROC_SCSI_HOST) == host &&
		fetch_match_integer(input, match_list, PROC_SCSI_TARGET) == target &&
		fetch_match_integer(input, match_list, PROC_SCSI_LUN) == lun;

}

static int
is_scsi_type_da(char *line_buffer, size_t size, FILE *fp)
{
	regmatch_t match_list[MATCH_COUNT_MAX];
	int rc;

	while (fgets(line_buffer, size, fp)) {
		rc = regexec(&proc_scsi_host_re, line_buffer,
			     NUM_ENTRIES(match_list), match_list, 0);
		if (!rc)
			return 0;
		if (!strcmp(line_buffer, "  Type: "))
			return strstr(line_buffer, "Direct-Access") != NULL;
	}
	return 0;
}

static int
is_disk_device(int host, int target, int lun)
{
	FILE *fp;
	char line_buffer[256];
	regmatch_t match_list[MATCH_COUNT_MAX];
	int is_da = 0;
	int rc;

	fp = fopen("/proc/scsi/scsi", "r");
	if (fp) {
		while (fgets(line_buffer, sizeof(line_buffer), fp)) {
			rc = regexec(&proc_scsi_host_re, line_buffer,
				     NUM_ENTRIES(match_list), match_list, 0);

			if (!rc) {
				if (proc_scsi_host_match(line_buffer,
							 match_list,
							 host, target, lun)) {
					is_da = is_scsi_type_da(line_buffer,
								sizeof(line_buffer),
								fp);
					break;
				}
			}
		}
		fclose(fp);
	}
	return is_da;
}
#endif

static void
remove_midlayer_device(int host, int device, int lun)
{
	FILE *fp;

	fp = fopen("/proc/scsi/scsi", "w");

	if (fp) {
		fprintf(fp, "scsi remove-single-device %d 0 %d %d\n",
			host + get_first_adapter_number(), device, lun);
		fclose(fp);
	}
}

static void
add_midlayer_device(int host, int device, int lun)
{
	FILE *fp;

	remove_midlayer_device(host, device, lun);
	sleep(2);
				/* We may at some point, want to only add
				 * devices that we've removed.  But for the
				 * moment, unconditionally adding doesn't seem
				 * to do any harm. */
	fp = fopen("/proc/scsi/scsi", "w");

	if (fp) {
		fprintf(fp, "scsi add-single-device %d 0 %d %d\n",
			host + get_first_adapter_number(), device, lun);
		fclose(fp);
	}
}

static int
find_path_htl(struct MplPathListEntry *path, int *host, int *target, int *lun)
{
	struct ElxDiDevId tmp;
	FILE *fp;
	char line_buffer[256];
	regmatch_t match_list[MATCH_COUNT_MAX];
	int  this_device = 0;
	int  is_disk;
	int  rc;

	if (path->DevId.m_type == ELX_DI_DEVID_SCSI_NEXUS) {
		*host = path->DevId.un.m_ScsiId.m_adapter;
		*target = path->DevId.un.m_ScsiId.m_target;
		*lun = path->DevId.un.m_ScsiId.m_lun;
		return 0;
	}

	if (path->DevId.m_type != ELX_DI_DEVID_NEXUS_LONG) {
		goto error_out;
	}
	
	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		goto error_out;
	}

	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		get_current_host(line_buffer);
		rc = regexec(&device_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc) {
			if (this_device)
				goto close_error_out;
			is_disk = is_device_disk(line_buffer, match_list);
			this_device |= is_device_eq(&path->ValidationId,
						    line_buffer, match_list);
		} else if (this_device && is_disk && rc == REG_NOMATCH) {
			rc = regexec(&path_re, line_buffer,
				     NUM_ENTRIES(match_list), match_list, 0);
			if (!rc && is_path_eq(&path->DevId, line_buffer,
					      match_list)) {
				make_scsi_path_entry(&tmp, line_buffer,
						     match_list);
				*host = tmp.un.m_ScsiId.m_adapter;
				*target = tmp.un.m_ScsiId.m_target;
				*lun = tmp.un.m_ScsiId.m_lun;
				break;
			}
		} 
		if (rc && rc != REG_NOMATCH) {
			goto close_error_out;
		}
	}
	
	fclose(fp);
	return 0;

close_error_out:
	fclose(fp);
error_out:
	return -1;
}

/***************************************************************************
 *
 * mplInitialize
 *
 * Initialize the MultiPulse API Library. 
 * Return the version of the MultiPulse library.
 *
 * Parameters:
 *
 * Version        - points to a buffer to receive the MultiPulse version
 *
 * Length         - contains the number of bytes available in the buffer
 *                  pointed to by Version 
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the initialization is successful. Other status
 * indicative of the failure if the initialization is not successful.
 *
 * Comments:
 *
 * This function must be called to initialize the MultiPulse library before
 * any other MultiPulse library calls are made.
 *
 ***************************************************************************/
static void
determine_version(char *buffer)
{
	regmatch_t match_list[2];
	char  version_string[32];
	int   rc;
#define VERSION_MATCH_VERSION 1

	rc = mpl_re_compile(&version_re, "[Vv]ersion[ :=]*([[:digit:]]+)");
	if (!rc) {
		rc = regexec(&version_re, buffer, NUM_ENTRIES(match_list),
			     match_list, 0);
	}
	if (!rc) {
		rc = !fetch_match(version_string, sizeof(version_string),
				  buffer, match_list + VERSION_MATCH_VERSION);
	}
}

MPL_STATUS mplInitialize(char *Version, 
                         unsigned int Length)
{
	FILE *fp;
	char buffer[64];
	int  rv = 0;

	fp = fopen(MPL_VERSION_NODE, "r");
	if (!fp)
		return MPL_STATUS_NO_DEVICE;
	fread(buffer, 1, sizeof(buffer), fp);
	if (ferror(fp))
		rv = MPL_STATUS_NO_DEVICE;
	fclose(fp);
	if (rv)
		return rv;
	determine_version(buffer);
	if (init_regexp())
		return MPL_STATUS_UNSUCCESSFUL;
	if (Length > 0) {
		snprintf(Version, Length, "%s", mpl_library_version);
		Version[Length - 1] = 0;
	}
	build_port_cache();
	initialized = 1;
	return MPL_STATUS_SUCCESS;
}

/***************************************************************************
 *
 * mplGetPortList
 *
 * Retreive a list of identifiers to be used to identify the different ports
 * available to the MultiPulse library.
 *
 * Parameters:
 *
 * PortList      - Points a GET_PORT_LIST structure for input and return values
 *                 If PortList->Count == 0, the current count of available
 *                 GET_PORT_LIST identifiers is returned.
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if list of port identifiers in GET_PORT_LIST is valid. 
 * MPL_STATUS_UNSUCCESSFUL completed unsuccessfully for an unspecified reason
 * MPL_STATUS_BUFFER_TOO_SMALL if the list to receive the identifiers is not
 *     large enough to contain all of the port identifiers available.
 *     Only PortList->Count identifiers are returned.
 *
 * Comments:
 *
 * Call this function to get the available port identifiers to be used in
 * subsequent MultiPulse API library calls.
 *
 * The Count member of GET_PORT_LIST structure must be initialized to the number
 * of elements in the Ports member array, or 0, before calling mplGetPortList.
 *
 ***************************************************************************/
static void
make_port_device(struct ElxDiDevId  *device, unsigned char wwpn[8])
{
	memset(device, 0, sizeof(*device));
	device->m_type = ELX_DI_DEVID_WWPN;
	device->m_len  = SIZEOF_WWPN;
	device->m_dflags = ELX_DI_DEVID_FLAGS_BINARY;
	memcpy(device->un.m_wwn, wwpn, sizeof(device->un.m_wwn));
}

static int decode_port_state(const char *state)
{
	static struct {
		char *text;
		int  value;
	} state_xlation[] = {
		{ "Down", MPL_PORT_LINK_DOWN },
		{ "Up - Discovery", MPL_PORT_LINK_UP },
		{ "Up - Ready", MPL_PORT_DISCOVERY_COMPLETE },
		{ "Up", MPL_PORT_LINK_UP },
	};
	int  i;
		       

	for (i = 0; i < NUM_ENTRIES(state_xlation); ++i) {
		if (!strcmp(state, state_xlation[i].text)) {
			return state_xlation[i].value;
		}
	}
	return MPL_PORT_OTHER;
}

static MPL_STATUS
build_port_entry(struct MplPortListEntry *port, const char *name)
{
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	FILE *fp;
	char line_buffer[256];
	regmatch_t match_list[5];
	char port_name[32];
	char port_state[32];
	unsigned char wwn[8];
	int  rc;

	fp = fopen(name, "r");
	if (!fp) {
		status = MPL_STATUS_NO_DEVICE;
		goto out;
	}

	status = MPL_STATUS_SUCCESS;
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		rc = regexec(&port_name_re, line_buffer,
			     NUM_ENTRIES(match_list), match_list, 0);
		if (!rc && fetch_match(port_name, sizeof(port_name),
				       line_buffer,
				       match_list + PORT_NAME_MATCH_WWN)) {
			parse_wwn(wwn, port_name);
			make_port_device(&port->DevId, wwn);
			continue;
		} else if (rc != REG_NOMATCH) {
			char buffer[64];

			regerror(rc, &port_name_re, buffer, sizeof(buffer));
			status = MPL_STATUS_UNSUCCESSFUL;
			break;
		}
		rc = regexec(&port_state_re, line_buffer,
			     NUM_ENTRIES(match_list), match_list, 0);
		if (!rc && fetch_match(port_state, sizeof(port_state),
				line_buffer,
				match_list + PORT_STATE_MATCH_STATE)) {
			port->State = decode_port_state(port_state);
		} else if (rc != REG_NOMATCH) {
			char buffer[64];

			regerror(rc, &port_state_re, buffer, sizeof(buffer));
			fprintf(stderr, "mpllib: port state, no match: %s\n",
				buffer);
			status = MPL_STATUS_UNSUCCESSFUL;
			break;
		}
	}
	
	fclose(fp);
out:
	return status;
}


MPL_STATUS port_list_2_6(struct MplGetPortList *PortList)
{
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	struct MplGetPortList *port_list = PortList;
	int  count = port_list->Count;
	struct dirent **namelist;
	char fname[64];
	int  i, p = 0;
	int  n;

	if (!initialized) {
		status = MPL_STATUS_UNSUCCESSFUL;
		goto out;
	}

	n  = scandir(SYS_SCSI_HOSTS, &namelist, ignore_dotted_files,
		     numeric_compare);
	if (n < 0) {
		perror("mplGetPortList scandir");
		status = MPL_STATUS_NO_DEVICE;
		goto out;
	}
	status = MPL_STATUS_SUCCESS;
	for (i = 0; i < n && p < port_count; ++i) {
				/* skip non lpfc scsi devices */
		snprintf(fname, sizeof(fname), "%s/%s/lpfc_drvr_version",
			 SYS_SCSI_HOSTS, namelist[i]->d_name);
		if (!access(fname, R_OK)) {
			snprintf(fname, sizeof(fname), "%s/%s/state",
				 SYS_SCSI_HOSTS, namelist[i]->d_name);
			if (p >= count) {
				status = MPL_STATUS_BUFFER_TOO_SMALL;
			} else {
				int rc;

				rc = build_port_entry(port_list->Ports + p,
						      fname);
				if (rc != MPL_STATUS_SUCCESS) {
					status = rc;
				}
				make_port_device(&port_list->Ports[p].DevId,
						 port_cache[p].port_wwpn);
			}
			++p;
		}
		free(namelist[i]);
	}
	free(namelist);

out:
 	port_list->Count = p;
	return status;
}

MPL_STATUS mplGetPortList(PMPL_GET_PORT_LIST PortList)
{
	return (port_list_2_6(PortList));
}

/***************************************************************************
 *
 * mplGetRouteList
 *
 * Retreive a list of identifiers to be used to identify the different routes
 * available to the MultiPulse library.
 *
 * Parameters:
 *
 * RouteList    - Points a GET_ROUTE_LIST structure for input and return values
 *                If RouteList->Count == 0, the current count of available
 *                GET_ROUTE_LIST identifiers is returned.
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if list of route identifiers in GET_ROUTE_LIST is valid. 
 * MPL_STATUS_UNSUCCESSFUL completed unsuccessfully for an unspecified reason
 * MPL_STATUS_BUFFER_TOO_SMALL if the list to received the identifiers is not
 *    large enough to contain all of the route identifiers available
 *    Only RouteList->Count identifiers are returned.
 * -1 if the config file cannot be opened, so as to distinguish between
 *    success and a failure.
 *
 * Comments:
 *
 * Call this function to get the available route identifiers to be used in
 * subsequent MultiPulse API library calls.
 *
 * The Count member of GET_ROUTE_LIST structure must be initialized to the number
 * of elements in the Ports member array, or 0, before calling mplGetRouteList.
 *
 ***************************************************************************/

MPL_STATUS mplGetRouteList(PMPL_GET_ROUTE_LIST RouteList)
{
	struct MplGetRouteList *route_list = RouteList;
	int  count = route_list->Count;
	int  i = 0;
	struct ElxDiDevId device;
	struct ElxDiDevId route;
	FILE *fp;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	char line_buffer[256];
	char sbuf[32];
	regmatch_t match_list[MATCH_COUNT_MAX];
	int  configured = 0;
	int  rc;

	if (!initialized) {
		status = MPL_STATUS_UNSUCCESSFUL;
		goto out;
	}
	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		status = -1;
		goto out;
	}
	
	status = MPL_STATUS_SUCCESS;
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		get_current_host(line_buffer);
		rc = regexec(&device_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc) {
			configured = is_device_configured(line_buffer,
							  match_list) &&
				is_device_disk(line_buffer, match_list);
			make_scsi_id(&device, line_buffer, match_list);
			make_route_id(&route, line_buffer, match_list);
			if (configured) {
				if (i >= count) {
					status = MPL_STATUS_BUFFER_TOO_SMALL;
				} else {
					route_list->Routes[i].ValidationId = device;
					route_list->Routes[i].RouteId = route;
					fetch_match(sbuf, sizeof(sbuf), line_buffer,
		    				match_list + DEVICE_MATCH_STATIC);
					route_list->Routes[i].StaticConfigPrimary = 0; 
					if (!strcmp(sbuf, "static_p")) {
						route_list->Routes[i].StaticConfig = 1; 
						route_list->Routes[i].StaticConfigPrimary = 1; 
					}
					else {
						route_list->Routes[i].StaticConfig =
							(strcmp(sbuf, "static") == 0);
					}
					fetch_match(sbuf, sizeof(sbuf), line_buffer,
		    				match_list + DEVICE_MATCH_STATE);
					route_list->Routes[i].Found = (strcmp(sbuf, "notfound") != 0);
					route_list->Routes[i].AutoFailBack =
						is_device_autofail(line_buffer, match_list);
				}
				++i;
			}
		} else if (configured && rc == REG_NOMATCH) {
			rc = regexec(&path_re, line_buffer,
				     NUM_ENTRIES(match_list), match_list, 0);
			if (!rc) {
				if (!is_path_configured(line_buffer,
							match_list)) {
					continue;
				}
			}
		} 
		if (rc && rc != REG_NOMATCH) {
			char buffer[64];

			regerror(rc, &device_re, buffer, sizeof(buffer));
			fprintf(stderr, "mpllib: route list, no match: %s\n",
				buffer);
			status = MPL_STATUS_UNSUCCESSFUL;
			break;
		}
	}
	
	fclose(fp);
out:
 	route_list->Count = i;
	return status;
}



/***************************************************************************
 *
 * mplGetPathList
 *
 * Retreive a list of identifiers to be used to identify the different paths
 * available to the MultiPulse library for the specified route. If no route is
 * specified, return a list of the unbound paths.
 *
 * Parameters:
 *
 * PathList     - Points a GET_PATH_LIST structure for input and return values
 *                If PathList->Count == 0, the current count of available
 *                GET_PATH_LIST identifiers is returned.
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if list of path identifiers in GET_PATH_LIST is valid. 
 * MPL_STATUS_UNSUCCESSFUL completed unsuccessfully for an unspecified reason
 * MPL_STATUS_BUFFER_TOO_SMALL if the list to received the identifiers is not
 *    large enough to contain all of the path identifiers available
 *    Only PathList->Count identifiers are returned.
 *
 * Comments:
 *
 * Call this function to get the available path identifiers to be used in
 * subsequent MultiPulse API library calls.
 *
 * The Count member of GET_PATH_LIST structure must be initialized to the number
 * of elements in the Ports member array, or 0, before calling mplGetPathList.
 *
 * The Route member of the GET_PATH_LIST structure should contain a valid route
 * identifier returned from a previous call to mplGetRouteList. If the list of
 * free paths is desired, the Route member should be initialized to the value
 * of MplNullIdentifier.
 *
 ***************************************************************************/

static MPL_STATUS
make_path_list_entry(struct MplGetPathList *path_list, int index,
		     struct ElxDiDevId *device,
		     char *line_buffer, regmatch_t *match_list)
{
	struct ElxDiDevId path;

	make_path_entry(&path, line_buffer, match_list);
	if (index >= path_list->Count) {
		if (index && is_path_primary(line_buffer, match_list)) {
			path_list->Paths[0].ValidationId = *device;
			path_list->Paths[0].DevId = path;
			make_path_list_details(path_list->Paths, line_buffer,
					       match_list);
		}
		return MPL_STATUS_BUFFER_TOO_SMALL;
	} else {
		path_list->Paths[index].ValidationId = *device;
		path_list->Paths[index].DevId = path;
		make_path_list_details(path_list->Paths + index,
				       line_buffer, match_list);
				/* Swap path entry with first path if
				 * this non first entry is primary */
		if (index && path_list->Paths[index].Default) {
			struct MplPathListEntry tmp;

			tmp = path_list->Paths[index];
			path_list->Paths[index] = path_list->Paths[0];
			path_list->Paths[0] = tmp;
		}
	}
	return MPL_STATUS_SUCCESS;
}



static MPL_STATUS
mpl_get_specific_path_list(struct MplGetPathList *path_list)
{
	struct ElxDiDevId device;
	FILE *fp;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	char line_buffer[256];
	regmatch_t match_list[MATCH_COUNT_MAX];
	int  this_device = 0;
	int  is_disk;
	int  i = 0;
	int  configured = 0;
	int  rc;

	if (path_list->RouteId.m_type != ELX_DI_DEVID_SCSI_NEXUS &&
	    path_list->RouteId.m_type != ELX_DI_DEVID_NEXUS_LONG &&
	    path_list->RouteId.m_type != ELX_DI_DEVID_INQSN) {
		status = MPL_STATUS_INVALID_PARAMETER;
		goto out;
	}
	
	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		status = MPL_STATUS_NO_DEVICE;
		goto out;
	}

	status = MPL_STATUS_SUCCESS;
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		get_current_host(line_buffer);
		rc = regexec(&device_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc) {
			if (this_device)
				break;
			i = 0;
			make_scsi_id(&device, line_buffer, match_list);
			is_disk = is_device_disk(line_buffer, match_list);
			this_device |= is_device_eq(&path_list->RouteId,
						    line_buffer, match_list);
		} else if (is_disk && rc == REG_NOMATCH) {
			rc = regexec(&path_re, line_buffer,
				     NUM_ENTRIES(match_list), match_list, 0);
			if (!rc) {
				this_device |= is_path_eq(&path_list->RouteId,
							  line_buffer,
							  match_list);
				configured = is_path_configured(line_buffer,
								match_list);
				if (!configured) {
					continue;
				}
				status = make_path_list_entry(path_list, i,
						      &device,
						      line_buffer,
						      match_list);
				path_list->Paths[i].Bound = 1;
				++i;
			}
		} 
		if (rc && rc != REG_NOMATCH) {
			char buffer[64];

			regerror(rc, &device_re, buffer, sizeof(buffer));
			status = MPL_STATUS_UNSUCCESSFUL;
			break;
		}
	}
	
	fclose(fp);
out:
 	path_list->Count = i;
	if (!this_device) {
		status = MPL_STATUS_ROUTE_NOT_FOUND;
	}
	return status;
}

static MPL_STATUS
mpl_get_unbound_path_list(struct MplGetPathList *path_list,
			  match_list_test filter)
{
	int  count = path_list->Count, i = 0;
	struct ElxDiDevId device;
	struct ElxDiDevId path;
	FILE *fp;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	char line_buffer[256];
	char parameter[32];
	regmatch_t match_list[MATCH_COUNT_MAX];
	int  is_disk;
	int  rc;

	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		status = MPL_STATUS_NO_DEVICE;
		goto out;
	}
	
	status = MPL_STATUS_SUCCESS;
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		get_current_host(line_buffer);
		rc = regexec(&device_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc) {
			make_scsi_id(&device, line_buffer, match_list);
			is_disk = is_device_disk(line_buffer, match_list);
		} else if (is_disk && rc == REG_NOMATCH) {
			rc = regexec(&path_re, line_buffer,
				     NUM_ENTRIES(match_list), match_list, 0);
			if (!rc) {
				if (!filter(line_buffer, match_list)) {
					continue;
				}
				make_path_entry(&path, line_buffer, 
						match_list);
				if (i >= count) {
					status = MPL_STATUS_BUFFER_TOO_SMALL;
				} else  {
					path_list->Paths[i].DevId = path;
					path_list->Paths[i].FailoverBitMask = 0;
					path_list->Paths[i].Bound = 0;
					fetch_match(parameter, sizeof(parameter), line_buffer,
		    				match_list + PATH_MATCH_STATE);
					path_list->Paths[i].Found =
							(strcmp(parameter, "notfound") != 0);
					path_list->Paths[i].StaticConfig = 1;
					if (path_list->Paths[i].Found)
						path_list->Paths[i].ValidationId = device;
				}
				++i;
			} 
			if (rc && rc != REG_NOMATCH) {
				char buffer[64];
				
				regerror(rc, &device_re, buffer, sizeof(buffer));
				fprintf(stderr, "mpllib: path list, no match: %s\n",
					buffer);
				status = MPL_STATUS_UNSUCCESSFUL;
				break;
			}
		}
	}
	
	fclose(fp);
out:
 	path_list->Count = i;
	return status;
}


MPL_STATUS mplGetPathList(PMPL_GET_PATH_LIST PathList)
{
	if (!initialized) {
		PathList->Count = 0;
		return MPL_STATUS_UNSUCCESSFUL;
	}

	return (memcmp(&mplNullIdentifier, &PathList->RouteId,
		       sizeof(mplNullIdentifier))
		? mpl_get_specific_path_list(PathList)
		: mpl_get_unbound_path_list(PathList, is_path_unconfigured));
}

/***************************************************************************
 *
 * mplCreateRoute
 *
 * Create a new route from a list of paths / define the behavior of the route.
 *
 * Parameters:
 *
 * PathList          - Points a GET_PATH_LIST structure indicating the
 *                     available paths for the route
 *
 * PathControl       - Points to a path control structure used to control how
 *                     the paths comprising the route will be controlled
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the route is created successfully.
 * Other error codes will indicate the reason for failure.
 *
 * Comments:
 *
 * The path list structure is created to describe the paths to the route.
 * The elements of the path list structure member Paths are obtained from
 * a call to mplGetPaths. The MPL_GET_PATH_LIST RouteId field is used to
 * specify the identifier used to identify the route. This must be unique on
 * the system it is defined. Creating two routes with the same route identifier
 * will lead to an unstable system. The MPL_GET_PATH_LIST Count field is used
 * the specify the number of MPL_PATH_LIST_ENTRYs that make up the route.
 * 
 * A path may be a member of one an only one route. If an attempt is made to
 * add a path to a route and the path is already in use, an error will occur.
 *
 * The PathControl argument should point to one of the various path control
 * structures. The type of multipath control will be obtained from the type
 * field contained in the header of the chosed path control structure.
 *
 ***************************************************************************/

static MPL_STATUS
set_device_failover(struct MplFailoverControl *failover)
{
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	FILE *fp;
	char path_id[PATH_ID_SIZE];
	char tmp[32];
	char serial_number[256];

	if (failover->PrimaryPathValidationId.m_type != ELX_DI_DEVID_INQSN) {
		return MPL_STATUS_INVALID_PARAMETER;
	}

	fp = fopen(MPL_CONTROL_NODE, "r+");
	if (!fp) {
		return MPL_STATUS_NO_DEVICE;
	}

	status = MPL_STATUS_SUCCESS;

	format_inqsn(serial_number, sizeof(serial_number),
		     failover->PrimaryPathValidationId.un.m_InquiryData,
		     failover->PrimaryPathValidationId.m_len);

	fprintf(fp, "device %s type-failover\n", serial_number);

	fprintf(fp, "device %s %s\n",
		serial_number,
		failover->Header.AutoFailBack ? "failback" : "nofailback");

	/* Give the path a chance to figure out its SAS, if applicable */
	sleep(1);

	if (failover->PrimaryPathDevId.m_type) {
		fprintf(fp, "device %s primary %s\n",
			serial_number,
			format_path_id(path_id, sizeof(path_id),
			       &failover->PrimaryPathDevId));
	}

	fflush(fp);
	rewind(fp);
	if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
		fprintf(stderr, "%s: no status\n", __FUNCTION__);
	} else {
		status = strtol(tmp, NULL, 0);
	}
	fclose(fp);
	return status;
}

static MPL_STATUS
set_device_traffic_ratios(struct MplTrafficRatioControl *control)
{
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	FILE *fp;
	char path_id[PATH_ID_SIZE];
	char tmp[32];
	int  i, count = control->Count;

	fp = fopen(MPL_CONTROL_NODE, "r+");
	if (!fp) {
		return MPL_STATUS_NO_DEVICE;
	}

	status = MPL_STATUS_SUCCESS;

	for (i = 0; status == MPL_STATUS_SUCCESS && i < count; ++i) {
		format_path_id(path_id, sizeof(path_id),
			       &control->PathRatios[i].DeviceId);
		fprintf(fp, "path %s ratio %d\n",
			path_id, control->PathRatios[i].Ratio);
		fflush(fp);
		rewind(fp);
		if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
			fprintf(stderr, "%s: no status\n", __FUNCTION__);
		} else {
			status = strtol(tmp, NULL, 0);
		}
	}
	fclose(fp);
	return status;		
}

static MPL_STATUS
set_device_queue_handler(struct ElxDiDevId *route, unsigned int type)
{
	static char *names[] = {
		[MPL_PC_TYPE_TRAFFIC_RATIO] = "ratio",
		[MPL_PC_TYPE_DYNAMIC_LUNQ] = "lunq",
		[MPL_PC_TYPE_DYNAMIC_TGTQ] = "targetq",
		[MPL_PC_TYPE_DYNAMIC_HBAQ] = "hbaq",
		[MPL_PC_TYPE_DYNAMIC_TIME] = "time",
	};
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	FILE *fp;
	char serial_number[256];
	char tmp[32];

	fp = fopen(MPL_CONTROL_NODE, "r+");
	if (!fp) {
		return MPL_STATUS_NO_DEVICE;
	}

	status = MPL_STATUS_SUCCESS;

	format_path_id(serial_number, sizeof(serial_number),
		       route);
	fprintf(fp, "device %s type-%s\n", serial_number, names[type]);
	fflush(fp);
	rewind(fp);
	if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
		fprintf(stderr, "%s: no status\n", __FUNCTION__);
	} else {
		status = strtol(tmp, NULL, 0);
	}
	fclose(fp);
	return status;
}

MPL_STATUS mplCreateRoute(PMPL_GET_PATH_LIST PathList,
                          PMPL_PATH_CONTROL_HEADER PathControl)
{
	FILE *fp;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	struct MplPathListEntry *path;
	struct ElxDiDevId *primary = NULL;
	char serial_number[256];
	char path_id[PATH_ID_SIZE];
	char tmp[32];
	int  count = PathList->Count;
	int  i;

	if (!initialized || count < 1) {
		return MPL_STATUS_UNSUCCESSFUL;
	}

	fp = fopen(MPL_CONTROL_NODE, "r+");
	if (!fp) {
		return MPL_STATUS_NO_DEVICE;
	}

	status = MPL_STATUS_SUCCESS;
	path = PathList->Paths;

	format_inqsn(serial_number, sizeof(serial_number),
		     path->ValidationId.un.m_InquiryData,
		     path->ValidationId.m_len);

	fprintf(fp, "path %s configure-device\n",
		format_path_id(path_id, sizeof(path_id), &path->DevId));
	fflush(fp);
	rewind(fp);
	if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
		fprintf(stderr, "%s: no status\n", __FUNCTION__);
	} else {
		status = strtol(tmp, NULL, 0);
	}
	if (status) {
		goto create_failed;
	}

	rewind(fp);
	for (i = 0; i < count; ++i) {
		fprintf(fp, "path %s configure\n",
			format_path_id(path_id, sizeof(path_id),
				       &path->DevId));
		fprintf(fp, "path %s status online\n",
			format_path_id(path_id, sizeof(path_id),
				       &path->DevId));
		++path;
		fflush(fp);
		rewind(fp);
		if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
			fprintf(stderr, "%s: no status\n", __FUNCTION__);
		} else {
			status = strtol(tmp, NULL, 0);
		}
		if (status) {
			goto cleanup_paths;
		}
	}

	rewind(fp);
	primary = &PathList->Paths->DevId;

	if (PathControl->Type == MPL_PC_TYPE_FAILOVER) {
		status = set_device_failover((struct MplFailoverControl *) 
					     PathControl);
	} else if (PathControl->Type == MPL_PC_TYPE_TRAFFIC_RATIO) {
		set_device_traffic_ratios((struct MplTrafficRatioControl *)
					  PathControl);
		status = set_device_queue_handler(&PathList->RouteId,
						  PathControl->Type);
	} else if (PathControl->Type == MPL_PC_TYPE_DYNAMIC_LUNQ ||
		   PathControl->Type == MPL_PC_TYPE_DYNAMIC_TGTQ ||
		   PathControl->Type == MPL_PC_TYPE_DYNAMIC_HBAQ ||
		   PathControl->Type == MPL_PC_TYPE_DYNAMIC_TIME) {
		status = set_device_queue_handler(&PathList->RouteId,
						  PathControl->Type);
	} else {
		status = MPL_STATUS_UNIMPLEMENTED;
	}

	if (status) {
		if (status == -EBUSY) {
			status = MPL_STATUS_ROUTE_NOT_FOUND;
cleanup_paths:
			rewind(fp);
			path = PathList->Paths;
			fprintf(fp, "path %s deconfigure-device\n",
				format_path_id(path_id, sizeof(path_id),
					       &path->DevId));
			fflush(fp);
			goto create_failed;
		}
	}

create_failed:
	fclose(fp);

	return status;
}



/***************************************************************************
 *
 * mplDeleteRoute
 *
 * Delete a previously created route 
 *
 * Parameters:
 *
 * RouteIdentifier   - Device identifier to be used to identify the route
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the route is created successfully.
 * Other error codes will indicate the reason for failure.
 *
 * Comments:
 *
 * The route identifier used to identify the route must be unique on the
 * system it is defined. 
 *
 ***************************************************************************/

typedef elxDiDeviceIdentifier fcDeviceIdentifier;

static void
add_single_path(char *line_buffer, regmatch_t *match_list)
{
	char number[16];
	int  host;
	int  target;
	int  lun;

	if (fetch_match(number, sizeof(number), line_buffer,
			match_list + PATH_MATCH_HOST)) {
		host = strtol(number, NULL, 0);
	}
	if (fetch_match(number, sizeof(number), line_buffer,
			match_list + PATH_MATCH_TARGET)) {
		target = strtol(number, NULL, 0);
	}
	if (fetch_match(number, sizeof(number), line_buffer,
			match_list + PATH_MATCH_LUN)) {
		lun = strtol(number, NULL, 0);
	}
	add_midlayer_device(host, target, lun);
}


static void
add_midlayer_paths(struct ElxDiDevId  *routeid)
{
	FILE *fp;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	char line_buffer[256];
	regmatch_t match_list[MATCH_COUNT_MAX];
	int  this_device = 0;
	int  is_disk;
	int  rc;

	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		return;
	}
	
	status = MPL_STATUS_SUCCESS;
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		get_current_host(line_buffer);
		rc = regexec(&device_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc) {
			if (this_device)
				break;
			this_device |= is_device_eq(routeid, line_buffer,
						    match_list);
			is_disk = is_device_disk(line_buffer, match_list);
		} else if (rc == REG_NOMATCH) {
			rc = regexec(&path_re, line_buffer,
				     NUM_ENTRIES(match_list), match_list, 0);
			if (!rc) {
				this_device |= is_path_eq(routeid,
							  line_buffer,
							  match_list);
				if (this_device && is_disk) {
					add_single_path(line_buffer, match_list);
				}
			}
		} 
		if (rc && rc != REG_NOMATCH) {
			char buffer[64];

			regerror(rc, &device_re, buffer, sizeof(buffer));
			status = MPL_STATUS_UNSUCCESSFUL;
			break;
		}
	}
	
	fclose(fp);
}

MPL_STATUS mplDeleteRoute(fcDeviceIdentifier *RouteIdentifier) 
{
	FILE *fp;
	char path_id[PATH_ID_SIZE];
	char tmp[32];
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;

	if (!initialized) {
		return MPL_STATUS_UNSUCCESSFUL;
	}
	if (RouteIdentifier->m_type != ELX_DI_DEVID_SCSI_NEXUS &&
	    RouteIdentifier->m_type != ELX_DI_DEVID_NEXUS_LONG &&
	    RouteIdentifier->m_type != ELX_DI_DEVID_INQSN) {
		return MPL_STATUS_INVALID_PARAMETER;
	}

	fp = fopen(MPL_CONTROL_NODE, "r+");
	if (!fp) {
		return MPL_STATUS_NO_DEVICE;
	}

	status = MPL_STATUS_SUCCESS;
	fprintf(fp, "path %s deconfigure-device\n",
		format_path_id(path_id, sizeof(path_id), RouteIdentifier));
	
	fflush(fp);
	rewind(fp);
	if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
		fprintf(stderr, "%s: no status\n", __FUNCTION__);
	} else {
		status = strtol(tmp, NULL, 0);
		if (status == -EBUSY) {
			status = MPL_STATUS_PATH_IN_USE;
		} else if (status == -ENODEV) {
			status = MPL_STATUS_ROUTE_NOT_FOUND;
		} else if (!status) {
			add_midlayer_paths(RouteIdentifier);
		}
	}

	fclose(fp);

	return status;
}


/***************************************************************************
 *
 * mplGetPathControl
 *
 * Retrieve the Path Controls for a specified route identifier.
 *
 * Parameters:
 *
 * RouteIdentifier  - A valid route identifier used to get the path controls
 *
 * PathControl      - A ptr to the buffer to contain the path controls struture
 *
 * BufferSize       - The size of the buffer pointed to by PathControl
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the PathControl structure is valid 
 * An error indicating the failure if the request is unsuccessful.
 *
 * Comments:
 *
 * The client application allocates a buffer to receive the path control
 * structure and initializes BufferSize with the size of the buffer.
 *
 * RouteId should be set to the identifier for the route from which the path
 * control structure is to be returned.
 *
 ***************************************************************************/

static void
do_traffic_ratio(struct MplTrafficRatioControl *traffic,
		 int path_index,
		 struct ElxDiDevId *device,
		 char *line_buffer,
		 regmatch_t *match_list)
{
	traffic->PathRatios[path_index].Ratio = get_traffic_ratio(line_buffer,
								  match_list);
	traffic->PathRatios[path_index].ValidationId = *device;
	make_path_entry(&traffic->PathRatios[path_index].DeviceId,
			line_buffer, match_list);
}

MPL_STATUS mplGetPathControl(fcDeviceIdentifier *Route,
                             PMPL_PATH_CONTROL_HEADER PathControl,
                             int BufferSize)
{
	FILE *fp;
	struct MplPathControlHeader *header;
	struct MplFailoverControl *failover;
	struct MplTrafficRatioControl *traffic;
	struct ElxDiDevId device;
	struct ElxDiDevId path;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	char line_buffer[256];
	char sbuf[32];
	regmatch_t match_list[MATCH_COUNT_MAX];
	int  this_device = 0;
	int  path_count = 0;
	int  first_path = 0;
	int  rc;

	if (!initialized) {
		return MPL_STATUS_UNSUCCESSFUL;
	}

	memset(PathControl, 0, BufferSize);
	header = (struct MplPathControlHeader *) PathControl;
	failover = (struct MplFailoverControl *) PathControl;
	traffic = (struct MplTrafficRatioControl *) PathControl;
	if (BufferSize < sizeof(*failover)) {
		return MPL_STATUS_BUFFER_TOO_SMALL;
	}

	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		status = MPL_STATUS_NO_DEVICE;
		goto out;
	}
	
	status = MPL_STATUS_SUCCESS;
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		get_current_host(line_buffer);
		rc = regexec(&device_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc) {
			first_path = 0;
			if (this_device) {
				break;
			}
			this_device = is_scsi_id_eq(Route, line_buffer,
						    match_list);
			header->Type = device_treatment_type(line_buffer,
							     match_list);
			make_scsi_id(&device, line_buffer, match_list);
			if (PathControl->Type == MPL_PC_TYPE_FAILOVER) {
				failover->Header.AutoFailBack = 
					is_device_autofail(line_buffer,
							   match_list);
			}

			fetch_match(sbuf, sizeof(sbuf), line_buffer,
		    		match_list + DEVICE_MATCH_STATIC);
			header->StaticConfigPrimary = 0; 
			if (!strcmp(sbuf, "static_p")) {
				header->StaticConfig = 1; 
				header->StaticConfigPrimary = 1; 
			}
			else {
				header->StaticConfig = (strcmp(sbuf, "static") == 0);
			}
			fetch_match(sbuf, sizeof(sbuf), line_buffer,
		    		match_list + DEVICE_MATCH_STATE);
			header->Found = (strcmp(sbuf, "notfound") != 0);

			header->InquiryHeartbeat = 1;
			header->FailoverAbortTaskSet = 0;
			header->FailoverLunReset = 0;
			header->FailoverTgtReset = 0;
			header->Pad = 0;
		} else if (rc == REG_NOMATCH) {
			rc = regexec(&path_re, line_buffer,
				     NUM_ENTRIES(match_list), match_list, 0);
			if (!rc) {
				if (!is_path_configured(line_buffer,
							match_list)) {
					continue;
				}
				this_device |= is_path_eq(Route, line_buffer,
							  match_list);
				if (!this_device)
					continue;
				if (!first_path ||
				    is_path_primary(line_buffer, match_list)) {
					make_path_entry(&path, line_buffer,
							match_list);
					first_path = 1;
				}
				if (header->Type == MPL_PC_TYPE_TRAFFIC_RATIO) {
					do_traffic_ratio(traffic, path_count,
							 &device,
							 line_buffer,
							 match_list);
				}
				++path_count;
			}
		}
		if (rc && rc != REG_NOMATCH) {
			char buffer[64];

			regerror(rc, &device_re, buffer, sizeof(buffer));
			status = MPL_STATUS_UNSUCCESSFUL;
			break;
		}
	}
	if (status == MPL_STATUS_SUCCESS &&
	    header->Type == MPL_PC_TYPE_FAILOVER) {
		failover->PrimaryPathValidationId = device;
		failover->PrimaryPathDevId = path;
		failover->Header.PathCount = path_count;
		failover->Header.Length = sizeof(failover->Header);
	} else if (status == MPL_STATUS_SUCCESS &&
		   header->Type == MPL_PC_TYPE_TRAFFIC_RATIO) {
		traffic->Count = path_count;
		header->PathCount = path_count;
	}
	
	fclose(fp);
out:
	return status;
}



/***************************************************************************
 *
 * mplSetPathControl
 *
 * Set the Path Controls for a specified route identifier.
 *
 * Parameters:
 *
 * RouteIdentifier  - A valid route identifier used to set the path controls
 *
 * PathControl      - A ptr to the buffer containing the path controls struture
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the PathControl had been updated.
 * An error indicating the failure if the request is unsuccessful.
 *
 * Comments:
 *
 * The client builds a Path Control structure and passes to mplSetPathControl
 * to set the new Path Control information for the specified route.
 *
 ***************************************************************************/

MPL_STATUS mplSetPathControl(fcDeviceIdentifier *RouteIdentifier,
                             PMPL_PATH_CONTROL_HEADER PathControl)
{
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;

	if (!initialized) {
		return MPL_STATUS_UNSUCCESSFUL;
	}


	if (RouteIdentifier->m_type != ELX_DI_DEVID_SCSI_NEXUS &&
	    RouteIdentifier->m_type != ELX_DI_DEVID_NEXUS_LONG &&
	    RouteIdentifier->m_type != ELX_DI_DEVID_INQSN) {
		return MPL_STATUS_INVALID_PARAMETER;
	}

	if (PathControl->Type == MPL_PC_TYPE_FAILOVER) {
		status = set_device_failover((struct MplFailoverControl *) 
					     PathControl);
	} else if (PathControl->Type == MPL_PC_TYPE_TRAFFIC_RATIO) {
		set_device_traffic_ratios((struct MplTrafficRatioControl *)
					  PathControl);
		status = set_device_queue_handler(RouteIdentifier,
						  PathControl->Type);
	} else if (PathControl->Type == MPL_PC_TYPE_DYNAMIC_LUNQ ||
		   PathControl->Type == MPL_PC_TYPE_DYNAMIC_TGTQ ||
		   PathControl->Type == MPL_PC_TYPE_DYNAMIC_HBAQ ||
		   PathControl->Type == MPL_PC_TYPE_DYNAMIC_TIME) {
		status = set_device_queue_handler(RouteIdentifier,
						  PathControl->Type);
	} else {
		return MPL_STATUS_UNIMPLEMENTED;
	}

	
	if (status == -EBUSY)
		status = MPL_STATUS_PATH_IN_USE;


	return status;
}



/*******************************************************************************************
 *
 * mplGetPathInfo
 *
 * Get information about the specified path
 *
 * Parameters:
 *
 * DevId             - The DevId portion of the path identifier
 *
 * ValidationId      - The Validation ID port of the path identifier
 *
 * PathInfo          - A pointer to a buffer to receive the response data
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the PathInfo structure is valid 
 * An error indicating the failure if the request is unsuccessful.
 *
 * Comments:
 *
 * The client initializes the PathInfo Length member with the size of the buffer
 * allocated for PathInfo.
 *
 * The client then indicates the path using the DevId and ValidationId values returned
 * in a path list entry from a call to mplGetPaths.
 *
 *******************************************************************************************/

static void
fetch_inquiry(char vendor[12], char product[20], char rev[8],
	      char *line_buffer, regmatch_t *match_list)
{
	fetch_match(vendor, 12, line_buffer,
		    match_list + DEVICE_MATCH_VENDOR);
	fetch_match(product, 20, line_buffer,
		    match_list + DEVICE_MATCH_PRODUCT);
	fetch_match(rev, 8, line_buffer,
		    match_list + DEVICE_MATCH_REV);
}


MPL_STATUS mplGetPathInfo(fcDeviceIdentifier *DevId,
                          fcDeviceIdentifier *ValidationId,
                          PMPL_PATH_INFO PathInfo)
{
	struct ElxDiDevId device;
	struct ElxDiDevId route;
	struct ElxDiDevId path;
	FILE *fp;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	char line_buffer[256];
	regmatch_t match_list[MATCH_COUNT_MAX];
	int  this_device = 0;
	int  rc;

	if (!initialized) {
		return MPL_STATUS_UNSUCCESSFUL;
	}

	if ((DevId->m_type != ELX_DI_DEVID_SCSI_NEXUS &&
	     DevId->m_type != ELX_DI_DEVID_NEXUS_LONG) ||
	    ValidationId->m_type != ELX_DI_DEVID_INQSN) {
		return MPL_STATUS_INVALID_PARAMETER;
	}

	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		status = MPL_STATUS_NO_DEVICE;
		goto out;
	}
	
	status = MPL_STATUS_SUCCESS;
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		get_current_host(line_buffer);
		rc = regexec(&device_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc) {
			if (this_device)
				break;
			this_device = is_scsi_id_eq(ValidationId, line_buffer,
						    match_list);
			make_scsi_id(&device, line_buffer, match_list);
			make_route_id(&route, line_buffer, match_list);
			fetch_inquiry(PathInfo->Vendor, PathInfo->Product,
				      PathInfo->Rev, line_buffer, match_list);
			PathInfo->Length = sizeof(PathInfo);
		} else if (!this_device) {
			continue;
		} else if (rc == REG_NOMATCH) {
			rc = regexec(&path_re, line_buffer,
				     NUM_ENTRIES(match_list), match_list, 0);
			if (!rc) {
				make_scsi_path_entry(&path, line_buffer, 
						     match_list);
				if (this_device && is_path_eq(DevId,
							      line_buffer,
							      match_list)) {
					PathInfo->Port = path.un.m_ScsiId.m_adapter;
					PathInfo->Bus = 0;
					PathInfo->Target = path.un.m_ScsiId.m_target;
					PathInfo->Lun = path.un.m_ScsiId.m_lun;
					PathInfo->RouteId = route;
					break;
				}
			}
		} 
		if (rc && rc != REG_NOMATCH) {
			char buffer[64];

			regerror(rc, &device_re, buffer, sizeof(buffer));
			status = MPL_STATUS_UNSUCCESSFUL;
			break;
		}
	}
	
	fclose(fp);
out:
	if (!this_device) {
		status = MPL_STATUS_UNSUCCESSFUL;
	}
	return status;
}


/***************************************************************************
 *
 * mplGetPathStatus
 *
 * Get the current status of the specified path
 *
 * Parameters:
 *
 * DevId             - The DevId portion of the path identifier
 *
 * ValidationId      - The Validation ID port of the path identifier
 *
 * PathStatus        - A pointer to a buffer to receive the response data
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the PathStatus structure is valid 
 * An error indicating the failure if the request is unsuccessful.
 *
 * Comments:
 *
 * The client initializes the PathStatus Length member with the size of the
 * buffer allocated for PathStatus.
 *
 * The client then indicates the path using the DevId and ValidationId values
 * returned in a path list entry from a call to mplGetPathList.
 *
 ***************************************************************************/

static unsigned int get_path_state(char *line_buffer, regmatch_t *match_list)
{
	char parameter[32];

	fetch_match(parameter, sizeof(parameter), line_buffer,
		    match_list + PATH_MATCH_STATE);
	return !strcmp(parameter, "online") ? MPL_PS_STATE_ONLINE :
		!strcmp(parameter, "standby") ? MPL_PS_STATE_STANDBY :
		!strcmp(parameter, "failed") ? MPL_PS_STATE_FAILED :
		!strcmp(parameter, "disabled") ? MPL_PS_STATE_OFFLINE : 0;
}

MPL_STATUS mplGetPathStatus(fcDeviceIdentifier *DevId,
                            fcDeviceIdentifier *ValidationId,
                            PMPL_PATH_STATUS PathStatus)
{
	struct ElxDiDevId path;
	FILE *fp;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	char line_buffer[256];
	regmatch_t match_list[MATCH_COUNT_MAX];
	int  this_device = 0;
	int  rc;

	if (!initialized) {
		return MPL_STATUS_UNSUCCESSFUL;
	}

	if ((DevId->m_type != ELX_DI_DEVID_SCSI_NEXUS &&
	     DevId->m_type != ELX_DI_DEVID_NEXUS_LONG) ||
	    ValidationId->m_type != ELX_DI_DEVID_INQSN) {
		return MPL_STATUS_INVALID_PARAMETER;
	}

	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		status = MPL_STATUS_NO_DEVICE;
		goto out;
	}
	
	status = MPL_STATUS_SUCCESS;
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		get_current_host(line_buffer);
		rc = regexec(&device_re, line_buffer, NUM_ENTRIES(match_list), 
			     match_list, 0);
		if (!rc) {
			if (this_device)
				break;
			this_device = is_scsi_id_eq(ValidationId, line_buffer,
						    match_list);
		} else if (!this_device) {
			continue;
		} else if (rc == REG_NOMATCH) {
			rc = regexec(&path_re, line_buffer,
				     NUM_ENTRIES(match_list), match_list, 0);
			if (!rc) {
				make_path_entry(&path, line_buffer, 
						match_list);
				if (this_device && is_path_eq(DevId,
							      line_buffer,
							      match_list)) {
					PathStatus->Length = sizeof(*PathStatus);
					PathStatus->State = get_path_state(line_buffer, match_list);
					PathStatus->RatioCount = 0;
					PathStatus->ActiveCount = 0;
					PathStatus->QueueDepth = 32;
					PathStatus->ErrorCount = 0;
					PathStatus->QueueFullCount = 0;
					PathStatus->TotalCmdCount = 0;
					PathStatus->TotalCmplCount = 0;
					PathStatus->FailoverBitMask = failover_bit_mask(line_buffer, match_list);
					PathStatus->Flags = 0;
					PathStatus->Mask = MPL_PATH_MASK_ALL;
					break;
				}
			}
		} 
		if (rc && rc != REG_NOMATCH) {
			char buffer[64];

			regerror(rc, &device_re, buffer, sizeof(buffer));
			status = MPL_STATUS_UNSUCCESSFUL;
			break;
		}
	}
	
	fclose(fp);
out:
	if (!this_device) {
		status = MPL_STATUS_UNSUCCESSFUL;
	}
	return status;
}


/***************************************************************************
 *
 * mplSetPathStatus
 *
 * Get the current status of the specified path
 *
 * Parameters:
 *
 * DevId           - The DevId portion of the path identifier
 *
 * ValidationId    - The Validation ID port of the path identifier
 *
 * PathStatus      - A ptr to a PathStatus structure with the new status values
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the update to the path status was successful.
 * An error indicating the failure if the request is unsuccessful.
 *
 * Comments:
 *
 * The client initializes the PathStatus Length member with the size of the
 * buffer allocated for PathStatus.
 *
 * The client sets the mask word to indicate which fields of the structure are
 * to be updated. For example, to update just the status, set the mask to
 * MPL_PATH_MASK_STATE and set the new status value in the status member of the
 * PathStatus structure.
 *
 * The client then indicates the path using the DevId and ValidationId values
 * returned in a path list entry from a call to mplGetPathList.
 *
 ***************************************************************************/

MPL_STATUS mplSetPathStatus(fcDeviceIdentifier *DevId,
                            fcDeviceIdentifier *ValidationId,
                            PMPL_PATH_STATUS PathStatus)
{
	FILE *fp;
	char path_id[PATH_ID_SIZE];
	char tmp[32];
	MPL_STATUS status = MPL_STATUS_SUCCESS;
	char *function = NULL;;

	if (!initialized) {
		return MPL_STATUS_UNSUCCESSFUL;
	}

	if ((DevId->m_type != ELX_DI_DEVID_SCSI_NEXUS &&
	     DevId->m_type != ELX_DI_DEVID_NEXUS_LONG) ||
	    ValidationId->m_type != ELX_DI_DEVID_INQSN) {
		return MPL_STATUS_INVALID_PARAMETER;
	}

	fp = fopen(MPL_CONTROL_NODE, "r+");
	if (!fp) {
		return MPL_STATUS_NO_DEVICE;
	}

	if (PathStatus->State == MPL_PS_STATE_ONLINE) {
		function = "online";
	} else 	if (PathStatus->State == MPL_PS_STATE_OFFLINE) {
		function = "offline";
	} else if (PathStatus->State) {
		status = MPL_STATUS_INVALID_PARAMETER;
	}
	if (function) {
		fprintf(fp, "path %s status %s\n",
			format_path_id(path_id, sizeof(path_id), DevId),
			function);
	}

	if (PathStatus->Mask == MPL_PATH_MASK_ALL) {
		fprintf(fp, "path %s status clear\n",
			format_path_id(path_id, sizeof(path_id), DevId));
	}

	fflush(fp);
	rewind(fp);
	if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
		fprintf(stderr, "%s: no status\n", __FUNCTION__);
	} else {
		status = strtol(tmp, NULL, 0);
	}

	fclose(fp);

	return status;
}



/***************************************************************************
 *
 * mplAddPathListToRoute
 *
 * Add the paths in the path list to the route specified as the route
 * identifier in the path list.
 *
 * Parameters:
 *
 * PathList     - Points a GET_PATH_LIST structure indicating the route
 *                and new path identifiers. PathList->Count is initialized
 *                to the number of paths to add to the route.
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the paths are successfully added to the route
 * MPL_STATUS_UNSUCCESSFUL completed unsuccessfully for an unspecified reason
 * MPL_STATUS_ROUTE_NOT_FOUND if the indicated route identifier is not valid
 * MPL_STATUS_PATH_NOT_FOUND if any of the indicated path identifiers are not
 * valid.
 *
 * Comments:
 *
 * The Count member of GET_PATH_LIST structure must be initialized to the number
 * of valid path elements in the path list array.
 *
 * The Route member of the GET_PATH_LIST structure should contain the route
 * identifier to indicate the route to be updated.
 *
 ***************************************************************************/

MPL_STATUS mplAddPathListToRoute(PMPL_GET_PATH_LIST PathList)
{
	FILE *fp;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	struct MplPathListEntry *path;
	char path_id[PATH_ID_SIZE];
	char tmp[32];
	int  count = PathList->Count;
	int  i;

	if (!initialized || count < 1) {
		return MPL_STATUS_UNSUCCESSFUL;
	}

	path = PathList->Paths;
	for (i = 0; i < count; ++i) {
		if ((path->DevId.m_type != ELX_DI_DEVID_SCSI_NEXUS &&
		     path->DevId.m_type != ELX_DI_DEVID_NEXUS_LONG) ||
		    path->ValidationId.m_type != ELX_DI_DEVID_INQSN) {
			return MPL_STATUS_INVALID_PARAMETER;
		}
	}

	fp = fopen(MPL_CONTROL_NODE, "r+");
	if (!fp) {
		return MPL_STATUS_NO_DEVICE;
	}

	status = MPL_STATUS_SUCCESS;
	path = PathList->Paths;
	for (i = 0; i < count; ++i) {
		fprintf(fp, "path %s configure\n",
			format_path_id(path_id, sizeof(path_id),
				       &path->DevId));
		fprintf(fp, "path %s status online\n",
			format_path_id(path_id, sizeof(path_id),
				       &path->DevId));
		++path;
		fflush(fp);
		rewind(fp);
		if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
			fprintf(stderr, "%s: no status\n", __FUNCTION__);
			break;
		} else {
			status = strtol(tmp, NULL, 0);
			if (status) {
				break;
			}
		}
	}
	

	fclose(fp);

	return status;
}


/***************************************************************************
 *
 * mplRemovePathListFromRoute
 *
 * Remove the paths in the path list from the route specified as the route
 * identifier in the path list.
 *
 * Parameters:
 *
 * PathList     - Points a GET_PATH_LIST structure indicating the route 
 *                identifier from which the paths in the path list array
 *                are to be removed. PathList->Count is initialized to the
 *                number of paths to remove from the route.
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the paths are successfully removed from the route
 * MPL_STATUS_UNSUCCESSFUL completed unsuccessfully for an unspecified reason
 * MPL_STATUS_ROUTE_NOT_FOUND if the indicated route identifier is not valid
 * MPL_STATUS_PATH_NOT_FOUND if any of the indicated path identifiers are not
 *
 * Comments:
 *
 * The Count member of GET_PATH_LIST structure must be initialized to the number
 * of valid path elements in the path list array.
 *
 * The Route member of the GET_PATH_LIST structure should contain the route
 * identifier to indicate the route to be updated.
 *
 ***************************************************************************/

static MPL_STATUS tweak_paths(struct MplGetPathList *path_list,
			      char *operation,
			      int add_flag)
{
	FILE *fp;
	MPL_STATUS status = MPL_STATUS_NO_DEVICE;
	struct MplPathListEntry *path;
	char path_id[PATH_ID_SIZE];
	char tmp[32];
	int  count = path_list->Count;
	int  i;

	if (!initialized || count < 1) {
		return MPL_STATUS_UNSUCCESSFUL;
	}

	fp = fopen(MPL_CONTROL_NODE, "r+");
	if (!fp) {
		return MPL_STATUS_NO_DEVICE;
	}

	status = MPL_STATUS_SUCCESS;
	path = path_list->Paths;
	for (i = 0; i < count; ++i) {
		fprintf(fp, "path %s %s\n",
			format_path_id(path_id, sizeof(path_id), &path->DevId),
			operation);
		++path;
	}

	fflush(fp);
	rewind(fp);
	if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
		fprintf(stderr, "%s: no status\n", __FUNCTION__);
	} else {
		status = strtol(tmp, NULL, 0);
		if (status == -EBUSY) {
			fclose(fp);
			return MPL_STATUS_UNSUCCESSFUL;
		}
	}

	fclose(fp);

	if (add_flag) {
		path = path_list->Paths;
		for (i = 0; i < count; ++i) {
			int host, target, lun;

			if (!find_path_htl(path, &host, &target, &lun)) {
				add_midlayer_device(host, target, lun);
			}
			++path;
		}
	}

	return status;
}

static int sanity_check_remove_path(struct MplGetPathList *path_list)
{
	fcDeviceIdentifier *route_id;
	struct MplPathListEntry *path;
	int  count = path_list->Count;
	int i;

	route_id = &path_list->RouteId;
	if (route_id->m_type != ELX_DI_DEVID_SCSI_NEXUS &&
	    route_id->m_type != ELX_DI_DEVID_NEXUS_LONG &&
	    route_id->m_type != ELX_DI_DEVID_INQSN) {
		return -1;
	}
	path = path_list->Paths;
	for (i = 0; i < count; ++i) {
				/* Reject removal of presentation path */
		if (path->DevId.m_type != ELX_DI_DEVID_SCSI_NEXUS &&
		    path->DevId.m_type != ELX_DI_DEVID_NEXUS_LONG &&
		    path->DevId.m_type != ELX_DI_DEVID_INQSN) {
			return -1;
		}

		if (route_id->m_len == path->DevId.m_len  &&
		    !memcmp(&route_id->un, &path->DevId.un,
			    route_id->m_len)) {
			return -1;
		}
		++path;
	}
	return 0;
}

MPL_STATUS mplRemovePathListFromRoute(PMPL_GET_PATH_LIST PathList)
{
	if (!initialized) {
		return MPL_STATUS_UNSUCCESSFUL;
	}
	if (sanity_check_remove_path(PathList)) {
		return MPL_STATUS_INVALID_PARAMETER;
	}

	return tweak_paths(PathList, "deconfigure", 1);
}

/***************************************************************************
 *
 * mplRegisterForEvent
 *
 * Register a callback function to respond to events
 *
 * Parameters:
 *
 * EventMask        - The events for which this function is to be called
 *
 * Function         - The function to be called when an event occurs
 *
 * Context          - An opaque value to be returned as an argument to Function
 * 
 * Return Values:
 *
 * A handle to the event function if successful, EMP_INVALID_HANDLE_VALUE if
 * not successful
 *
 * Comments:
 *
 * The client sets the mask for the types of events the function that is being
 * registered will handle. The function is of the type:
 *
 * VOID (*Function)(PVOID Arg,
 *                  MPL_EVENT_TYPE EventType,
 *                  PVOID EventData)
 *
 * In the callback, the Arg parameter has the same value that was passed in to
 * mplRegisterForEvent. The EventType parameter indicates the type of event.
 * The EventData argument is a pointer to the data associated with the event.
 * There are separate structures for each event type.
 * The structure at EventData is only valid until the handler function returns.
 * Any data that is to be used later must be saved by the handler function.
 *
 * The event handler creates a thread to distribute events as they occur. The
 * event handler should not take a long amount of time to execute as this could
 * cause other events to be delayed.
 *
 ***************************************************************************/

MPL_EVENT_HANDLE mplRegisterForEvent(MPL_EVENT_MASK EventMask,
                                     fcDeviceIdentifier *DevId,
                                     void (*Function)(MPL_EVENT_TYPE EventType,
                                                      void *EventData,
                                                      void *Arg),
                                     void *Arg)
{
	return NULL;
}





/***************************************************************************
 *
 * mplUnregisterForEvent
 *
 * Unregister a callback function to respond to events
 *
 * Parameters:
 *
 * EventHandle      - Event handle returned from a previous call to
 *                    mplRegisterForEvent
 * 
 * Return Values:
 *
 * None.
 *
 * Comments:
 *
 * The client calls this function to indicate to the library the library should
 * stop indicating events on the particular callback function.
 *
 ***************************************************************************/

void mplUnregisterForEvent(MPL_EVENT_HANDLE EventHandle)
{
}


/***************************************************************************
 *
 * mplCommitRouteToNonvolatileStorage
 *
 * Cause information to be written into non-volatile storage so the route can be
 * reconstructed at the next system boot.
 *
 * Parameters:
 *
 * RouteId        - Device Identifier identifying the route
 * 
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the request was successful. If the request was not
 * successful, then an error code indicating the failure.
 *
 * Comments:
 *
 * The client calls this function to cause the indicated route to be saved in
 * non-volatile storage so the route can be reconstructed at the next system
 * restart. The information that will be saved is the current configuration of
 * the route.
 *
 ***************************************************************************/

MPL_STATUS mplCommitRouteToNonvolatileStorage(fcDeviceIdentifier *RouteId)
{
	return MPL_STATUS_UNIMPLEMENTED;
}


/***************************************************************************
 *
 * mplRemoveRouteFromNonvolatileStorage
 *
 * Cause the route information to be removed from non-volatile storage.
 * This will not remove the route from the current instance of the driver.
 *
 * Parameters:
 *
 * RouteId        - Device Identifier identifying the route
 * 
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the request was successful. If the request was not
 * successful, then an error code indicating the failure.
 *
 * Comments:
 *
 * The client calls this function to cause the indicated route to be removed
 * from non-volatile storage so the route will not be reconstructed at the 
 * next system restart. 
 *
 ***************************************************************************/

MPL_STATUS mplRemoveRouteFromNonvolatileStorage(fcDeviceIdentifier *RouteId)
{
	return MPL_STATUS_UNIMPLEMENTED;
}


/***************************************************************************
 *
 * mplGetRouteControlTypes
 *
 * Retrieve the different types of route controls available.
 *
 * Parameters:
 *
 * None
 * 
 * Return Values:
 *
 * A pointer to the Route Control Type table. NULL is returned if there are
 * no route control types available
 * 
 * Comments:
 *
 * The client calls this function to retrieve the different route control types
 * avaiable.
 * An example of the different types are Simple Failover, Lun balancing, etc.
 *
 ***************************************************************************/

PMPL_ROUTE_CONTROL_TYPES mplGetRouteControlTypes(void)
{
	static struct MplRouteControlTypes mpl_control_types = {
		MPL_PC_TYPE_COUNT, 0,
		{
			{
				"Simple Failover",
				MPL_PC_TYPE_FAILOVER,
			},
			{
				"Round Robin",
				MPL_PC_TYPE_TRAFFIC_RATIO,
			},
			{
				"Dynamic LUN queue depth",
				MPL_PC_TYPE_DYNAMIC_LUNQ,
			},
			{
				"Dynamic target queue depth",
				MPL_PC_TYPE_DYNAMIC_TGTQ,
			},
			{
				"Dynamic HBA queue depth",
				MPL_PC_TYPE_DYNAMIC_HBAQ,
			},
			{
				"Least service time",
				MPL_PC_TYPE_DYNAMIC_TIME,
			},
		},
	};


	return (PMPL_ROUTE_CONTROL_TYPES) &mpl_control_types;
}

/***************************************************************************
 *
 * mplClose
 *
 * The client application calls this routine to indicate that it is finished
 * using the MultiPulse API library. This function frees resources allocated
 * during mplInitialize.
 *
 * Parameters:
 *
 * None.
 *
 * Return Values:
 *
 * None.
 *
 * Comments:
 *
 * This function should be called only when there will be no more calls into
 * the MultiPulse API library by the client application. After this function
 * has been called, the application would have to call mplInitialize to use
 * any MultiPulse API library function calls.
 *
 ***************************************************************************/

void mplClose()
{
	regfree(&hex_wwn_re);
	regfree(&colon_separated_wwn_re);
	regfree(&version_re);
	regfree(&port_name_re);
	regfree(&port_state_re);
	regfree(&device_re);
	regfree(&path_re);
	regfree(&host_re);
	regfree(&cfg_param1_re);
	regfree(&cfg_param2_re);
	regfree(&cfg_param3_2_re);
	regfree(&cfg_param3_1_re);
	initialized = 0;
}

/***************************************************************************
 *
 * mplExportPaths
 *
 * Cause the paths in the path list structure to not be used by the 
 * MultiPulse driver.
 *
 * Parameters:
 *
 * PathList     - Points a GET_PATH_LIST structure indicating the paths  
 *                that are to be removed. PathList->Count is initialized to 
 *                the number of paths to remove from the MultiPulse subsystem.
 *                The route identifier is ignored.
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the paths are successfully removed
 * MPL_STATUS_UNSUCCESSFUL completed unsuccessfully for an unspecified reason
 * MPL_STATUS_PATH_NOT_FOUND if any of the indicated path identifiers are not
 * found
 *
 * Comments:
 *
 * If MPL_STATUS_PATH_NOT_FOUND is returned, some of the paths may be changed
 * to the not filter status. It is up to the caller to determine if any of the 
 * paths were successfully updated.
 *
 * The Count member of GET_PATH_LIST structure must be initialized to the number
 * of valid path elements in the path list array.
 *
 * Once the paths are removed from the MultiPulse driver, they will no longer be
 * visible via the normal MultiPulse library calls. In order to get a list
 * of paths that are being exported, use mplGetExportedPaths.
 *
 ***************************************************************************/

MPL_STATUS
mplExportPaths(PMPL_GET_PATH_LIST PathList)
{
	return tweak_paths(PathList, "export", 0);
}


/***************************************************************************
 *
 * mplImportPaths
 *
 * Cause the paths in the path list structure to be used by the 
 * MultiPulse filter driver.
 *
 * Parameters:
 *
 * PathList     - Points a GET_PATH_LIST structure indicating the paths  
 *                that are to be restored. PathList->Count is initialized to 
 *                the number of paths to restore to the MultiPulse driver. The 
 *                route identifier is ignored.
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the paths are successfully restored
 * MPL_STATUS_UNSUCCESSFUL completed unsuccessfully for an unspecified reason
 * MPL_STATUS_PATH_NOT_FOUND if any of the indicated path identifiers are not
 * found
 *
 * Comments:
 *
 * If MPL_STATUS_PATH_NOT_FOUND is returned, some of the paths may have been 
 * restored to filtering. It is up to the caller to determine if any of the 
 * paths were successfully updated.
 *
 * The Count member of GET_PATH_LIST structure must be initialized to the number
 * of valid path elements in the path list array.
 *
 ***************************************************************************/

MPL_STATUS
mplImportPaths(PMPL_GET_PATH_LIST PathList)
{
	return tweak_paths(PathList, "deconfigure", 0);
}


/***************************************************************************
 *
 * mplGetExportedPaths
 *
 * Obtain a list of paths that are currently not being filtered by the driver.
 *
 * Parameters:
 *
 * PathList     - Points a GET_PATH_LIST structure indicating the paths  
 *                that are being exporteded. PathList->Count is initialized 
 *                to the number entries available in the path list array. The 
 *                route identifier is ignored.
 *
 * Return Values:
 *
 * MPL_STATUS_SUCCESS if the paths are successfully restored
 * MPL_STATUS_UNSUCCESSFUL completed unsuccessfully for an unspecified reason
 * MPL_STATUS_BUFFER_TOO_SMALL if PathList is too small to hold all export paths
 *
 * Comments:
 *
 * The Count member of GET_PATH_LIST structure must be initialized to the number
 * of available path elements in the path list array.
 *
 ***************************************************************************/

MPL_STATUS
mplGetExportedPaths(PMPL_GET_PATH_LIST PathList)
{
	return mpl_get_unbound_path_list(PathList, is_path_exported);
}


int
mplGetCfgParam(MPL_CFGPARAM *cfgparam)
{
	FILE *fp;
	int status = 0;
	MPL_CFGPARAM *cfgp;
	char line_buffer[256];
	regmatch_t match_list[5];
	char cfg_value[32];
	int  rc;

	if (!initialized) {
		goto out;
	}
	fp = fopen(MPL_CONFIG_NODE, "r");
	if (!fp) {
		goto out;
	}
	
	while (fgets(line_buffer, sizeof(line_buffer), fp)) {
		rc = regexec(&cfg_param1_re, line_buffer,
			NUM_ENTRIES(match_list), match_list, 0);
		if (!rc) {
			if (fetch_match(cfg_value, sizeof(cfg_value),
				       line_buffer,
				       match_list + CFG_MATCH_LOG_VERBOSE)) {
				cfgp = cfgparam + MPL_CFG_LOG_VERBOSE;
				strncpy(cfgp->a_string, "log-verbose",
					sizeof(cfgp->a_string));
				cfgp->a_string[sizeof(cfgp->a_string) - 1] = 0;
				cfgp->a_help[0] = 0;
				cfgp->a_current = strtol(cfg_value, 0, 16);
				cfgp->a_low = 0;
				cfgp->a_hi = 0xffff;
				cfgp->a_default = 0;
				cfgp->a_flag = MPL_CFG_EXPORT;
				cfgp->a_changestate = MPL_CFG_DYNAMIC;
				status = MPL_CFG_LOG_VERBOSE+1;
			}
			if (fetch_match(cfg_value, sizeof(cfg_value),
				       line_buffer,
				       match_list + CFG_MATCH_HBEAT_INTERVAL)) {
				cfgp = cfgparam + MPL_CFG_HBEAT_INTERVAL;
				strncpy(cfgp->a_string, "hbeat-interval",
					sizeof(cfgp->a_string));
				cfgp->a_string[sizeof(cfgp->a_string) - 1] = 0;
				cfgp->a_help[0] = 0;
				cfgp->a_current = atoi(cfg_value);
				cfgp->a_low = 1;
				cfgp->a_hi = 256;
				cfgp->a_default = 30;
				cfgp->a_flag = MPL_CFG_EXPORT;
				cfgp->a_changestate = MPL_CFG_DYNAMIC;
				status = MPL_CFG_HBEAT_INTERVAL+1;
			}
			if (fetch_match(cfg_value, sizeof(cfg_value),
				       line_buffer,
				       match_list + CFG_MATCH_HBEAT_FAIL)) {
				cfgp = cfgparam + MPL_CFG_HBEAT_FAIL_INTERVAL;
				strncpy(cfgp->a_string, "hbeat-fail",
				       sizeof(cfgp->a_string));
				cfgp->a_string[sizeof(cfgp->a_string) - 1] = 0;
				cfgp->a_help[0] = 0;
				cfgp->a_current = atoi(cfg_value);
				cfgp->a_low = 1;
				cfgp->a_hi = 256;
				cfgp->a_default = 1;
				cfgp->a_flag = MPL_CFG_EXPORT;
				cfgp->a_changestate = MPL_CFG_DYNAMIC;
				status = MPL_CFG_HBEAT_FAIL_INTERVAL+1;
			}
			if (fetch_match(cfg_value, sizeof(cfg_value),
				       line_buffer,
				       match_list + CFG_MATCH_HBEAT_RETRY)) {
				cfgp = cfgparam + MPL_CFG_HBEAT_RETRY;
				strncpy(cfgp->a_string, "hbeat-retry",
					sizeof(cfgp->a_string));
				cfgp->a_string[sizeof(cfgp->a_string) - 1] = 0;
				cfgp->a_help[0] = 0;
				cfgp->a_current = atoi(cfg_value);
				cfgp->a_low = 1;
				cfgp->a_hi = 256;
				cfgp->a_default = 60;
				cfgp->a_flag = MPL_CFG_IGNORE;
				cfgp->a_changestate = MPL_CFG_DYNAMIC;
				status = MPL_CFG_HBEAT_RETRY+1;
			}
		}
		rc = regexec(&cfg_param2_re, line_buffer,
			NUM_ENTRIES(match_list), match_list, 0);
		if (!rc) {
			if (fetch_match(cfg_value, sizeof(cfg_value),
				       line_buffer,
				       match_list + CFG_MATCH_AUTO_CONFIGURE)) {
				cfgp = cfgparam + MPL_CFG_AUTO_CONFIGURE;
				strncpy(cfgp->a_string, "auto-configure",
					sizeof(cfgp->a_string));
				cfgp->a_string[sizeof(cfgp->a_string) - 1] = 0;
				cfgp->a_help[0] = 0;
				cfgp->a_current = atoi(cfg_value);
				cfgp->a_low = 1;
				cfgp->a_hi = MPL_PC_TYPE_COUNT;
				cfgp->a_default = MPL_PC_TYPE_FAILOVER;
				cfgp->a_flag = MPL_CFG_IGNORE;
				cfgp->a_changestate = MPL_CFG_REBOOT;
				status = MPL_CFG_AUTO_CONFIGURE+1;
			}
			if (fetch_match(cfg_value, sizeof(cfg_value),
				       line_buffer,
				       match_list + CFG_MATCH_AUTO_BALANCE)) {
				cfgp = cfgparam + MPL_CFG_AUTO_LUN_BALANCE;
				strncpy(cfgp->a_string, "auto-balance",
					sizeof(cfgp->a_string));
				cfgp->a_string[sizeof(cfgp->a_string) - 1] = 0;
				cfgp->a_help[0] = 0;
				cfgp->a_current = atoi(cfg_value);
				cfgp->a_low = 0;
				cfgp->a_hi = 1;
				cfgp->a_default = 1;
				cfgp->a_flag = MPL_CFG_IGNORE;
				cfgp->a_changestate = MPL_CFG_REBOOT;
				status = MPL_CFG_AUTO_LUN_BALANCE+1;
			}
		}
		rc = regexec(&cfg_param3_2_re, line_buffer,
			NUM_ENTRIES(match_list), match_list, 0);
		if (!rc) {
			if (fetch_match(cfg_value, sizeof(cfg_value),
				       line_buffer,
				       match_list + CFG_MATCH_ENABLE_TGT_QDEPTH)) {
				cfgp = cfgparam + MPL_CFG_ENABLE_TGT_QDEPTH;
				strncpy(cfgp->a_string, "enable-tgt-qdepth",
					sizeof(cfgp->a_string));
				cfgp->a_string[sizeof(cfgp->a_string) - 1] = 0;
				cfgp->a_help[0] = 0;
				cfgp->a_current = atoi(cfg_value);
				cfgp->a_low = 0;
				cfgp->a_hi = 1;
				cfgp->a_default = 1;
				cfgp->a_flag = MPL_CFG_EXPORT;
				cfgp->a_changestate = MPL_CFG_DYNAMIC;
				status = MPL_CFG_ENABLE_TGT_QDEPTH+1;
			}
			if (fetch_match(cfg_value, sizeof(cfg_value),
				       line_buffer,
				       match_list + CFG_MATCH_HANDLE_QFULL)) {
				cfgp = cfgparam + MPL_CFG_HANDLE_QFULL;
				strncpy(cfgp->a_string, "handle-qfull",
					sizeof(cfgp->a_string));
				cfgp->a_string[sizeof(cfgp->a_string) - 1] = 0;
				cfgp->a_help[0] = 0;
				cfgp->a_current = atoi(cfg_value);
				cfgp->a_low = 0;
				cfgp->a_hi = 1;
				cfgp->a_default = 1;
				cfgp->a_flag = MPL_CFG_EXPORT;
				cfgp->a_changestate = MPL_CFG_DYNAMIC;
				status = MPL_CFG_HANDLE_QFULL+1;
			}
		}
		rc = regexec(&cfg_param3_1_re, line_buffer,
			NUM_ENTRIES(match_list), match_list, 0);
		if (!rc) {
			if (fetch_match(cfg_value, sizeof(cfg_value),
				       line_buffer,
				       match_list + CFG_MATCH_ENABLE_TGT_QDEPTH)) {
				cfgp = cfgparam + MPL_CFG_ENABLE_TGT_QDEPTH;
				strncpy(cfgp->a_string, "enable-tgt-qdepth",
					sizeof(cfgp->a_string));
				cfgp->a_string[sizeof(cfgp->a_string) - 1] = 0;
				cfgp->a_help[0] = 0;
				cfgp->a_current = atoi(cfg_value);
				cfgp->a_low = 0;
				cfgp->a_hi = 1;
				cfgp->a_default = 1;
				cfgp->a_flag = MPL_CFG_EXPORT;
				cfgp->a_changestate = MPL_CFG_DYNAMIC;
				status = MPL_CFG_ENABLE_TGT_QDEPTH+1;
			}
		}
	}
	fclose(fp);
out:
	return status;
}

int
mplSetCfgParam(int offset, int value)
{
	FILE *fp;
	char tmp[32];
	int  rc = 0;

	if (!initialized) {
		goto out;
	}
	fp = fopen(MPL_CONTROL_NODE, "r+");
	if (!fp) {
		goto out;
	}
	
	switch (offset) {
	case MPL_CFG_LOG_VERBOSE:        /* mpl_log_verbose */
		fprintf(fp, "debug level %d\n", value);
		break;
	case MPL_CFG_HBEAT_INTERVAL:     /* mpl_hbeat_interval */
		fprintf(fp, "hbeat %d\n", value);
		break;
	case MPL_CFG_HBEAT_FAIL_INTERVAL:/* mpl_fail_hbeat_interval */
		fprintf(fp, "hbeat fail %d\n", value);
		break;
	case MPL_CFG_AUTO_CONFIGURE:     /* mpl_auto_configure */
	case MPL_CFG_AUTO_LUN_BALANCE:   /* mpl_auto_lun_balance */
		/* can't change dynamically */
		rc = 1;
		break;
	case MPL_CFG_HBEAT_RETRY:        /* mpl_max_hbeat_retry */
		fprintf(fp, "hbeat retry %d\n", value);
		break;
	case MPL_CFG_ENABLE_TGT_QDEPTH:  /* mpl_enable_tgt_qdepth */
		fprintf(fp, "tqdepth %d\n", value);
		break;
	case MPL_CFG_HANDLE_QFULL:       /* mpl_handle_qfull */
		fprintf(fp, "qfull %d\n", value);
		break;
	}

	if (!rc) {
		fflush(fp);
		rewind(fp);
		if (fread(tmp, 1, sizeof(tmp), fp) <= 0) {
			fprintf(stderr, "%s: no status\n", __FUNCTION__);
			rc = 1;
		} else {
			if (strtol(tmp, NULL, 0)) {
				rc = 1;
			}
		}
	}
	fclose(fp);
out:
	return rc;
}

