/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* mpl.h: MultiPulse for Linux definitions and prototypes
 * $Id: mpl.h 1340 2007-05-07 15:01:42Z jinfante $
 */

#include <linux/ctype.h>
#include "mplib.h"
#include "mpl_uid.h"

#if !defined(list_first)	/* 2.4.21 ism */
#define list_first(head)      (((head)->next != (head)) ? (head)->next: (struct list_head *) 0)
#endif

#if !defined(WARN_ON)
#define WARN_ON(condition) do { \
		if (unlikely((condition)!=0)) {				\
			printk("Badness in %s at %s:%d\n", __FUNCTION__, \
			       __FILE__, __LINE__);			\
			dump_stack();					\
		}							\
	} while (0)
#endif

#if !defined(__user)
#define __user
#endif
#if !defined(list_for_each_entry_safe)
#define list_for_each_entry_safe(pos, n, head, member)			\
	for (pos = list_entry((head)->next, typeof(*pos), member),	\
		n = list_entry(pos->member.next, typeof(*pos), member);	\
	     &pos->member != (head); 					\
	     pos = n, n = list_entry(n->member.next, typeof(*n), member))
#endif


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static inline u64 wwn_to_u64(u8 *wwn)
{
	return (u64)wwn[0] << 56 | (u64)wwn[1] << 48 |
	    (u64)wwn[2] << 40 | (u64)wwn[3] << 32 |
	    (u64)wwn[4] << 24 | (u64)wwn[5] << 16 |
	    (u64)wwn[6] <<  8 | (u64)wwn[7];
}
#endif

#define ScsiResult(host_code, scsi_code) (((host_code) << 16) | scsi_code)

#define MPL_INTERNAL_CMD_TMO  60
#define MPL_INQUIRY_CMD_TMO   20
#define MPL_TUR_CMD_TMO       30
#define MPL_RETRY_CMD_TMO     10

enum deferral_reason {
	MPL_OOS_INQ_CMD,
	MPL_INQ_CMD,
	MPL_INT_CMD,
	MPL_HB_CMD,
	MPL_FAILED_CMD,
	MPL_RETRY_CMD,
	MPL_HEARTBEAT_CMD,
	MPL_INTERNAL_CMD,
	MPL_ABORTED_CMD,
	MPL_DO_ABORT,
	MPL_DO_DEV_RESET,
	MPL_DO_DEV_SCAN,
	MPL_DO_TMO,
	MPL_DO_INT_TMO,
	MPL_DO_HBEAT_TMO
};

enum mpl_stub_state {
	STUB_ALLOC,		/* Initial state, allocated */
	STUB_DQ,		/* Not in any q */
	STUB_LPFCQ,		/* Sent to lpfc base driver for processing */
	STUB_PENDQ,		/* On pending queue for device */
	STUB_DEFERQ,		/* Background processing for cmpl */
	STUB_EH_DEFERQ,		/* Background processing for timeout */
	STUB_FREE		/* Freed */
};

/* Defines for mpl_find_scsi_cmnd routines */
#define MPL_INTERNAL_CMND 1
#define MPL_SCSI_CMND     2

/* Defines for mpl abort handling */
#define MPL_ABORT_EXACT             0x0
#define MPL_ABORT_DEVICE            0x1
#define MPL_ABORT_HOST              0x2
#define MPL_ABORT_TYPE_MASK         0xf
#define MPL_ABORT_IGNORE_ABORTED    0x10


struct io_stub 
{
	unsigned long magic;         /* Magic number to mark start of stub */
#define MPL_STUB_MAGIC  (unsigned long)(0x12345678)
	struct list_head stub_queue; /* linkage for I/Os sent to lpfc driver */
				     /* linkage for deferral or device I/O
				      * pending queue */
	enum deferral_reason reason;
	enum   mpl_stub_state  state;
        uint8_t flag;
#define MPL_DEC_CAN_QUEUE     1      /* Stub already decremented canq */
#define MPL_INTERNAL_CMD_STUB 2      /* Internal command */
#define MPL_SCSI_CMD_STUB     4      /* SCSI Layer initiated command */
#define MPL_ABORT_FAILED      8	     /* Abort issued, but failed */
	struct scsi_cmnd *cmdp;      /* mid-layer command structure */
	void (*done)(struct scsi_cmnd *); /* mid-layer scsii_done routine */
	struct scsi_device *sdev;    /* mid-layer scsi_device */
	struct mpl_path *path;       /* pointer back to path, device, host */
	struct Scsi_Host *scsi_host; /* pointer to associated mid-layer data */
	unsigned int target;	     /* path identifier: target no. */
	unsigned int lun;	     /* path identifier: lun */
	unsigned long start_time;    /* jiffies when IO started */
	void (*timeout_handler)(unsigned long); /* used for restoring scsi cmds
                                                 * timeout handler */
	struct scsi_cmnd cmd;        /* copy of mid-layer command structure */
};

#define MPL_LOCATE_STUB_SZ (sizeof(struct io_stub) - sizeof(struct scsi_cmnd))
#define mpl_locate_stub(cmd) (struct io_stub *)(((unsigned char *)cmd) - MPL_LOCATE_STUB_SZ)

struct mpl_host
{
	struct list_head host_list;  /* list of all recognized hosts */
	struct list_head lpfc_queue; /* linkage for I/Os sent to lpfc driver */
	struct Scsi_Host *scsi_host; /* pointer to associated mid-layer data */
	struct list_head devices;    /* list of all devices on this host */
	void *hba;	             /* pointer to lld hba structure */
	struct timer_list host_block;/* timer for host being blocked */
	struct timer_list host_scan; /* timer for host being scanned */
	struct timer_list host_hbeat;/* timer for heartbeat */
	uint8_t wwpn[8];             /* WWPN for this HBA */
	uint16_t host_no;	     /* host number (the X in path:
				      * lpfcXtYdZ) */
	uint16_t dev_count;	     /* count of devices */
	uint32_t ref_count;	     /* reference count for scsi_host_get */
	uint32_t can_queue;          /* Original value of scsi_host can_queue */
	uint32_t blocked;            /* Host block count */ 

	/* host flags */
	unsigned int flags; 
#define MPL_HOST_SCAN_INP	1    /* Host is being scanned */
#define MPL_HOST_SCAN_BLOCKED	2    /* scsi scan is blocked for host */
#define MPL_HOST_HBEAT_START    8    /* Host hbeat timer is running */

	atomic_t io_count;	     /* outstanding I/O operations */
	atomic_t tot_count;	     /* total ongoing I/O operations */
};

enum mpl_device_state {
	DEV_ONLINE,		/* Nomal state */
	DEV_FAILING,		/* Temporarily not available.  Recovery 
				 * is in process. INITIAL device state.
				 * Inquiry can go down.
				 * All new cmds go to device's pend queue.
				 * Old cmds that failed go to pend queue.
				 */ 

	DEV_FAILED,		/* dev failed, no path available.  all
				 * new cmds and those that failed get
				 * returned with DID_NO_CONNECT.
				 * Internal cmds can go down.  Only a
				 * path coming ONLINE can bring it
				 * back*/
	DEV_DISABLED,		/* Administratively down */
};

enum mpl_device_event {
	DEV_PATH_UP,		/* Path coming online */
	DEV_PATH_DOWN,		/* Path has failed */
	DEV_DISABLE,		/* Device going administratively down. */
};

enum mpl_config_state {		/* Used for app lib management */
	CONFIG_UNCONFIGURED,	/* Not configured (discovered) */
	CONFIG_CONFIGURED,	/* Configured (lpfc.conf or api lib) */
	CONFIG_UNAVAILABLE,	/* Unavailable to MultiPulse from UI POV */
};

/* The presentation id represents how devices are identified between the SCSI mid-layer
 * and the driver.
 */
struct mpl_presentation
{
	int    host_no;		/* Route ID defined by API */
	int    tgt_no;		/* Route ID defined by API */
	int    lun_no;		/* Route ID defined by API */
};


#define MPL_MAX_STATIC_DEVICES 1024 /* max # of static devices that can be
				     * defined in configuration file.
				     * Be careful as this value, 1024, is
				     * hardcoded in mplize_conf.pl
				     */

#define MPL_MAX_EXCLUSION_DEVICES 1024 /* max # of exclusion entries that can be
				     * defined in configuration file.
				     * Be careful as this value, 1024, is
				     * hardcoded in mplize_conf.pl
				     */

struct mpl_device
{
	struct mpl_device *next;      /* hash collision chain (keyed by
				       * presentation path id) */
	struct list_head devices;     /* list of devices on host */
	struct mpl_host  *host;	      /* parent host (presentation path) */
	struct mpl_plugin *plugin;    /* plugin signed up to support this dev */
	struct mpl_presentation id;   /* presentation id for device */
	uint8_t scsi_type;	      /* scsi dev type (da, tape, etc.) */
	uint8_t vendor[8];	      /* inquiry vendor id */
	uint8_t product[16];	      /* inquiry product id */
	uint8_t rev[4];		         /* inquiry revision code */

	/* Device flags */
	unsigned int tpg_cmd_inp: 1;         /* RTPG / STPG cmd is in progress */
	unsigned int alua_support: 1;        /* Device supports ALUA */
	unsigned int pref_controller: 1;     /* Device supports a preferred
					      * controller */
	unsigned int active_passive: 1;      /* Device is active-passive 
					      * as opposed to active-active */
	unsigned int auto_failback: 1;       /* failback to primary is automatic */
	unsigned int optimized_only: 1;      /* load balance active optimized paths */
	unsigned int inquiry_heartbeat: 1;   /* shouldn't this be per path? */
   	unsigned int StaticConfig:        1;  /* Device is Statically Configured */
   	unsigned int StaticConfigPrimary: 1;  /* Device has primary path Statically Configured */
	unsigned int dev_need_inqsn_data :1; /* need to capture serial num */
	unsigned int on_reaperq :1;          /* On reaper q to be deleted */
	unsigned int just_created :1;        /* device was just created */
	unsigned int path_cntl_type : 4;     /* path control type */
	unsigned int reset_pending :1;       /* device was just reset */
	unsigned int deconfig_api : 1;       /* device was deconfigured thru user api */
	unsigned int pad : 14;

	unsigned short queue_max;     /* device initial queue depth */
	atomic_t opt_path_cnt;	      /* optimized path count */
	atomic_t ref_count;	      /* reference count */
	atomic_t io_count;            /* # of Linux scsi cmd we owe upper layer */ 
	atomic_t tot_count;	     /* total ongoing I/O operations */
	atomic_t pending_count;       /* # of Linux scsi cmd we 
                                       * have on pend list */
	atomic_t deferral_count;      /* # of Linux scsi cmd we 
                                       * have on deferral list */
	atomic_t eh_deferral_count;      /* # of Linux scsi cmd we 
                                       * have on eh_deferral list */
	struct list_head pending_ios; /* io_stub list */
	struct mpl_path *last_path;   /* path used to send last I/O req */
	struct mpl_path *failing_path; /* last_path at time of failover */
	struct list_head paths;	      /* list of paths to this
				       * device, presentation path
				       * is first path on list. */
	enum mpl_device_state state;  /* up, down, in recovery */
	size_t   serial_size;	      /* actual bytes in serial number */
	uint8_t  serial_number[256];
	enum mpl_config_state   configured; /* configured or discovered? */
				      /* API related fields */
	uint8_t pending_lun_migration;
	struct timer_list nopath;     /* timer for extra nopath tmo */
	unsigned int tgt_queue_depth; /* tgt queue depth assigned by plugin */
	unsigned int nopath_tmo;      /* extra timed allowed for no paths */
};


struct mpl_target
{
	struct list_head target_list; /* list of all recognized targets */
	struct list_head paths;       /* list of paths to this target */
	struct mpl_host *host;	      /* host this target connects through */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
	struct fc_rport *rport;	      /* rport this target connects through */
	struct timer_list dev_loss;   /* timer to simulate dev_loss tmo */
	uint32_t roles;	              /* used to save rport info */
	enum fc_port_state port_state;/* used to save rport info */
#endif
	uint8_t wwpn[8];              /* WWPN for this target NPort */
	uint8_t wwnn[8];              /* WWPN for this target NPort */
	unsigned int id;	      /* scsi identifier: target no. */
	atomic_t path_ref_count;      /* total paths assoc with tgt */
	atomic_t prim_path_ref_count; /* total primary paths assoc with tgt */
	atomic_t io_count;	      /* outstanding I/O operations */
	atomic_t tot_count;	      /* total ongoing I/O operations */
};

enum mpl_path_state {
	PATH_ONLINE,		     /* healthy path - in I/O path */
	PATH_FAILED,		     /* unhappy path - not operational */
	PATH_STANDBY,                /* healthy path - not currently in
				      * I/O path */
	PATH_DISABLED,		     /* operationally offline */
};

enum mpl_path_event {
	PATH_DOWN,		     /* Path is failing */
	PATH_UP,		     /* Path is up */
	PATH_DISABLE,		     /* Set administratively down */
	PATH_ENABLE,		     /* Set administratively up */
	PATH_FAILBACK,		     /* Path will fail back */
	PATH_FAILOVER,		     /* Path will fail over */
};

enum mpl_hb_state {
	HB_STOPPED,		     /* initialization */
	HB_TUR,	                     /* sent TUR for hbeat */
	HB_RTPG,	             /* sent RTPG for hbeat */
	HB_INQSN,                    /* sent INQ SN for hbeat */
	HB_WAITING,		     /* waiting for send hbeat tmr to expire */
};

enum mpl_hb_event {
	HB_START,		      /* Start heart beat if stopped */
	HB_STOP,		      /* no heartbeat required */
	HB_SUCCEED,		      /* Successful hb, inq s/n matched */
	HB_FAIL,		      /* inq s/n didn't match or tur failed */
	HB_TIMEOUT,		      /* inquiry timed out */
	HB_BUSY,		      /* device queue full; perhaps blocked */
};

struct mpl_path 
{
	struct mpl_path  *next;	      /* hash collision chain (keyed by
				       * host/targ/lun) */
	struct list_head path_list;   /* other paths, same device */
	struct list_head tgt_path_list; /* list of all paths on this target */
	struct mpl_device *device;    /* associated device */
	struct mpl_target *tgt;       /* associated target */

	unsigned int channel;         /* path identifier: channel */
	unsigned int target;	      /* path identifier: target no. */
	unsigned int lun;	      /* path identifier: lun */
	struct mpl_host *host;	      /* host this path connects through */

	enum   mpl_path_state  state;
	enum   mpl_hb_state    hbstate;

	/* Path flags */
	unsigned int is_default :    1;	    /* configured as default path */
   	unsigned int StaticConfig :  1;     /* Path is Statically Configured */
	unsigned int seen_ext_inq :  1;      /* mid-layer scan detected */
	unsigned int got_wwpn_data : 1;     /* wwpn for target copied into path */
	unsigned int active_controller : 2; /* path is to active controller */
	unsigned int pref_controller : 1;   /* path is to preferred controller */
	uint32_t flag;

	/* Used for Target Port Group handling */
	uint32_t tgt_port_group;
	uint32_t relative_tgt_port;
	uint8_t *tgt_port_info;

	unsigned long avg_time;       /* avg time, jiffies, for an IO */
	struct timer_list hbeat;      /* timer for heartbeat */
	struct timer_list sched_cmd;  /* timer to send recovery cmds */
	int    retries;               /* retry count for cmd to be retried */
	int    failures;              /* retry count for failed hbeat cmd */
	atomic_t avg_cnt;	      /* I/O count used for avg time */
	atomic_t io_count;	      /* outstanding I/O count */
	atomic_t internal_count;      /* active internal I/O count */
	atomic_t tot_count;	      /* total ongoing I/O operations */
	uint8_t traffic_ratio;	      /* load balancing ratio */
	uint8_t traffic_count;        /* load balancing I/O count */
	enum mpl_config_state configured; /* configured or discovered? */
	struct scsi_device    *sdev; /* scsi device used by slave_alloc */
	struct io_stub *delayed_cmd; /* mid-layer command being held until
				      * an internal command completes */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	struct timer_list sdev_block;/* Timer for target block */
#endif
	uint8_t lld_blocked;         /* Flag to indicate Low-layer driver
                                        blocked this path.              */
	uint8_t tpgs;	             /* Target Port Group Support - ALUA */
#define MPL_ALUA_TPGS_MASK	0x30
#define MPL_ALUA_NOT_SUP	0x0
#define MPL_ALUA_ACTIVE_ACTIVE	0x10 /* or 0x30 */
#define MPL_ALUA_ACTIVE_PASSIVE	0x20
	uint8_t sas;	             /* Symetric Access State - ALUA */
#define MPL_ALUA_SAS_MASK	0xf
#define MPL_ALUA_ACTIVE_OPT	0x0
#define MPL_ALUA_ACTIVE_NOT_OPT	0x1
#define MPL_ALUA_STANDBY	0x2
#define MPL_ALUA_UNAVAILABLE	0x3
	uint8_t refcnt;	             /* PSM reference count */
};

/* path's flag define */
#define PATH_NEEDS_TUR           0x2    /* need to send TUR for this path */
#define PATH_NEEDS_INQUIRYP0     0x4    /* need to send Inq Page 0 for this path */
#define PATH_NEEDS_STARTSTOP     0x8    /* need to send Start for this path */
#define PATH_NEEDS_SET_TGT_PORT  0x10   /* need to send Set Tgt Port Group for this path */
#define PATH_NEEDS_ALL_CMDS      0x1e    /* mask for all "needs" cmds */
#define PATH_OUT_CHG_LUN_OWN_CMD 0x100	/* need to call startstop cmpl for this cmd */
#define PATH_JUST_ENABLED        0x200   /* path enabled from disabled state */
#define PATH_FAST_POLL           0x400   /* path in fast hbeat poll mode */
#define PATH_DEV_QUIESCE         0x800   /* need to quiese device before issuing
					 * path command.
					 */
#define PATH_ASSIGN_TPG          0x1000  /* path has a tgt port group assigned */
#define PATH_FAIL_BACK_EVT       0x2000  /* path has failback outstanding */
#define PATH_FAIL_OVER_EVT       0x4000  /* path has failover outstanding */
#define PATH_SLAVE_DESTROY       0x8000  /* path has been slave destroy'ed */
#define PATH_ALWAYS_EXCLUDED     0x10000 /* array type is not supported */
#define PATH_HBEAT_TMO           0x20000 /* hbeat command just timed out */

/* modprobe configuration path flags define */
/* Flags 0-0xff are viewed as traffic ratios */
#define MPL_CFG_PATH_PRIMARY	0x100	 /* use as primary path */
#define MPL_CFG_PATH_EXPORT	0x200    /* put path on export list */


/* used in the scsi_cmd's "owner" field to signal a hbeat inquirysn */
#define HBEAT_SIG               0xAB

/* HeartBeat defines */
#define MPL_DFT_HBEAT_INTERVAL		30
#define MPL_DFT_HBEATFAIL_INTERVAL	1
#define MPL_DFT_HBEAT_RETRY		20
#define MPL_MAX_HBEAT_INTERVAL		256
#define MPL_MAX_HBEATFAIL_INTERVAL	256
#define MPL_MAX_HBEAT_RETRY		256

struct config_device {
	struct config_device *next;
	struct config_path *path;
	size_t    serial_size;	      /* actual bytes in serial number */
	uint8_t   serial_number[256];
	uint16_t  flags;
	uint16_t  found;
};

struct config_path {
	struct config_path *next;
	struct config_device *device;
	struct mpl_path *path;
	uint8_t  hba_wwpn[8];
	uint8_t  target_wwpn[8];
	int      lun_no;
	uint16_t flags;
	uint16_t found;
};

struct mpl_plugin
{
	char *name;
	int  (*probe)(struct scsi_cmnd *, struct mpl_path *);
	int  (*pre_queue_filter)(struct mpl_path *,
				 struct scsi_cmnd *,
				 void (*done)(struct scsi_cmnd *));
	int  (*diagnose_results)(struct io_stub *, int);
	void (*path_down)(struct mpl_path *);
	int  (*path_up)(struct mpl_path *);
	int  (*inq_validation)(struct mpl_path *, uint8_t *, uint8_t *, int *);
};

/* diagnose plugin flag */
#define PLUG_INTERNAL_CMD   1
#define PLUG_HBEAT_CMD      2
#define PLUG_SCSI_CMD       3

/* misc scsi defines */
#define SCSI_STAT_CHECK_COND 2
#undef  REPORT_LUNS
#define REPORT_LUNS 0xa0
#undef  REPORT_TGT_PORT_GROUP
#define REPORT_TGT_PORT_GROUP 0xa3
#undef  SET_TGT_PORT_GROUP
#define SET_TGT_PORT_GROUP 0xa4

#ifndef SAM_STAT_BUSY
#define SAM_STAT_BUSY 0x8
#endif
#ifndef SAM_STAT_TASK_SET_FULL
#define SAM_STAT_TASK_SET_FULL 0x28
#endif

/* Plugin module definition.  This generates a table that the plugin
 * code uses to locate available plugin modules to poll (using probe
 * routine address
 */
#define DEFINE_PLUGIN(id, name, probe, prequeue, diagnose, down, up, valid) \
struct mpl_plugin plugin_def_##id                               	\
__attribute__((section("plugin.table"), aligned(sizeof(void *)))) =	\
{                                                               	\
	name, probe, prequeue, diagnose, down, up, valid		\
};


/* Extra sanity checks for stub handling */
// #define MPL_STUB_SANITY 1

/* trim_to: similar to memcpy, but null terminates and deletes
 * trailing space */
static inline char *trim_to(char *dest, const char *src, size_t size)
{
	int n;

	memcpy(dest, src, size - 1);
	dest[size - 1] = 0;
	n = strlen(dest);
	while (n > 0 && dest[--n] == ' ')
		dest[n] = 0;
	return dest;
}


static inline struct mpl_path *
mpl_first_path(struct mpl_device *device)
{
	return  list_empty(&device->paths)
		? NULL
		: list_entry(list_first(&device->paths), struct mpl_path,
			     path_list);
}

static inline int
mpl_set_device_configured(struct mpl_device *device,
			  enum mpl_config_state configured)
{
	device->configured = configured;
	return 0;
}

static inline int
mpl_set_device_treatment(struct  mpl_device *device, unsigned int treatment)
{
	device->path_cntl_type = treatment;
	return 0;
}

static inline int
mpl_is_auto_failback(struct mpl_device *device)
{
	return (device->path_cntl_type == MPL_PC_TYPE_FAILOVER &&
		device->auto_failback);
}
static inline int
mpl_is_optimized(struct mpl_device *device)
{
	return (device->path_cntl_type != MPL_PC_TYPE_FAILOVER &&
		device->optimized_only);
}

static inline int
mpl_set_path_ratio(struct mpl_path *path, int ratio)
{
	path->traffic_ratio = ratio;
	return 0;
}

static inline void
mpl_set_device_static_config(struct mpl_device *device)
{
	device->StaticConfig = 1;
}

static inline void
mpl_set_path_static_config(struct mpl_path *path)
{
	path->StaticConfig = 1;
}

static inline int 
mpl_is_device_configured(struct mpl_device *device)
{
	return device->configured == CONFIG_CONFIGURED;
}

static inline int 
mpl_is_path_configured(struct mpl_path *path)
{
	return path->configured == CONFIG_CONFIGURED;
}

static inline int 
mpl_is_path_unconfigured(struct mpl_path *path)
{
	return path->configured == CONFIG_UNCONFIGURED;
}

static inline int
mpl_is_path_primary(struct mpl_path *path)
{
	return (path->is_default && (path->configured == CONFIG_CONFIGURED));
}

static int inline
mpl_is_queue_full(int result)
{
	return (status_byte(result) == QUEUE_FULL);
}

static int inline
mpl_is_queue_busy(int result)
{
	return ((status_byte(result) == BUSY) ||
		(host_byte(result) == DID_BUS_BUSY));
}

static int inline
mpl_is_power_on_reset(int result, uint8_t *sense)
{
	return (status_byte(result) == CHECK_CONDITION &&
 		((sense[2] & 0xF) == 0x6) && 
		sense[12] == 0x29 &&
		((sense[13] == 0x0) || (sense[13] == 0x1)));
}

static int inline
mpl_is_sas_changed(int result, uint8_t *sense)
{
	return (status_byte(result) == CHECK_CONDITION &&
 		((sense[2] & 0xF) == 0x6) && 
		sense[12] == 0x2a &&
		(sense[13] == 0x6 || sense[13] == 0x7));
}

static int inline
mpl_is_lu_not_ready(int result, uint8_t *sense)
{
	return (status_byte(result) == CHECK_CONDITION &&
 		((sense[2] & 0xF) == 0x2) && 
		(sense[12] == 0x4) &&
		(sense[13] < 0x2));
}

static inline int
mpl_is_inquiry_page_83(struct scsi_cmnd *cmd)
{
	return cmd->cmnd[0] == INQUIRY && cmd->cmnd[2] == 0x83;
}

static inline int
mpl_is_inquiry_page_0(struct scsi_cmnd *cmd)
{
	return cmd->cmnd[0] == INQUIRY && cmd->cmnd[2] == 0;
}

static inline unsigned
mpl_xtobin(unsigned char c)
{
	return isdigit(c) ? c - '0' : tolower(c) - 'a' + 10;
}

static inline void
mpl_crack_wwn(unsigned char wwn[8], const char *cp, const char **wwn_buffer)
{
	int  i;

	memset(wwn, 0, 8);
	for (i = 0; i < 8; ++i) {
		while (isspace(*cp) || *cp == ':') {
			++cp;
		}
		if (!isxdigit(*cp)) {
			break;
		}
		wwn[i] = mpl_xtobin(*cp);
		++cp;
		if (isxdigit(*cp)) {
			wwn[i] = (wwn[i] << 4) + mpl_xtobin(*cp);
			++cp;
		}
	}
	if (wwn_buffer) {
		*wwn_buffer = cp;
	}
}

static inline struct Scsi_Host *
mpl_scsi_host_from_cmd(struct scsi_cmnd *cmd)
{
	return cmd->device->host;
}

/* Helper routine mpl_host_no_from_cmd() defined as a macro to avoid
 * needing to include scsi/host.h in all plugin routines */
#define mpl_host_no_from_cmd(cmd)  \
	mpl_scsi_host_from_cmd(cmd)->host_no

static inline int
mpl_target_no_from_cmd(struct scsi_cmnd *cmd)
{
	return cmd->device->id;
}

static inline int
mpl_lun_no_from_cmd(struct scsi_cmnd *cmd)
{
	return cmd->device->lun;
}

extern int  mpl_create_proc(void);
extern void mpl_remove_proc(void);
extern struct mpl_host *mpl_find_host(int n);
struct mpl_path *mpl_find_path(struct mpl_host *host, int target, int lun);

extern struct mpl_plugin plugin_default;

extern struct mpl_host   *mpl_find_or_create_host(int);
extern struct mpl_host   *mpl_find_host_by_scsi_host(struct Scsi_Host *);
extern struct mpl_device *mpl_find_or_create_device(struct mpl_host *,
						    int channel,
						    int target, int lun);
extern struct mpl_path   *mpl_add_path(struct mpl_device *, struct mpl_host *,
				       int, int, int);
extern struct mpl_path   *mpl_find_or_create_path(struct mpl_host *, 
						  int, int);
extern struct mpl_path *mpl_get_primary_path(struct mpl_device *device);
extern struct mpl_path *mpl_device_pick_primary_path(struct mpl_device *device);
extern int  mpl_set_device_auto_failback(struct mpl_device *device, int boolean);
extern int  mpl_set_device_optimized(struct mpl_device *device, int boolean);
extern int  mpl_set_path_primary(struct mpl_path *path, int boolean);
extern int  mpl_set_path_configured(struct mpl_path *path,
						     enum mpl_config_state configured);

extern int  mpl_is_path_msa1000_secondary(struct mpl_path *path);
extern int  mpl_set_path_ratio(struct mpl_path *, int);
extern void mpl_set_dev_ctrl_header(struct mpl_device *, int);
extern void mpl_add_stub_to_lpfcq(struct io_stub *);
extern int  mpl_is_presentation_path(struct mpl_path *);

extern int  mpl_set_path_online(struct mpl_path *path);
extern int  mpl_set_path_offline(struct mpl_path *path);
extern int  mpl_clear_path_statistics(struct mpl_path *path);
extern void mpl_activate_path(struct mpl_path *path);
extern int  mpl_do_base_queuecommand(struct mpl_device *device,
				     struct mpl_path *path,
				     struct io_stub *stub);

extern int  mpl_queuecommand_alternate_path(struct mpl_device *device,
					    struct mpl_path *path,
					    struct io_stub *stub);
extern void mpl_add_deferral(struct io_stub *stub, enum deferral_reason reason);
extern void mpl_add_eh_deferral(struct io_stub *stub, enum deferral_reason reason);
extern int  mpl_quiesce_device(struct mpl_device *device, int tmo);
extern void mpl_free_stub(struct io_stub *stub);

extern kmem_cache_t *io_stub_cache;
