/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_ds400.c 902 2005-08-19 22:13:46Z jinfante $
 */

#include <linux/version.h>
#include <linux/pci.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>

#include <linux/string.h>

#include "mpl.h"
#include "mpl_crtn.h"
#include "mpl_log.h"

static uint8_t ds400_str[] = "Adaptec FC";
extern struct scsi_id_search_values mpl_id_search_list[12];

static int
mpl_ds400_probe(struct scsi_cmnd *cmd, struct mpl_path *path)
{
	uint8_t vendor[8 + 1];
	uint8_t product[16 + 1];
	uint8_t rev[4 + 1];
	uint8_t *request;
	uint8_t *prod;
	int len = 0;

	request = mpl_find_cmnd_data(cmd, &len);
	if ((request == NULL) || (len < 36)) {
		mpl_print(0, MPL_ERR "No data in DS400 probe len %d: %s %s\n",
      			len, mpl_print_device(path->device),
      			mpl_print_path(path))
		return 0;
	}
	prod = request + 16;

	if ((memcmp(prod, ds400_str, sizeof(ds400_str) - 1) == 0)) { 
		trim_to(vendor, request + 8, sizeof(vendor));
		trim_to(product, request + 16, sizeof(product));
		trim_to(rev, request + 32, sizeof(rev));
		mpl_release_cmnd_data(cmd);
		mpl_print(MPL_INIT, MPL_INFO " DS400: plugin support for: "
		  	"tgt %d lun %d: %s, %s, %s\n",
		  	cmd->device->id, cmd->device->lun,
			vendor, product, rev);
		return 1;
	} else {
		mpl_release_cmnd_data(cmd);
		return 0;
	}
}

static int
mpl_ds400_pre_queue_filter(struct mpl_path *path, struct scsi_cmnd *cmd,
			     void (*done)(struct scsi_cmnd *))
{
	if (cmd->cmnd[0] == SET_TGT_PORT_GROUP) {
		memset(&cmd->sense_buffer[0], 0, SCSI_SENSE_BUFFERSIZE);
		cmd->result = 0;
		done(cmd);
		return -1;
	}
	return 0;
}

static void
mpl_set_tgt_port_done(struct io_stub *stub)
{
	struct mpl_path *path = stub->path, *p;
	struct mpl_device *device = path->device, *dev;
	struct scsi_cmnd *cmd = &stub->cmd;
	struct mpl_host *host  = path->device->host;

	if (cmd->result == ScsiResult(DID_NO_CONNECT, 0)) {
		list_for_each_entry(dev, &host->devices, devices) {
			list_for_each_entry(p, &dev->paths, path_list) {
				if (p->host->host_no == path->host->host_no &&
					p->tgt_port_group == path->tgt_port_group &&
					p->target == path->target) {
					mpl_path_state_machine(dev->last_path, PATH_DOWN);
					dev->pending_lun_migration = 0;
					break;
				}
			}
		}
	} else if (cmd->result == 0 ||
		   cmd->result == ScsiResult(DID_ERROR, 0)) {
		/* Failing_path and path may be on the same or on opposite
		 * controllers. Do same extra checking to determine how to
		 * set active_controller.
		 */

		mpl_print(MPL_RECOVERY, MPL_INFO "Cmpl Start Unit status x%x: %s %s act %d\n",
			cmd->result,
		  	mpl_print_device(device),
		  	mpl_print_path(path),
			path->active_controller);

		if(device->failing_path && (path->active_controller !=
			device->failing_path->active_controller)) {
			mpl_path_flip_all_active(path);
		}
		else {
			if(!path->active_controller) {
				mpl_path_flip_all_active(path);
			}
		}
		list_for_each_entry(dev, &host->devices, devices) {
			list_for_each_entry(p, &dev->paths, path_list) {
				if (p->host->host_no == path->host->host_no &&
					p->target == path->target &&
					p->tgt_port_group == path->tgt_port_group) {
					mpl_path_state_machine(dev->last_path, PATH_UP);
					dev->pending_lun_migration = 0;
					break;
				}
			}
		}
	}
	else {
		mpl_print(0,
			  MPL_ERR "Unknown Set Tgt Port status x%x: %s %s\n",
			cmd->result,
		  	mpl_print_device(device),
		  	mpl_print_path(path));

		/* Attempt to cleanup devices associated with the command */
		list_for_each_entry(dev, &host->devices, devices) {
			list_for_each_entry(p, &dev->paths, path_list) {
				if (p->host->host_no == path->host->host_no &&
					p->target == path->target &&
					p->tgt_port_group == path->tgt_port_group) {
					mpl_path_state_machine(dev->last_path, PATH_DOWN);
					dev->pending_lun_migration = 0;
					break;
				}
			}
		}
	}
	return;
}


static void
mpl_ds400_path_down(struct mpl_path *path)
{
	mpl_print(MPL_RECOVERY,
		  MPL_INFO "Path Down Plugin %s %s: %s\n",
		  mpl_print_device(path->device),
		  mpl_print_path(path),
		  (path->active_controller ? "active_cntrlr" : "inactive_cntrlr"));
}

static int
mpl_ds400_path_up(struct mpl_path *path)
{
	struct mpl_device *device = path->device, *dev;
	uint32_t start_lun_migration = 1, cnt = 0;
	uint32_t non_active_ctlr_up;
	struct mpl_host *host = path->device->host;
	struct mpl_path *p, *flip_p;
	struct mpl_device *flip_device;

	p = path;
	list_for_each_entry(p, &device->paths, path_list) {
		cnt++;
	}

	mpl_print(MPL_RECOVERY,
		  MPL_INFO "Path Up Plugin %s %s: %s cnt %d lpath %d\n",
		  mpl_print_device(device),
		  mpl_print_path(path),
		  (path->active_controller ? "active_cntrlr" : "inactive_cntrlr"),
		  cnt, (path == device->last_path));

	if (device->scsi_type) {
		path->active_controller = 0;
		return 1;
	}

	if ((path->active_controller) ||
	    (path != device->last_path)) {
		return 1;
	}

	/* if lun migration started do nothing */
	if (path->device->pending_lun_migration == 2)
		return 0;


	/* 
	   check if all the luns on this target has atleast one good path
	   to the non-active controller. If there exist such a device, 
	   start the heart beat for this device and delay the lun 
	   migration until a path becomes online to this controller.
	*/
	list_for_each_entry(dev, &host->devices, devices) {
		list_for_each_entry(p, &dev->paths, path_list) {
			if (p->host->host_no == path->host->host_no &&
			    p->target == path->target &&
			    p->tgt_port_group == path->tgt_port_group &&
			    p->state != PATH_DISABLED) {

				/* 
				   Check if this device has atleast one path to a non_active
				   controller 
				*/
				non_active_ctlr_up = 0;
				flip_device = p->device;
				/* Lun migration waiting for all luns to have atleast one 
				   path available to the new active controller. */
				flip_device->pending_lun_migration = 1;
				list_for_each_entry(flip_p, &flip_device->paths, path_list) {
					if ((!flip_p->active_controller) &&
					    ((flip_p->state == PATH_ONLINE) ||
					     (flip_p->state == PATH_STANDBY)))
						non_active_ctlr_up = 1;

					/* If this is a failed path to inactive controller
					   do heart beat
					*/
					if ((!flip_p->active_controller) &&
					    (flip_p->state == PATH_FAILED) &&
					    (flip_p->hbstate == HB_WAITING))
						mpl_hbeat_state_machine(flip_p, HB_START);
				}
				
				/* There is a device with no good path to the non_active controller */
				if (!non_active_ctlr_up) {
					start_lun_migration = 0;
				}
			}
		}
	}

	if ( start_lun_migration ) {

		list_for_each_entry(dev, &host->devices, devices) {
			list_for_each_entry(p, &dev->paths, path_list) {
				if (p->host->host_no == path->host->host_no &&
				    p->target == path->target &&
			    	    p->tgt_port_group == path->tgt_port_group &&
				    p->state != PATH_DISABLED) {
					/* Lun migration command sent to target */
					p->device->pending_lun_migration = 2;
					p->device->state = DEV_FAILING;
				}
			}
		}

		mpl_print(MPL_RECOVERY, MPL_INFO "Issue Set Tgt Port Group: %s %s\n",
		  	mpl_print_device(device),
		  	mpl_print_path(path));

		path->flag |= (PATH_NEEDS_SET_TGT_PORT | PATH_DEV_QUIESCE);
		path->retries = 1;
		mpl_restart_sched_cmd(path, jiffies + 1);
	}

	return 0;
}

static inline int
mpl_initializing_command_required(uint8_t *sense)
{
	uint8_t sk, asc, ascq;

	sk = sense[2] & 0xF;
	asc = sense[12];
	ascq = sense[13];

	return (sk == 0x02 && asc == 0x04 && (ascq == 0xb || ascq == 0xa)); 
}

static int
mpl_ds400_diagnose(struct io_stub *stub, int flag)
{
	struct scsi_cmnd *cmd = &stub->cmd;
	struct mpl_path *path = stub->path;
	int    cmd_success, need_init_command;
	int    rv = 0;
	uint8_t sk;

	sk = cmd->sense_buffer[2] & 0xF;
	cmd_success = cmd->result == 0 && sk == 0;
	need_init_command = mpl_initializing_command_required(cmd->sense_buffer);
	/* Assume 80% utilization of max (512) */
	mpl_print(MPL_RECOVERY, MPL_INFO "Diagnose Plugin "
		"%s %s%scmd %x result %x sk %x ini %d\n",
		mpl_print_device(path->device), mpl_print_path(path),
		(stub->done) ? " " : " internal ",
		(unsigned int)cmd->cmnd[0], cmd->result, sk, need_init_command);

	/* Setup alua / active-passive on first command */
	path->device->active_passive = 1;
	path->device->alua_support = 1; /* supports ALUA mode */

	switch (flag) {
	case PLUG_INTERNAL_CMD:
		/* mpl driver ignores the return value for this case */
		switch (cmd->cmnd[0]) {
		case SET_TGT_PORT_GROUP:
			/* Only for internally generated command */
	    		if(mpl_is_power_on_reset(cmd->result,
			   cmd->sense_buffer) ||
	    		   mpl_is_lu_not_ready(cmd->result,
			   cmd->sense_buffer)) {
				path->flag |= (PATH_NEEDS_SET_TGT_PORT | PATH_DEV_QUIESCE);
				path->retries++;
				mpl_restart_sched_cmd(path, jiffies+1);
			}
			else {
				mpl_set_tgt_port_done(stub);
			}
			break;
		default:
			break;
		}
		break;

	case PLUG_HBEAT_CMD:
		/* mpl driver assumes a return value of 0 means the hbeat
		 * command succeeded.
		 */
		switch (cmd->cmnd[0]) {
		case TEST_UNIT_READY:
			if (path->lun == 0) {
				/*  LUN 0 on MSA1000 is never available */
				path->active_controller = 0;
				break;
			}
			if (cmd_success) {
				mpl_path_set_active(path, 1);
			} else if (need_init_command) {
				mpl_path_set_active(path, 0);
			}
			else {
				rv = 1;  /* error */
			}
			break;
		case INQUIRY:
			/* Setup tgt qdepth after first INQUIRY */
			path->device->tgt_queue_depth = 409; 

			if (!cmd_success) {
				rv = 1;  /* error */
			} 
			break;
		}
		break;

	case PLUG_SCSI_CMD:
		/* mpl driver assumes a return value of 1 means to initiate
		 * a failover and retry the command.
		 */
		if (cmd->cmnd[0] == TEST_UNIT_READY) {
			if (cmd_success) {
				mpl_path_set_active(path, 1);
			} else if (need_init_command) {
				mpl_path_set_active(path, 0);

				/* To the upper SCSI layer, this command
				 * really succeeded
				 */
				memset(&cmd->sense_buffer[0], 0,
					SCSI_SENSE_BUFFERSIZE);
				cmd->result = 0;
			}
		} 
		if (cmd->result == 2 && sk == 0x02 &&
			cmd->sense_buffer[12] == 0x04) {
			if (path->device->pending_lun_migration)
				rv = 2;  /* retry without failover */
			else
				rv = 1;  /* retry with failover */
		}
		break;
	}
	return rv;
}

/* Return 0 if we find a valid validation id, else return 1 */
int mpl_ds400_validation_id(
	struct mpl_path *path,
	uint8_t *page_83,
	uint8_t *serial, int *serial_size)
{
	uint32_t offset = 0;
	uint32_t size = 0, relport;
	uint16_t total_size, id_size;
	uint8_t *idptr, type, tpcnt, old_sas;
	uint16_t tpg;
	uint32_t *idwptr;

	total_size = *((uint16_t *) &page_83[2]);
	total_size = be16_to_cpu(total_size);
	offset = 4;

	path->flag &= ~PATH_ASSIGN_TPG;
	while (offset < total_size) {
		idptr = page_83 + offset;
		id_size = (uint16_t) *(idptr+3);
		offset += (id_size + 4);
		
		idwptr = (uint32_t *)(idptr + 4);

		type = *(idptr+1);
		if (((type & 0xf) == 5) && (id_size == sizeof(uint32_t))) {
			size = *idwptr;
			path->tgt_port_group = be32_to_cpu(size);
			path->flag |= PATH_ASSIGN_TPG;
			mpl_print(MPL_ALUA, MPL_INFO "Tgt Port Group: x%x"
				" %s %s\n",
				path->tgt_port_group,
				mpl_print_device(path->device),
				mpl_print_path(path));
		}
		if (((type & 0xf) == 4) && (id_size == sizeof(uint32_t))) {
			size = *idwptr;
			path->relative_tgt_port = be32_to_cpu(size);
			mpl_print(MPL_ALUA, MPL_INFO "Relative Tgt Port: x%x"
				" %s %s\n",
				path->relative_tgt_port,
				mpl_print_device(path->device),
				mpl_print_path(path));
		}

		if (id_size == 0)
			break;
	}

	old_sas = path->sas;
	if (path->tgt_port_info) {
		idwptr = (uint32_t *)path->tgt_port_info;
		size = *idwptr;
		size = be32_to_cpu(size);
		offset = 0;
		idptr = (uint8_t *)(idwptr + 1);
		while (offset < size) {
			type = *(idptr + offset);
			offset += 2;
			tpg = *((uint16_t *)(idptr + offset));
			tpg = be16_to_cpu(tpg);
			offset += 5;
			tpcnt = *(idptr + offset);
			offset++;
			/* For our tpg, set sas if paths
			 * asymmetric access state is optimised
			 */
			if (tpg == path->tgt_port_group) {
				path->sas = (type & MPL_ALUA_SAS_MASK);
				mpl_print(MPL_ALUA, MPL_INFO "Asymetric Access State "
		  		" %s %s: %d\n",
		  		mpl_print_device(path->device), mpl_print_path(path),
		  		path->sas);
			}
			idwptr = (uint32_t *)(idptr + offset);
			while (tpcnt--) {
				relport = *idwptr++;
				relport = be32_to_cpu(relport);
				offset += 4;
			}
		}
	}

	if (old_sas != path->sas) {
		if (path->sas == MPL_ALUA_ACTIVE_OPT) {
			atomic_inc(&path->device->opt_path_cnt);
		}
		else {
			if ((old_sas == MPL_ALUA_ACTIVE_OPT) &&
			    (atomic_read(&path->device->opt_path_cnt))) {
				atomic_dec(&path->device->opt_path_cnt);
			}
		}
		mpl_print(MPL_ALUA, MPL_INFO "Asymetric Access State Change"
		  " %s %s: tpg %d From %d to %d optcnt %d\n",
		  mpl_print_device(path->device), mpl_print_path(path),
		  path->tgt_port_group, old_sas, path->sas,
		  atomic_read(&path->device->opt_path_cnt));
	}

	return mpl_get_scsi_page83_serial(page_83, mpl_id_search_list,
		sizeof(mpl_id_search_list)/sizeof(mpl_id_search_list[0]),
		serial, serial_size);
}

/* DEFINE_PLUGIN(id, name, probe, prequeue, diagnose, down, up, valid)
 */

DEFINE_PLUGIN(ds400, "ds400",
	      mpl_ds400_probe,
	      mpl_ds400_pre_queue_filter,
	      mpl_ds400_diagnose, 
	      mpl_ds400_path_down,
	      mpl_ds400_path_up,
	      mpl_ds400_validation_id);
