/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_eva.c 1261 2006-11-03 19:41:13Z jinfante $
 */


#include <linux/version.h>
#include <linux/pci.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>

#include <linux/string.h>

#include "mpl.h"
#include "mpl_crtn.h"
#include "mpl_log.h"

static uint8_t eva5000_str[] = "HSV110";
static uint8_t eva3000_str[] = "HSV100";

extern int mpl_enable_nopath;

int mpl_exclude_eva = 1;

static int
mpl_eva_probe(struct scsi_cmnd *cmd, struct mpl_path *path)
{
	uint8_t vendor[8 + 1];
	uint8_t product[16 + 1];
	uint8_t rev[4 + 1];
	uint8_t *request;
	uint8_t *prod;
	int len = 0;

	request = mpl_find_cmnd_data(cmd, &len);
	if ((request == NULL) || (len < 36)) {
		mpl_print(0, MPL_ERR "No data in EVA probe len %d: %s %s\n",
      			len, mpl_print_device(path->device),
      			mpl_print_path(path))
		return 0;
	}
	prod = request + 16;

	if (!mpl_vendor_hp_check(cmd)) {
		mpl_release_cmnd_data(cmd);
		return 0;
	}

	if ((memcmp(prod, eva5000_str, sizeof(eva5000_str)-1) == 0) || 
	    (memcmp(prod, eva3000_str, sizeof(eva3000_str)-1) == 0)) { 
		trim_to(vendor, request + 8, sizeof(vendor));
		trim_to(product, request + 16, sizeof(product));
		trim_to(rev, request + 32, sizeof(rev));
		mpl_release_cmnd_data(cmd);
		if (mpl_exclude_eva) {
			mpl_set_path_configured(path, CONFIG_UNAVAILABLE);
			path->flag |= PATH_ALWAYS_EXCLUDED;
			mpl_print(MPL_INIT, MPL_INFO "EVA: not supported for: "
		  		"tgt %d lun %d: %s, %s, %s\n",
		  		cmd->device->id, cmd->device->lun,
				vendor, product, rev);
		}
		else {
			mpl_print(MPL_INIT, MPL_INFO "EVA: plugin support for: "
		  		"tgt %d lun %d: %s, %s, %s\n",
		  		cmd->device->id, cmd->device->lun,
				vendor, product, rev);
		}

		return 1;
	} else {
		mpl_release_cmnd_data(cmd);
		return 0;
	}
}

/* pre-queue filter returns zero to allow the command to be queued to
 * base driver, non-zero to indicate that the done has been called
 * and that command processing is complete.
 */

static int
mpl_eva_pre_queue_filter(struct mpl_path *path, struct scsi_cmnd *cmd,
			     void (*done)(struct scsi_cmnd *))
{
	if (!mpl_exclude_eva &&
		(cmd->cmnd[0] == START_STOP) && !path->active_controller) {
		memset(&cmd->sense_buffer[0], 0, SCSI_SENSE_BUFFERSIZE);
		cmd->result = 0;
		done(cmd);
		return -1;
	}
	return 0;
}

static void
mpl_start_stop_done(struct io_stub *stub)
{
	struct mpl_path *path = stub->path;
	struct mpl_device *device = path->device;
	struct scsi_cmnd *cmd = &stub->cmd;

	if (cmd->result == ScsiResult(DID_NO_CONNECT, 0)) {
		mpl_path_state_machine(path, PATH_DOWN);
	} else if (cmd->result == 0) {
		/* Failing_path and path may be on the same or on opposite
		 * controllers. Do same extra checking to determine how to
		 * set active_controller.
		 */
		path->active_controller = 1;
		mpl_path_state_machine(path, PATH_UP);
	}
	else {
		mpl_print(0,
		  	MPL_ERR "Unknown Start Unit status x%x: %s %s\n",
			cmd->result,
		  	mpl_print_device(device),
		  	mpl_print_path(path));
	}
	return;
}


static void
mpl_eva_path_down(struct mpl_path *path)
{
	mpl_print(MPL_RECOVERY,
		  MPL_INFO "Path Down Plugin %s %s: %s\n",
		  mpl_print_device(path->device),
		  mpl_print_path(path),
		  (path->active_controller ? "active_cntrlr" : "inactive_cntrlr"));

	if (mpl_exclude_eva)
		mpl_set_path_configured(path, CONFIG_UNAVAILABLE);
}

static int
mpl_eva_path_up(struct mpl_path *path)
{
	struct mpl_device *device = path->device;

	mpl_print(MPL_RECOVERY,
		  MPL_INFO "Path Up Plugin %s %s: %s lpath %d\n",
		  mpl_print_device(device),
		  mpl_print_path(path),
		  (path->active_controller ? "active_cntrlr" : "inactive_cntrlr"),
		  (path == device->last_path));

	if (mpl_exclude_eva) {
		mpl_set_path_configured(path, CONFIG_UNAVAILABLE);
		return 0;
	}

	if (device->scsi_type) {
		path->active_controller = 0;
		return 1;
	}

	if ((path->active_controller) ||
		 (path != device->last_path)) {
		return 1;
	}

	device->state = DEV_FAILING;
	path->flag |= (PATH_NEEDS_STARTSTOP | PATH_DEV_QUIESCE |
		    PATH_OUT_CHG_LUN_OWN_CMD);
	path->retries = 1;
	mpl_restart_sched_cmd(path, jiffies + 1);
	return 0;
}

static inline int
mpl_initializing_command_required(uint8_t *sense)
{
	uint8_t sk, asc, ascq;

	sk = sense[2] & 0xF;
	asc = sense[12];
	ascq = sense[13];

	return sk == 0x02 && asc == 0x04 && ascq == 0x02;
}

static inline int
mpl_eval_sense(uint8_t *sense)
{
	uint8_t sk, asc, ascq;

	sk = sense[2] & 0xF;
	asc = sense[12];
	ascq = sense[13];
				/* return non zero for:
				 * sk/asc/ascq    description
				 * -----------    -----------
				 * 02/04/00     not ready, cause not reportable
				 * 02/04/02     not ready, initializing cmd req
				 * 02/04/03     not ready, manual intervtn req
				 * 06/0a/08     unit attention, ???
				 * 06/04/02     unit attention, ???
				 */
	return ((sk == 0x02 && asc == 0x04 && 
		 (ascq == 0x00 || ascq == 0x02 || ascq == 0x03)) ||
		(sk == 0x06 && ((asc == 0x0A && ascq == 0x08) ||
				(asc == 0x04 && ascq == 0x02))));
}

static void
mpl_eva_set_active(struct mpl_path *path, int boolean)
{
	if (path->device->scsi_type) {
		path->active_controller = 0;
		return;
	}

	path->active_controller = boolean;
	return;
}


static int
mpl_eva_diagnose(struct io_stub *stub, int flag)
{
	struct scsi_cmnd *cmd = &stub->cmd;
	struct mpl_path *path = stub->path;
	struct mpl_device *device = path->device;
	int    cmd_success, need_init_command;
	int    rv = 0;
	uint8_t sk;

	mpl_print(MPL_RECOVERY, MPL_INFO "Diagnose Plugin "
		"%s %s%scmd %x result %x\n",
		mpl_print_device(device), mpl_print_path(path),
		(stub->done) ? " " : " internal ",
		(unsigned int)cmd->cmnd[0], cmd->result);

	sk = cmd->sense_buffer[2] & 0xF;
	cmd_success = cmd->result == 0 && sk == 0;
	need_init_command = mpl_initializing_command_required(cmd->sense_buffer);
	/* Setup alua / active-passive on first command */
	device->active_passive = 1;

	switch (flag) {
	case PLUG_INTERNAL_CMD:
		if (mpl_exclude_eva) {
			mpl_set_path_configured(path, CONFIG_UNAVAILABLE);
			return 0;
		}
		/* mpl driver ignores the return value for this case */
		switch (cmd->cmnd[0]) {
		case START_STOP:
			if (path->flag & PATH_OUT_CHG_LUN_OWN_CMD) {
				/* Only for internally generated command */
	    			if(mpl_is_power_on_reset(cmd->result,
				   cmd->sense_buffer) ||
	    			   mpl_is_lu_not_ready(cmd->result,
				   cmd->sense_buffer)) {
					path->flag |= (PATH_NEEDS_STARTSTOP |
					    PATH_DEV_QUIESCE |
					    PATH_OUT_CHG_LUN_OWN_CMD);
					path->retries++;
					mpl_restart_sched_cmd(path, jiffies+1);
				}
				else {
					path->flag &= ~PATH_OUT_CHG_LUN_OWN_CMD;
					mpl_start_stop_done(stub);
				}
			}
			break;
		default:
			break;
		}
		break;

	case PLUG_HBEAT_CMD:
		if (mpl_exclude_eva) {
			mpl_set_path_configured(path, CONFIG_UNAVAILABLE);
			return 0;
		}
		/* mpl driver assumes a return value of 0 means the hbeat
		 * command succeeded.
		 */
		switch (cmd->cmnd[0]) {
		case TEST_UNIT_READY:
			if (path->lun == 0) {
				/*  LUN 0 on MSA1000 is never available */
				path->active_controller = 0;
				break;
			}
			if (cmd_success) {
				mpl_eva_set_active(path, 1);
			} else if (need_init_command) {
				mpl_eva_set_active(path, 0);
			}
			else {
				rv = 1;  /* error */
			}
			break;

		case INQUIRY:
			/* Setup tgt qdepth after first INQUIRY */
			/* Assume 80% utilization of max (256) */
			device->tgt_queue_depth = 205; 
 			if (mpl_enable_nopath && !device->nopath_tmo) {
				device->nopath_tmo = 60;
			}
			if (!cmd_success) {
				rv = 1;  /* error */
			} 
			break;
		}
		break;

	case PLUG_SCSI_CMD:
		if (mpl_exclude_eva) {
			return 0;
		}
		/* mpl driver assumes a return value of 1 means to initiate
		 * a failover and retry the command.
		 */
		if (cmd->cmnd[0] == TEST_UNIT_READY) {
			if (cmd_success) {
				mpl_eva_set_active(path, 1);
			} else if (need_init_command) {
				mpl_eva_set_active(path, 0);

				/* To the upper SCSI layer, this command
				 * really succeeded
				 */
				memset(&cmd->sense_buffer[0], 0,
					SCSI_SENSE_BUFFERSIZE);
				cmd->result = 0;
			}
		} 
		if (cmd->result == 2 && sk == 0x02 &&
			cmd->sense_buffer[12] == 0x04) {
			rv = 1;  /* failover */
		}
		break;
	}
	return rv;
}

/* DEFINE_PLUGIN(id, name, probe, prequeue, diagnose, down, up, valid)
 */

DEFINE_PLUGIN(eva, "eva",
	      mpl_eva_probe,
	      mpl_eva_pre_queue_filter,
	      mpl_eva_diagnose, 
	      mpl_eva_path_down,
	      mpl_eva_path_up,
	      NULL);


