/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_mem.c 1263 2006-11-04 00:46:57Z jinfante $
 */

#include <linux/version.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>

#include <linux/ctype.h>
#include <linux/string.h>

#include "mpl.h"
#include "mpl_log.h"
#include "mpl_crtn.h"

extern atomic_t total_internal_count;
kmem_cache_t *io_stub_cache = NULL;
kmem_cache_t *data_buf_cache = NULL;

#define MAX_MEM_CNT      32
struct mpl_mem_pool {
	void *buf_array[MAX_MEM_CNT];	
	uint16_t   max_count;
	uint16_t   cur_count;
};

int mpl_data_cnt = 0;
int mpl_stub_cnt = 0;
int mpl_data_save_cmd = 0;
int mpl_data_int_cmd = 0;

struct mpl_mem_pool buf_list;
struct mpl_mem_pool stub_list;
/* lock for buf_list and stub_list */
static spinlock_t int_cmd_data_lock = SPIN_LOCK_UNLOCKED;

int
mpl_create_mem_pool(void)
{
	void *buf;
	int i;
	int rc = 0;

	io_stub_cache = kmem_cache_create("mpl_stub_io",
					  sizeof(struct io_stub), 0,
					  (SLAB_HWCACHE_ALIGN
#ifdef CONFIG_DEBUG_SLAB
					   | SLAB_POISON | SLAB_RED_ZONE
#endif
						  ),
					  NULL, NULL);
	if (!io_stub_cache)
		goto mem_alloc_error;
	mpl_stub_cnt = 0;

	data_buf_cache = kmem_cache_create("mpl_data_buf",
					  512, 0,
					  SLAB_HWCACHE_ALIGN,
					  NULL, NULL);

	if (!data_buf_cache)
		goto mem_alloc_error;
	mpl_data_cnt = 0;

	memset(&buf_list, 0, sizeof(buf_list));
	memset(&stub_list, 0, sizeof(stub_list));

	for (i = 0; i < MAX_MEM_CNT; i++) {
		buf = kmem_cache_alloc(data_buf_cache, SLAB_KERNEL);
		if (buf == NULL) {
			break;
		} else {
			mpl_data_cnt++;
			buf_list.buf_array[i] = buf;
			buf_list.max_count++;
			buf_list.cur_count++;
		}
	}

	for (i = 0; i < MAX_MEM_CNT; i++) {
		buf = kmem_cache_alloc(io_stub_cache, SLAB_KERNEL);
		if (buf == NULL) {
			break;
		} else {
			mpl_stub_cnt++;
			stub_list.buf_array[i] = buf;
			stub_list.max_count++;
			stub_list.cur_count++;
		}
	}

	if ((buf_list.cur_count == 0) ||
	    (stub_list.cur_count == 0)) {
mem_alloc_error:
		/* one of the list is empty so free all and return error */
		rc = -ENOMEM;
		mpl_destroy_mem_pool();
	}

	return rc;
}


void
mpl_destroy_mem_pool(void)
{
	unsigned long flags;
	int i;
	kmem_cache_t *tmp;

	spin_lock_irqsave(&int_cmd_data_lock, flags); 

	for (i = 0; i < buf_list.cur_count; i++) {
		mpl_data_cnt--;
		kmem_cache_free(data_buf_cache, buf_list.buf_array[i]);
	}
	buf_list.cur_count = 0;

	if (data_buf_cache) {
		tmp = data_buf_cache;
		data_buf_cache = NULL;
		if (mpl_data_cnt) {
			mpl_print(0, MPL_ERR
			"kmem_cache: data_buf:%d %d,%d destroy error\n",
			mpl_data_cnt, mpl_data_save_cmd, mpl_data_int_cmd);
		}
		spin_unlock_irqrestore(&int_cmd_data_lock, flags); 
		kmem_cache_destroy(tmp);
		spin_lock_irqsave(&int_cmd_data_lock, flags); 
	}

	for (i = 0; i < stub_list.cur_count; i++) {
		mpl_stub_cnt--;
		kmem_cache_free(io_stub_cache, stub_list.buf_array[i]);
		stub_list.buf_array[i] = 0;
	}
	stub_list.cur_count = 0;

	if (io_stub_cache) {
		tmp = io_stub_cache;
		io_stub_cache = NULL;
		if (mpl_stub_cnt) {
			mpl_print(0, MPL_ERR
			"kmem_cache: io_stub:%d destroy error\n", mpl_stub_cnt);
		}
		spin_unlock_irqrestore(&int_cmd_data_lock, flags); 
		kmem_cache_destroy(tmp);
		spin_lock_irqsave(&int_cmd_data_lock, flags); 
	}
	spin_unlock_irqrestore(&int_cmd_data_lock, flags); 
	return;
}


void *
mpl_get_scsi_buf(void)
{
	unsigned long flags;
	void *buf;	

	buf = kmem_cache_alloc(data_buf_cache, SLAB_ATOMIC);
	if (buf == NULL) {
		spin_lock_irqsave(&int_cmd_data_lock, flags);
		if (buf_list.cur_count) {
			buf_list.cur_count--;
			mpl_print(MPL_INIT, MPL_INFO "POOL: Getting buf at "
				  "index %x\n",
				  buf_list.cur_count);
			buf = buf_list.buf_array[buf_list.cur_count];
		}
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	}
	else {
		mpl_data_cnt++;
	}

	return buf;
}


void
mpl_put_scsi_buf(void *buf)
{
	unsigned long flags;

	spin_lock_irqsave(&int_cmd_data_lock, flags);
	if (buf_list.cur_count < buf_list.max_count) {
		mpl_print(MPL_INIT, MPL_INFO "POOL: Putting buf at index "
			  "%x\n",
			  buf_list.cur_count);
		buf_list.buf_array[buf_list.cur_count] = buf;
		buf_list.cur_count++;
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	} else {
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
		mpl_data_cnt--;
		kmem_cache_free(data_buf_cache, buf);
	}	
}



static struct io_stub *
mpl_get_stub(void)
{
	unsigned long flags;
	struct io_stub *stub;

	stub = kmem_cache_alloc(io_stub_cache, SLAB_ATOMIC);
	if (stub == NULL) {
		spin_lock_irqsave(&int_cmd_data_lock, flags);
		if(stub_list.cur_count) {
			stub_list.cur_count--;
			mpl_print(MPL_INIT, MPL_INFO "POOL: Getting stub at "
				  "index %x\n",
				  stub_list.cur_count);
			stub = stub_list.buf_array[stub_list.cur_count];
			stub_list.buf_array[stub_list.cur_count] = 0;
		}
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	}
	else {
		mpl_stub_cnt++;
	}

	return stub;
}

static void
mpl_put_stub(struct io_stub *stub)
{
	unsigned long flags;

	spin_lock_irqsave(&int_cmd_data_lock, flags);
	if (stub_list.cur_count < stub_list.max_count) {
		mpl_print(MPL_INIT, MPL_INFO "POOL: Putting stub at index "
			  "%x\n",
			  stub_list.cur_count);
		stub_list.buf_array[stub_list.cur_count] = stub;
		stub_list.cur_count++;
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
	} else {
		spin_unlock_irqrestore(&int_cmd_data_lock, flags);
		mpl_stub_cnt--;
		kmem_cache_free(io_stub_cache, stub);
	}	
}


extern  uint32_t mpl_max_can_queue;

void
mpl_inc_can_queue(struct io_stub *stub)
{
	struct mpl_path   *path = stub->path;
	struct mpl_device *device = path->device;
	struct Scsi_Host * sh;

	if (!(stub->flag & MPL_DEC_CAN_QUEUE))
		return;
	stub->flag &= ~MPL_DEC_CAN_QUEUE;

	sh = path->host->scsi_host;
	if (sh->can_queue < mpl_max_can_queue)
		sh->can_queue++;
	if (device->host != path->host) {
		sh = device->host->scsi_host;
		if (sh->can_queue < mpl_max_can_queue)
			sh->can_queue++;
	}
	return;
}

void
mpl_dec_can_queue(struct io_stub *stub)
{
	struct mpl_path   *path = stub->path;
	struct mpl_device *device = path->device;
	struct Scsi_Host * sh;

	if (stub->flag & MPL_DEC_CAN_QUEUE)
		return;
	stub->flag |= MPL_DEC_CAN_QUEUE;
	sh = path->host->scsi_host;
	if (sh->can_queue)
		sh->can_queue--;

	if (device->host != path->host) {
		sh = device->host->scsi_host;
		if (sh->can_queue)
			sh->can_queue--;
	}
	return;
}

struct io_stub *
mpl_stub_alloc(struct scsi_cmnd *cmd, struct mpl_path *path,
               void (*done)(struct scsi_cmnd *))
{
	struct io_stub *stub;
	static unsigned long serial_number;

	stub = mpl_get_stub(); 
	if (!stub) {
		return NULL;
	}
	memset(stub, 0, sizeof(*stub));
	wmb();
	atomic_inc(&path->device->ref_count);
	stub->magic = MPL_STUB_MAGIC;
	stub->cmdp = cmd;
	stub->done = done;
	stub->path = path;
	stub->state = STUB_ALLOC;
	if (cmd) {
		stub->sdev = cmd->device;
		/* Make a copy of the SCSI cmnd, this may be
		 * modified if we send the IO down an alternate
		 * path. The original cmnd NEVER gets modified.
		 */
		memcpy(&stub->cmd, cmd, sizeof(struct scsi_cmnd));
		stub->cmd.scsi_done = mpl_done;
		INIT_LIST_HEAD(&stub->cmd.list);
		INIT_LIST_HEAD(&stub->cmd.eh_entry);
		atomic_inc(&path->io_count);
		atomic_inc(&path->tgt->io_count);
		atomic_inc(&path->device->io_count);
		atomic_inc(&path->host->io_count);
		atomic_inc(&path->tot_count);
		atomic_inc(&path->tgt->tot_count);
		atomic_inc(&path->device->tot_count);
		atomic_inc(&path->host->tot_count);
	}
	else {
		/* For each internal command we send, thats one less SCSI 
		 * command we can send.
		 */
		mpl_dec_can_queue(stub);

		cmd = &stub->cmd;
		cmd->serial_number = ++serial_number;
		cmd->allowed = 5;
		init_timer(&cmd->eh_timeout);
		cmd->eh_timeout.function = mpl_cmd_timeout;
		cmd->eh_timeout.expires = jiffies + 3600 * HZ;
		cmd->eh_timeout.data = 0;
		cmd->scsi_done = mpl_done;

		atomic_inc(&stub->path->internal_count);
		atomic_inc(&total_internal_count);
	}
	return stub;
}

void
mpl_stub_free(struct io_stub *stub)
{
	struct mpl_path *path = stub->path;

	stub->state = STUB_FREE;

	if (stub->flag & MPL_INTERNAL_CMD_STUB) {
		if (stub->cmd.request_buffer) {
			mpl_data_int_cmd--;
			mpl_put_scsi_buf(stub->cmd.request_buffer);
		}
		atomic_dec(&path->internal_count);
		atomic_dec(&total_internal_count);
		mpl_inc_can_queue(stub);
		mpl_module_unlock();
		del_timer_sync(&stub->cmd.eh_timeout);
		mpl_module_lock();
#ifdef MPL_STUB_SANITY
		if (stub->done || stub->cmdp) {
			mpl_print(0, MPL_ERR "Bad Internal stub free: %p from  %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub, (&stub)[-1],
		       mpl_print_device(path->device),
		       mpl_print_path(path),
		       stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
#endif
	}
	else {
#ifdef MPL_STUB_SANITY
		if (!(stub->done) || !(stub->cmdp) ||
			!(stub->flag & MPL_SCSI_CMD_STUB)) {
			mpl_print(0, MPL_ERR "Bad SCSI stub free: %p from  %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub, (&stub)[-1],
		       mpl_print_device(path->device),
		       mpl_print_path(path),
		       stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
#endif
		atomic_dec(&path->io_count);
		atomic_dec(&path->tgt->io_count);
		atomic_dec(&path->device->io_count);
		atomic_dec(&path->host->io_count);
	}
	atomic_dec(&path->device->ref_count);
	mpl_put_stub(stub);
}

void
mpl_stub_switch_path(struct io_stub *stub, struct mpl_path *new_path)
{
	struct mpl_path *path = stub->path;

	if (path == new_path)  /* sanity check */
		return;

	if (stub->flag & MPL_INTERNAL_CMD_STUB) {
		atomic_dec(&path->internal_count);
		atomic_inc(&new_path->internal_count);
		if (path->device != new_path->device) {
			atomic_dec(&path->device->ref_count);
			atomic_inc(&new_path->device->ref_count);
		}
	}
	else {
		atomic_dec(&path->io_count);
		atomic_dec(&path->tgt->io_count);
		atomic_inc(&new_path->io_count);
		atomic_inc(&new_path->tgt->io_count);
		if (path->device != new_path->device) {
			atomic_dec(&path->device->io_count);
			atomic_inc(&new_path->device->io_count);
			atomic_dec(&path->device->ref_count);
			atomic_inc(&new_path->device->ref_count);
		}
		if (path->host != new_path->host) {
			atomic_dec(&path->host->io_count);
			atomic_inc(&new_path->host->io_count);
		}
	}
	stub->path = new_path;
	return;
}

