/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* MultiPulse for Linux /proc file support
 * $Id: mpl_proc.c 1313 2007-02-02 19:30:20Z jinfante $
 */

#include <linux/version.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/cdev.h>
#include <linux/seq_file.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/blkdev.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>

#include "mpl.h"
#include "mpl_version.h"
#include "mpl_log.h"
#include "mpl_crtn.h"

#define CONFIG_NODE    "config"
#define EXCLUSION_NODE "exclusion"
#define CONTROL_NODE   "control"
#define STATS_NODE     "stats"
#define STATS_TGT_NODE "stats_tgt"
#define DEBUG_NODE     "debug"
#define VERSION_NODE   "version"
#define PROC_ENTRY_FLAGS (S_IFREG | S_IRUGO | S_IWUSR | S_IWGRP)

#define NUM_ENTRIES(x) ((sizeof(x) / sizeof(x[0])))

#define seq_from_file(x)  ((struct seq_file *) file->private_data)
#define seq_private_from_file(x) (seq_from_file(file)->private)

extern struct list_head mpl_targets;
extern int mpl_hbeat_interval;
extern int mpl_hbeat_fail_interval;
extern int mpl_max_hbeat_retry;
extern int mpl_auto_configure;
extern int mpl_auto_lun_balance;
extern int mpl_enable_tgt_qdepth;
extern int mpl_handle_qfull;
extern atomic_t total_internal_count;
extern struct config_device *mpl_config_devices;
extern struct config_path *mpl_config_paths;
extern struct config_device *mpl_exclusion_device;
extern int mpl_num_exclusion_devices;

static struct proc_dir_entry *mpl_root;
static int mpl_cdev_first = 0;

/* Life would be simple if we could depend on the validity of
 * the iterator's host, device, and path pointers.
 * Unfortunately, we cannot.  We hold the mpl_module_lock during
 * config_seek, config_show and stats_show.  That ensures list
 * integrity while we fill the current seq_file buffer.  But
 * when that buffer fills, our iteration is interrupted as the
 * buffer is returned to user space.  So the next time
 * config_start is called, we need to reconstruct where we are
 * in the enumeration.  So we have to carry host, device, and
 * path numbers around to ensure that we can seek back into
 * the middle of the lists.
 */

struct config_iterator 
{
	enum { M_HOST, M_DEVICE, M_TARGET, M_PATH, M_CDEVICE, M_CPATH, M_MODULE }
		which;
	struct mpl_host   host;
	struct mpl_device device;
	struct mpl_target target;
	struct mpl_path   path;
	struct mpl_path   *path_address;
				/* In the interest of keeping all the
				 * formatted strings off the stack, we
				 * allocate space for them here,
				 * rather than as automatics. */
	struct config_device cdevice;
	struct config_path   cpath;
	uint8_t dev_type[12];
	uint8_t state_buffer[32];
	uint8_t hb_state_buffer[32];
	uint8_t serial_buffer[256];
	uint8_t vendor[8 + 1];
	uint8_t product[16 + 1];
	uint8_t rev[4 + 1];
	uint8_t wwn[8];
	uint8_t wwn_buf[32];
	uint8_t config_buf[32];
	uint8_t status[12];
	uint8_t module;
};

static struct config_iterator *
exclusion_seek(struct config_iterator *config, loff_t *offset)
{
	struct config_iterator *rv = NULL;
	struct config_path *cpath  = NULL;
	loff_t p = *offset;

	mpl_module_lock();

	for (cpath = mpl_config_paths; cpath; cpath = cpath->next) {
		if (cpath->flags & MPL_CFG_PATH_EXPORT) {
			if (!p--) {
				config->which = M_CPATH;
				rv = config;
				goto done;
			}
		}
	}

done:
	if (cpath) {
		config->cpath = *cpath;
	}
	mpl_module_unlock();
	return rv;
}

static void *
exclusion_start(struct seq_file *file, loff_t *offset)
{
	struct config_iterator *config;

	config = file->private;
	return exclusion_seek(config, offset);
}

static void *
exclusion_next(struct seq_file *file, void *arg, loff_t *offset)
{
	struct config_iterator *config = arg;

	++*offset;
	return exclusion_seek(config, offset);
}

static struct config_iterator *
tgt_seek(struct config_iterator *config, loff_t *offset)
{
	struct list_head *l, *n;
	struct mpl_path   *path = NULL;
	struct mpl_target *tgtp = NULL;
	struct config_iterator *rv = NULL;
	loff_t p = *offset;

	mpl_module_lock();

        list_for_each_safe(l, n, &mpl_targets) {
                tgtp = list_entry(l, struct mpl_target, target_list);
		if (!p--) {
			config->which = M_TARGET;
			rv = config;
			goto done;
		}
		list_for_each_entry(path, &tgtp->paths, tgt_path_list) {
			if (path->flag & PATH_ALWAYS_EXCLUDED)
				continue;
			if (!p--) {
				config->which = M_PATH;
				rv = config;
				goto done;
			}
		}
        }

done:
	if (tgtp) {
		config->target = *tgtp;
		if (path) {
			config->path = *path;
			config->path_address = path;
		}
	}
	mpl_module_unlock();
	return rv;
}

static void *
tgt_start(struct seq_file *file, loff_t *offset)
{
	struct config_iterator *config;

	config = file->private;
	return tgt_seek(config, offset);
}

static void *
tgt_next(struct seq_file *file, void *arg, loff_t *offset)
{
	struct config_iterator *config = arg;

	++*offset;
	return tgt_seek(config, offset);
}

static struct config_iterator *
config_seek(struct config_iterator *config, loff_t *offset)
{
	struct mpl_host   *host = NULL;
	struct mpl_device *device = NULL;
	struct mpl_path   *path = NULL;
	struct mpl_path   *presentation = NULL;
	struct config_iterator *rv = NULL;
	struct config_device *cdevice = NULL;
	struct config_device *loopdevice;
	struct config_path *cpath = NULL;
	loff_t p = *offset;
	int  i;

	mpl_module_lock();
	for (i = 0; (host = mpl_find_host(i)) != NULL; ++i) {
		if (host->flags & ( MPL_HOST_SCAN_INP | MPL_HOST_SCAN_BLOCKED)) {
			if (!config->module) {
				config->module = 1;
				config->which = M_MODULE;
				rv = config;
			}
			mpl_module_unlock();
			return rv;
                }
		if (!p--) {
			mpl_get_hba_wwpn(host->hba, &host->wwpn[0]);
			config->which = M_HOST;
			rv = config;
			goto done;
		}
		for (cpath = mpl_config_paths; cpath; cpath = cpath->next) {
			if ((cpath->flags & MPL_CFG_PATH_EXPORT) && !cpath->found &&
			   (!memcmp( host->wwpn, cpath->hba_wwpn, 8))) {
				if (!p--) {
					config->which = M_CPATH;
					rv = config;
					goto done;
				}
			}
		}
		list_for_each_entry(device, &host->devices, devices) {
			presentation = mpl_first_path(device);
			if (presentation->flag & PATH_ALWAYS_EXCLUDED)
				continue;
			if (!p--) {
				config->which = M_DEVICE;
				rv = config;
				goto done;
			}
			list_for_each_entry(path, &device->paths, path_list) {
				if (path->flag & PATH_ALWAYS_EXCLUDED)
					continue;
				if (!p--) {
					config->which = M_PATH;
					rv = config;
					goto done;
				}
			}
			/* Make sure we found the static primary path.
			 * If not, we need to display it as notfound.
			 */

			for (loopdevice = mpl_config_devices; loopdevice;
						loopdevice = loopdevice->next) {
				if (loopdevice == mpl_exclusion_device)
					continue;
				if ((loopdevice->serial_size == device->serial_size) &&
					(!memcmp(device->serial_number,
					loopdevice->serial_number,
					device->serial_size))) {
					cpath = loopdevice->path;
					if (cpath && !(device->StaticConfigPrimary)) {
						if (!p--) {
							config->which = M_CPATH;
							rv = config;
							goto done;
						}
						cpath = NULL;
					}
				}
			}
		}
	}
	for (cdevice = mpl_config_devices; cdevice; cdevice = cdevice->next) {
		if (cdevice == mpl_exclusion_device)
			continue;
		if ((cdevice->serial_size) && !cdevice->found) {
			if (!p--) {
				config->which = M_CDEVICE;
				rv = config;
				goto done;
			}
			if (cdevice->path) {
				if (!p--) {
					config->which = M_CPATH;
					rv = config;
					goto done;
				}
			}
		}
	}


	if(!config->module) {
		config->module = 1;
		config->which = M_MODULE;
		rv = config;
	}
done:
	if (host) {
		config->host = *host;
		if (device) {
			config->device = *device;
			if (path) {
				config->path = *path;
				config->path_address = path;
			}
			if (cpath) {
				config->cpath = *cpath;
			}
		}
	}
	if (cdevice) {
		config->cdevice = *cdevice;
		if (cdevice->path) {
			config->cpath = *(cdevice->path);
		}
	}
	if (cpath) {
		config->cpath = *cpath;
	}
	mpl_module_unlock();
	return rv;
}

static void *
config_start(struct seq_file *file, loff_t *offset)
{
	struct config_iterator *config;

	config = file->private;
	return config_seek(config, offset);
}

static void *
config_next(struct seq_file *file, void *arg, loff_t *offset)
{
	struct config_iterator *config = arg;

	++*offset;
	return config_seek(config, offset);
}

static void
config_stop(struct seq_file *file, void *arg)
{
	return;
}

static char *path_state[] = {
	[PATH_ONLINE] = "online",
	[PATH_DISABLED] = "disabled",
	[PATH_FAILED] = "failed",
	[PATH_STANDBY] = "standby",
};

static char *device_state[] = {
	[DEV_ONLINE] = "online",
	[DEV_FAILED] = "failed",
	[DEV_FAILING] = "failing",
	[DEV_DISABLED] = "disabled",
};

static char *hb_state[] = {
	[HB_STOPPED] = "stopped",
	[HB_TUR] = "tur",
	[HB_RTPG] = "rtpg",
	[HB_INQSN] = "inqsn",
	[HB_WAITING] = "waiting",
};

static int
format_path_state(unsigned char *buf, size_t size, enum mpl_path_state state)
{
	if (state < NUM_ENTRIES(path_state)) {
		return snprintf(buf, size, "%s", path_state[state]);
	}
	return snprintf(buf, size, "Invalid path state: %d", state);
}

static int
format_device_state(unsigned char *buf, size_t size,
		    enum mpl_device_state state)
{
	if (state < NUM_ENTRIES(device_state)) {
		return snprintf(buf, size, "%s", device_state[state]);
	}
	return snprintf(buf, size, "Invalid device state: %d", state);
}

static int
format_hb_state(unsigned char *buf, size_t size, enum mpl_hb_event state)
{
	if (state < NUM_ENTRIES(hb_state)) {
		return snprintf(buf, size, "%s", hb_state[state]);
	}
	return snprintf(buf, size, "Invalid heartbeat state: %d", state);
}

static int
format_dev_type(unsigned char *buf, size_t size, int scsi_type)
{
	static char *type_names[] = {
		[0] = "disk",
		[1] = "tape",
		[2] = "printer",
		[3] = "processor",
		[4] = "worm",
		[5] = "cd/dvd",
		[8] = "changer",
		[12] = "raid",
		[13] = "enclosure",
		[14] = "simple-disk",
		[15] = "ocr",
	};

	if (scsi_type < NUM_ENTRIES(type_names)) {
		return snprintf(buf, size, "%s", type_names[scsi_type]);
	}
	return snprintf(buf, size, "%d", scsi_type);
}

static int
format_serial(unsigned char *buf, size_t size, unsigned char *serial,
	      size_t serial_size)
{
	unsigned char *cp = buf;
	size_t  n;
	size_t  i;

	for (i = 0; size && i < serial_size; ++i) {
		if (cp > buf && size > 1) {
			*cp++ = ':';
			--size;
		}
		n = snprintf(cp, size, "%02x", serial[i]);
		if (n > size) {
			n = size;
		}
		size -= n;
		cp += n; 
	}
	return cp - buf;
}

static int
format_configured(unsigned char *buf, size_t size,
		  enum mpl_config_state configured)
{
	static char *names[] = {
		[CONFIG_UNCONFIGURED] = "unconfigured",
		[CONFIG_CONFIGURED] = "configured",
		[CONFIG_UNAVAILABLE] = "unavailable",
	};
	int n;

	if (configured < NUM_ENTRIES(names)) {
		n = snprintf(buf, size, "%s", names[configured]);
	} else {
		n = snprintf(buf, size, "invalid config state %d\n",
			     configured);
	}
	return n;
}

static char *
device_treatment(struct mpl_device *device)
{
	static char *names[] = {
		[MPL_PC_TYPE_TRAFFIC_RATIO] = "ratio",
		[MPL_PC_TYPE_DYNAMIC_LUNQ] =  "lunq",
		[MPL_PC_TYPE_DYNAMIC_TGTQ] =  "tgtq",
		[MPL_PC_TYPE_DYNAMIC_HBAQ] =  "hbaq",
		[MPL_PC_TYPE_DYNAMIC_TIME] =  "time",
	};
	static char *names_opt[] = {
		[MPL_PC_TYPE_TRAFFIC_RATIO] = "ratio, opt",
		[MPL_PC_TYPE_DYNAMIC_LUNQ] =  "lunq, opt",
		[MPL_PC_TYPE_DYNAMIC_TGTQ] =  "tgtq, opt",
		[MPL_PC_TYPE_DYNAMIC_HBAQ] =  "hbaq, opt",
		[MPL_PC_TYPE_DYNAMIC_TIME] =  "time, opt",
	};

	if (!mpl_is_device_configured(device))
		return "unknown";

	if (device->path_cntl_type == MPL_PC_TYPE_FAILOVER) {
		return (mpl_is_auto_failback(device)
			? "autofail"
			: "noautofail");
	}
	else {
		char * ptr;

		ptr = names[device->path_cntl_type];
		if (!(mpl_is_optimized(device))) {
			return names[device->path_cntl_type];
		}
		else {
			return names_opt[device->path_cntl_type];
		}
	}
	return "invalid ctl type";
}

static char *
cdevice_treatment(struct config_device *cdevice)
{
	static char *names[] = {
		[MPL_PC_TYPE_TRAFFIC_RATIO] = "ratio",
		[MPL_PC_TYPE_DYNAMIC_LUNQ] =  "lunq",
		[MPL_PC_TYPE_DYNAMIC_TGTQ] =  "tgtq",
		[MPL_PC_TYPE_DYNAMIC_HBAQ] =  "hbaq",
		[MPL_PC_TYPE_DYNAMIC_TIME] =  "time",
	};
	uint32_t flags;

	flags = (cdevice->flags & MPL_PC_TYPE_MASK);
	if (flags == MPL_PC_TYPE_FAILOVER) {
		return ((cdevice->flags & MPL_PC_CFG_AUTO_FAILBACK)
			? "autofail"
			: "noautofail");
	}
	else {
		return names[flags];
	}
	return "invalid ctl type";
}

static char *
format_path_status(unsigned char *buf, size_t size, struct mpl_device *device,
		   struct mpl_path *path)
{
	switch (device->path_cntl_type) {
	case MPL_PC_TYPE_FAILOVER:
		snprintf(buf, size, (mpl_is_path_primary(path)
				     ? "primary"
				     : "secondary"));
		break;

	case MPL_PC_TYPE_TRAFFIC_RATIO:
		snprintf(buf, size, "%d", path->traffic_ratio);
		break;

	case MPL_PC_TYPE_DYNAMIC_LUNQ:
	case MPL_PC_TYPE_DYNAMIC_TGTQ:
	case MPL_PC_TYPE_DYNAMIC_HBAQ:
	case MPL_PC_TYPE_DYNAMIC_TIME:
		snprintf(buf, size, "traffic");
		break;

	default:
		snprintf(buf, size, "invalid");
		break;
	}
	return buf;
}

extern  uint32_t mpl_max_can_queue;
static int
config_show(struct seq_file *file, void *arg)
{
	struct config_iterator *config = arg;
	struct mpl_host *host = &config->host;
	struct mpl_device *device = &config->device;
	struct mpl_path *path = &config->path;
	struct config_device *cdevice = &config->cdevice;
	struct config_path *cpath = &config->cpath;
	char *static_dynamic;
	char *active_inactive;
	char *last_path_flag;

	mpl_module_lock();
	switch (config->which) {
	case M_HOST:
		mpl_cdev_first = 1;
		seq_printf(file, "host(%d): %p, scsi_host(%d) %p, "
			   "hba %p, io_count %u",
			   host->host_no, host,
			   (host->scsi_host
			    ? host->scsi_host->host_no
			    : 0),
			   host->scsi_host, host->hba,
			   atomic_read(&host->io_count));
		seq_printf(file, ", WWPN %s", mpl_print_wwpn(host->wwpn));
	        if (host->scsi_host)
		  seq_printf(file, " canq %d mcanq %d hcanq %d", host->scsi_host->can_queue, mpl_max_can_queue, host->can_queue);
		seq_printf(file, "\n");
		break;
	case M_DEVICE:
		trim_to(config->vendor, device->vendor,
			sizeof(config->vendor));
		trim_to(config->product, device->product,
			sizeof(config->product));
		trim_to(config->rev, device->rev, sizeof(config->rev));
		format_dev_type(config->dev_type, sizeof(config->dev_type),
				device->scsi_type);
		if (mpl_is_device_configured(device)) {
			format_device_state(config->state_buffer, 
				    sizeof(config->state_buffer),
				    device->state);
		}
		else {
			snprintf(config->state_buffer, 8, "unknown");
		}
		memset(config->serial_buffer, 0, sizeof(config->serial_buffer));
		format_serial(config->serial_buffer,
			      sizeof(config->serial_buffer),
			      device->serial_number, device->serial_size);
		format_configured(config->config_buf,
				  sizeof(config->config_buf),
				  device->configured);
		seq_printf(file, "  device: %d:%d/%d/%d %s, %s, %s, %s, %s, %s",
			   device->host->host_no,
			   device->id.host_no,
			   device->id.tgt_no,
			   device->id.lun_no,
			   config->dev_type,
			   config->vendor, config->product, config->rev,
			   config->serial_buffer,
			   config->config_buf);
		seq_printf(file, ", %s, %s, %s, %s",
			   config->state_buffer,
			   device_treatment(device),
			   device->StaticConfig ?
				(device->StaticConfigPrimary ?
				"static_p" : "static") : "dynamic",
			   device->plugin ? device->plugin->name : "<null>");
		if(mpl_enable_tgt_qdepth && device->tgt_queue_depth) {
			seq_printf(file, ", tqdepth(%d)",
			   device->tgt_queue_depth);
		}
	        if (device->host->scsi_host)
		  seq_printf(file, " canq %d", device->host->scsi_host->can_queue);
		seq_printf(file, " optcnt %d\n",
			atomic_read(&device->opt_path_cnt));
		break;
	case M_PATH:
		if (mpl_is_path_configured(path)) {
			format_path_state(config->state_buffer,
				  sizeof(config->state_buffer),
				  path->state);
			format_path_status(config->status, sizeof(config->status),
				   device, path);
			last_path_flag = (config->path_address == device->last_path
				  ? "*"
				  : "");
		}
		else {
			snprintf(config->state_buffer, 8, "unknown");
			snprintf(config->status, 8, "unknown");
			last_path_flag = "";
		}
		format_configured(config->config_buf,
				  sizeof(config->config_buf),
				  path->configured);
		static_dynamic = path->StaticConfig ? "static" : "dynamic";
		active_inactive = ((path->state == PATH_FAILED ||
				    path->state == PATH_DISABLED)
				   ? "cntrlr_unknown"
				   : (path->active_controller
				      ? "active_cntrlr"
				      : "inactive_cntrlr"));
		seq_printf(file, "      path: %d/%d/%d, %s",
			   path->host->host_no, path->target, path->lun,
			   config->config_buf);
		seq_printf(file, ", %s, %s, %s, %s",
			   config->state_buffer,
			   config->status,
			   static_dynamic,
			   active_inactive);
		seq_printf(file, ", WWPN %s %s",
			   mpl_print_wwpn(path->tgt->wwpn), last_path_flag);
	        if (path->host->scsi_host)
		  seq_printf(file, " canq %d", path->host->scsi_host->can_queue);
		seq_printf(file, ",tpg x%x:alua x%x x%x%s\n",
			path->tgt_port_group, path->tpgs, path->sas,
			path->lld_blocked ? ", blocked" : "");
		break;
	case M_CDEVICE:
		if (mpl_cdev_first) {
			mpl_cdev_first = 0;
			seq_printf(file, "config:\n");
		}
		memset(config->serial_buffer, 0, sizeof(config->serial_buffer));
		format_serial(config->serial_buffer,
			      sizeof(config->serial_buffer),
			      cdevice->serial_number, cdevice->serial_size);
		seq_printf(file, "  device: U:U/U/U %s, %s, %s, %s, %s, %s, %s, %s, %s, %s\n",
			"disk",
			"UNK", "UNK", "UNK",
			config->serial_buffer,
			"configured",
			((cdevice->found == 1) ? "found" : "notfound"),
			cdevice_treatment(cdevice),
			((cdevice->path) ? "static_p" : "static"),
			"<null>");
		break;
	case M_CPATH:
		seq_printf(file, "      path: U/U/%d,", cpath->lun_no);
		if ((cpath->flags & MPL_CFG_PATH_EXPORT) && !cpath->found) {
			seq_printf(file, " %s",
				"unavailable, notfound, unknown, static, cntrlr_unknown");
		}
		else {
			seq_printf(file, " %s",
				"configured, notfound, unknown, static, cntrlr_unknown");
		}
		seq_printf(file, ", WWPN %s  ",
			   mpl_print_wwpn(cpath->target_wwpn));
		seq_printf(file, "HOST %s\n",
			   mpl_print_wwpn(cpath->hba_wwpn));
		break;
	case M_MODULE:
		seq_printf(file, "module: ");
		seq_printf(file, "LOG: verbose(x%x) ", mpl_log_verbose);
		seq_printf(file, "HBEAT: interval(%d) fail(%d) retry(%d)\n",
			   mpl_hbeat_interval, mpl_hbeat_fail_interval,
			   mpl_max_hbeat_retry);
		seq_printf(file, "   AUTO: configure(%d) lun_balance(%d)\n",
		 	   mpl_auto_configure, mpl_auto_lun_balance);
		seq_printf(file, " QDEPTH: enable_tgt(%d) ",
			   mpl_enable_tgt_qdepth);
		seq_printf(file, "handle_qfull(%d) ", mpl_handle_qfull);
		seq_printf(file, "tot_int_cnt(%d)", atomic_read(&total_internal_count));
		seq_printf(file, "\n");
		break;
	default:
		seq_printf(file, "Ack! should never have gotten here.\n");
		break;
	}
	mpl_module_unlock();
	return 0;
}

static int
exclusion_show(struct seq_file *file, void *arg)
{
	struct config_iterator *config = arg;
	struct config_path *cpath = &config->cpath;

	mpl_module_lock();
	switch (config->which) {
	case M_CPATH:
		if (cpath->path && cpath->found) {
			seq_printf(file, "      path: %d/%d/%d, excluded,    found",
			   cpath->path->host->host_no, cpath->path->target, cpath->lun_no);
		}
		else {
			seq_printf(file, "      path: U/U/%d, excluded, notfound",
				cpath->lun_no);
		}
		seq_printf(file, ", WWPN %s  ",
			   mpl_print_wwpn(cpath->target_wwpn));
		seq_printf(file, "HOST %s\n",
			   mpl_print_wwpn(cpath->hba_wwpn));
		break;
	default:
		seq_printf(file, "Ack! should never have gotten here.\n");
		break;
	}
	mpl_module_unlock();
	return 0;
}

static int
stats_show(struct seq_file *file, void *arg)
{
	struct config_iterator *config = arg;
	struct mpl_host *host = &config->host;
	struct mpl_device *device = &config->device;
	struct mpl_target *tgt = &config->target;
	struct mpl_path *path = &config->path;

	mpl_module_lock();
	switch (config->which) {
	case M_HOST:
		seq_printf(file, "host: number %d, io_count %u, tot %u\n",
			   host->host_no,
			   atomic_read(&host->io_count),
			   atomic_read(&host->tot_count));
		break;
	case M_TARGET:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
		seq_printf(file, "lpfc%dt%d, paths %u, prim %u, io_count %u, tot %u\n",
			   tgt->host->host_no,
			   tgt->id,
			   atomic_read(&tgt->path_ref_count),
			   atomic_read(&tgt->prim_path_ref_count),
			   atomic_read(&tgt->io_count),
			   atomic_read(&tgt->tot_count));
#else
		seq_printf(file, "lpfc%dt%d, paths %u, prim %u, io_count %u, tot %u rport %p\n",
			   tgt->host->host_no,
			   tgt->id,
			   atomic_read(&tgt->path_ref_count),
			   atomic_read(&tgt->prim_path_ref_count),
			   atomic_read(&tgt->io_count),
			   atomic_read(&tgt->tot_count), tgt->rport);
#endif
		break;
	case M_DEVICE:
		seq_printf(file, "  device: %d/%d/%d,\ttot %u, io_count %u, "
			   "pending  %u, defer %u eh_defer %u\n",
			   device->id.host_no,
			   device->id.tgt_no,
			   device->id.lun_no,
			   atomic_read(&device->tot_count), 
			   atomic_read(&device->io_count), 
			   atomic_read(&device->pending_count),
			   atomic_read(&device->deferral_count),
			   atomic_read(&device->eh_deferral_count));
		break;
	case M_PATH:
		seq_printf(file, "      path: %d/%d/%d,\ttot %u, io_count %u, "
			   "internal %u avg(%u):%ld\n",
			   path->host->host_no, path->target, path->lun,
			   atomic_read(&path->tot_count),
			   atomic_read(&path->io_count),
			   atomic_read(&path->internal_count),
			   atomic_read(&path->avg_cnt),
			   path->avg_time);
		break;
	case M_MODULE:
		break;
	default:
		seq_printf(file, "Ack! should never have gotten here.\n");
		break;
	}
	mpl_module_unlock();
	return 0;
}

static int
debug_show(struct seq_file *file, void *arg)
{
	struct config_iterator *config = arg;
	struct mpl_host *host = &config->host;
	struct mpl_device *device = &config->device;
	struct mpl_path *path = &config->path;

	switch (config->which) {
	case M_HOST:
		seq_printf(file, "host: number %d, blocked %d\n",
			   host->host_no,
			   host->blocked);
		break;
	case M_DEVICE:
		if (device->configured != CONFIG_CONFIGURED)
			break;
		format_device_state(config->state_buffer, 
				    sizeof(config->state_buffer),
				    device->state);
		seq_printf(file, "device: %d/%d/%d,state: %s, "
			   "pend migration: %d\n",
			   device->id.host_no,
			   device->id.tgt_no,
			   device->id.lun_no,
			   config->state_buffer,
			   device->pending_lun_migration);
		break;
	case M_PATH:
		format_path_state(config->state_buffer,
				  sizeof(config->state_buffer),
				  path->state);
		format_hb_state(config->hb_state_buffer,
				sizeof(config->hb_state_buffer),
				path->hbstate);
		seq_printf(file, "  path: %d/%d/%d, flg x%x sdev %p state: %s, hb state: %s, "
			   "hb timer: %ld, shed_cmd timer: %ld, "
			   " blocked: %d, sdev_block: %ld\n",
			   path->host->host_no, path->target, path->lun,
			   path->flag,
			   path->sdev,
			   config->state_buffer,
			   config->hb_state_buffer,
			   (long) (path->hbeat.expires
				   ? path->hbeat.expires - jiffies
				   : 0),
			   (long) (path->sched_cmd.expires
				   ? path->sched_cmd.expires - jiffies
				   : 0),
			   path->lld_blocked,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
			   (long) (path->sdev_block.expires
				   ? path->sdev_block.expires - jiffies
				   : 0)
#else
				   0L
#endif
				   );
		break;
	case M_MODULE:
		break;
	default:
		seq_printf(file, "Ack! should never have gotten here.\n");
		break;
	}
	return 0;
}

struct seq_operations config_seq_ops = {
	.start = config_start,
	.next  = config_next,
	.stop  = config_stop,
	.show  = config_show
};

struct seq_operations exclusion_seq_ops = {
	.start = exclusion_start,
	.next  = exclusion_next,
	.stop  = config_stop,
	.show  = exclusion_show
};

struct seq_operations stats_seq_ops = {
	.start = config_start,
	.next  = config_next,
	.stop  = config_stop,
	.show  = stats_show
};

struct seq_operations stats_tgt_seq_ops = {
	.start = tgt_start,
	.next  = tgt_next,
	.stop  = config_stop,
	.show  = stats_show
};

struct seq_operations debug_seq_ops = {
	.start = config_start,
	.next  = config_next,
	.stop  = config_stop,
	.show  = debug_show
};

static int
common_open(struct inode *inode, struct file *file,
	    struct seq_operations *seq_ops)
{
	struct config_iterator *config;
	int  rv;

	config = kmalloc(sizeof(*config), GFP_KERNEL);
	if (!config)
		return -ENOMEM;
	memset(config, 0, sizeof(*config));
	rv = seq_open(file, seq_ops);
	seq_private_from_file(file) = config;
	return rv;
}

static int
stats_open(struct inode *inode, struct file *file)
{
	return common_open(inode, file, &stats_seq_ops);
}

static int
stats_tgt_open(struct inode *inode, struct file *file)
{
	return common_open(inode, file, &stats_tgt_seq_ops);
}

static int
config_open(struct inode *inode, struct file *file)
{
	struct mpl_host   *host;
	int  i, retry;

	mpl_module_lock();
	retry = 5;
	for (i = 0; (host = mpl_find_host(i)) != NULL; ++i) {
		while (host->flags & ( MPL_HOST_SCAN_INP | MPL_HOST_SCAN_BLOCKED)) {
			if (!retry) {
				mpl_module_unlock();
				return -EBUSY;
			}
			retry--;
			mpl_module_unlock();
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout(HZ);
			mpl_module_lock();
		}
	}
	mpl_module_unlock();
	return common_open(inode, file, &config_seq_ops);
}

static int
exclusion_open(struct inode *inode, struct file *file)
{
	return common_open(inode, file, &exclusion_seq_ops);
}

static int
debug_open(struct inode *inode, struct file *file)
{
	return common_open(inode, file, &debug_seq_ops);
}

static int
common_close(struct inode *inode, struct file *file)
{
	struct config_iterator *config;

	config = seq_private_from_file(file);
	kfree(config);
	return seq_release(inode, file);
}

static uint32_t
get_number(char **line, char *name)
{
	char *cp;

	if ((cp = strsep(line, " \t/")) == NULL) {
		mpl_print(0, MPL_ERR "missing %s\n",
		       name);
		return ~0;
	}
	return simple_strtoul(cp, NULL, 10);
}

int
mpl_crack_serial_number(uint8_t *serial, size_t size, char *src)
{
	int n = 0;

	memset(serial, 0, size);
	while (isxdigit(src[0]) && isxdigit(src[1])) {
		if (!size) {
			return 0;
		}
		--size;
		*serial = (mpl_xtobin(src[0]) << 4) + mpl_xtobin(src[1]);
		++serial;
		++n;
		src += 2;
		if (*src == ':') {
			++src;
		}
	}
	if (*src) {		/* non parsable data at end of string? */
		return 0;
	}
	return n;
}

static struct mpl_device *
mpl_find_device_by_serial(char *serial, size_t size)
{
	struct mpl_host *host;
	struct mpl_device *device;
	int  i;

	for (i = 0; (host = mpl_find_host(i)) != NULL; ++i) {
		list_for_each_entry(device, &host->devices, devices) {
			if (device->serial_size >= size &&
			    !memcmp(device->serial_number, serial,
				    device->serial_size)) {
				return device;
			}
		}
	}
	return NULL;
}

static struct mpl_path *
lookup_first_path_by_inqsn(char **s)
{
	struct mpl_device *device = NULL;
	uint8_t serial[sizeof(((struct mpl_device *) 0)->serial_number)];
	char *serial_number = *s;
	int  n;

	if ((serial_number = strsep(s, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "missing path id\n");
	} else {
		n = mpl_crack_serial_number(serial, sizeof(serial), serial_number);
		if (!n) {
			mpl_print(0, MPL_ERR "Invalid serial number: %s\n",
				  serial_number);
		} else {
			device = mpl_find_device_by_serial(serial, n);
			if (!device) {
				mpl_print(0, MPL_ERR "Cannot find device: %s\n",
					  serial_number);
			}
		}
	}
	return device ? mpl_first_path(device) : NULL;
}

static struct mpl_path *
find_path_by_wwpns(unsigned char hba_wwpn[8], unsigned char target_wwpn[8],
		   unsigned int lun)
{
	struct mpl_host *host;
	struct mpl_device *device;
	struct mpl_path *path;
	int  i;

	for (i = 0; (host = mpl_find_host(i)) != NULL; ++i) {
		list_for_each_entry(device, &host->devices, devices) {
			list_for_each_entry(path, &device->paths, path_list) {
				if (!memcmp(path->host->wwpn, hba_wwpn, 8) &&
				    !memcmp(path->tgt->wwpn, target_wwpn, 8) &&
				    path->lun == lun) {
					return path;
				}
			}
		}
	}
	return NULL;
}

static struct mpl_path *
lookup_path_by_wwpn(char **s)
{
	char wwpn_buffer[48];
	const char *ccp = wwpn_buffer;
	unsigned char hba_wwpn[8];
	unsigned char target_wwpn[8];
	unsigned int  lun;

	strncpy(wwpn_buffer, *s, sizeof(wwpn_buffer));
	wwpn_buffer[sizeof(wwpn_buffer) - 1] = 0;
	mpl_crack_wwn(hba_wwpn, ccp, &ccp);
	mpl_crack_wwn(target_wwpn, ccp, &ccp);
	while (isspace(*ccp) || *ccp == ':') {
		++ccp;
	}
	lun = simple_strtol(ccp, NULL, 0);
	return find_path_by_wwpns(hba_wwpn, target_wwpn, lun);
}

static struct mpl_path *
lookup_path_by_htl(char **s)
{
	struct mpl_host *host;
	uint32_t host_no;
	uint32_t target;
	uint32_t lun;

	host_no = get_number(s, "adapter number");
	target = get_number(s, "target id");
	lun = get_number(s, "lun number");
	if (host_no == ~0 || target == ~0 || lun == ~0)
		return NULL;
	host = mpl_find_host(host_no);
	if (!host) {
		mpl_print(0, MPL_ERR "Cannot find host %d for %d/%d/%d %s\n",
		       host_no, host_no, target, lun, *s);
		return NULL;
	}
	return mpl_find_path(host, target, lun);
}

static struct mpl_path *
lookup_path(char **line)
{
	struct mpl_path *path;
	char *cp;

	if ((cp = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "missing path id\n");
		path = NULL;
	} else if (cp[2] == ':') {
		path = lookup_first_path_by_inqsn(&cp);
	} else if (strchr(cp, ':')) {
		path = lookup_path_by_wwpn(&cp);
	} else {
		path = lookup_path_by_htl(&cp);
	}
	return path;
}

static int
configure_device(struct mpl_device *device, enum mpl_config_state configured)
{
	struct mpl_path *path;

	mpl_set_device_configured(device, configured);
	if (configured == CONFIG_UNCONFIGURED) {
		list_for_each_entry(path, &device->paths, path_list) {
			if (path->configured == CONFIG_CONFIGURED)
				mpl_set_path_configured(path, configured);
		}
	}
	return 0;
}

static int
set_primary_path(struct mpl_device *device, char *serial_number, char **line)
{
	struct mpl_path *path, *primary;
	char *cp = *line;
	int  status = -EINVAL;

	path = lookup_path(line);
	if (!path) {
		mpl_print(0, MPL_ERR "Cannot find path: %s, %s\n",
		       serial_number, cp);
		return -EINVAL;
	}
	/* This this time, we was aloowing active or inactive controllers
	 * to be configured to be the primary path.
	 */
	primary = mpl_get_primary_path(device);
	if (path == primary) {
		path->device->StaticConfigPrimary = 1;
		path->StaticConfig = 1;
		return 0;	/* done. */
	}

	status = mpl_set_path_primary(path, 1);
	return status;
}

static int
deconfigure_paths(struct mpl_device *device)
{
	struct mpl_path *path;

	list_for_each_entry(path, &device->paths, path_list) {
		if (path->configured == CONFIG_CONFIGURED)
			mpl_set_path_configured(path, CONFIG_UNCONFIGURED);
	}
	return 0;
}

static int
do_device_command(char **line)
{
	struct  mpl_device *device;
	uint8_t serial[sizeof(((struct mpl_device *) 0)->serial_number)];
	int     status = -EINVAL;
	char    *serial_number;
	char    *cp;
	int     n;

	if ((serial_number = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "No device parameters\n");
		return -EINVAL;
	}
	n = mpl_crack_serial_number(serial, sizeof(serial), serial_number);
	if (!n) {
		mpl_print(0, MPL_ERR "Invalid device serial number: %s\n",
		       serial_number);
		return -EINVAL;
	}
	device = mpl_find_device_by_serial(serial, n);
	if (!device) {
		mpl_print(0, MPL_ERR "Cannot find device: %s for %s\n",
		       serial_number, *line);
		return -EINVAL;
	}
	if ((cp = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "Do what with device? %s\n",
		       serial_number);
		return -EINVAL;
	}
	if (!strcmp(cp, "configure")) {
		status = configure_device(device, CONFIG_CONFIGURED);
		device->path_cntl_type = MPL_PC_TYPE_FAILOVER; /* default */
		device->deconfig_api = 0;
	} else if (!strcmp(cp, "deconfigure")) {
		status = configure_device(device, CONFIG_UNCONFIGURED);
		device->deconfig_api = 1;
		if (!status) {
			status = deconfigure_paths(device);
		}
	} else if (!strcmp(cp, "unavailable")) {
		status = configure_device(device, CONFIG_UNAVAILABLE);
	} else if (!strcmp(cp, "notoptimized")) {
		status = mpl_set_device_optimized(device, 0);
	} else if (!strcmp(cp, "optimized")) {
		status = mpl_set_device_optimized(device, 1);
	} else if (!strcmp(cp, "failback")) {
		status = mpl_set_device_auto_failback(device, 1);
		if (!status)
			mpl_set_device_static_config(device);
	} else 	if (!strcmp(cp, "nofailback")) {
		status = mpl_set_device_auto_failback(device, 0);
		if (!status)
			mpl_set_device_static_config(device);
	} else 	if (!strcmp(cp, "primary")) {
		status = set_primary_path(device, serial_number, line);
		if (!status)
			mpl_set_device_static_config(device);
	} else 	if (!strcmp(cp, "type-failover")) {
		status = mpl_set_device_treatment(device,
						  MPL_PC_TYPE_FAILOVER);
		if (!status)
			mpl_set_device_static_config(device);
	} else 	if (!strcmp(cp, "type-ratio")) {
		status = mpl_set_device_treatment(device,
						  MPL_PC_TYPE_TRAFFIC_RATIO);
		if (!status)
			mpl_set_device_static_config(device);
	} else 	if (!strcmp(cp, "type-lunq")) {
		status = mpl_set_device_treatment(device,
						  MPL_PC_TYPE_DYNAMIC_LUNQ);
		if (!status)
			mpl_set_device_static_config(device);
	} else 	if (!strcmp(cp, "type-targetq")) {
		status = mpl_set_device_treatment(device,
						  MPL_PC_TYPE_DYNAMIC_TGTQ);
		if (!status)
			mpl_set_device_static_config(device);
	} else 	if (!strcmp(cp, "type-hbaq")) {
		status = mpl_set_device_treatment(device,
						  MPL_PC_TYPE_DYNAMIC_HBAQ);
		if (!status)
			mpl_set_device_static_config(device);
	} else 	if (!strcmp(cp, "type-time")) {
		status = mpl_set_device_treatment(device,
						  MPL_PC_TYPE_DYNAMIC_TIME);
		if (!status)
			mpl_set_device_static_config(device);
	} else {
		mpl_print(0, MPL_ERR "Unrecognized action for device %s: %s\n",
		       serial_number, cp);
	}
	return status;
}

static int
do_path_status(struct mpl_path *path, char **line)
{
	int     status = -EINVAL;
	char *cp;

	if ((cp = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "Missing status parameter\n");
		return -EINVAL;
	}
	if (!strcmp(cp, "online")) {
		status = mpl_set_path_online(path);
	} else if (!strcmp(cp, "offline")) {
		status = mpl_set_path_offline(path);
	} else if (!strcmp(cp, "clear")) {
		status = mpl_clear_path_statistics(path);
	} else {
		mpl_print(0, MPL_ERR "Unrecognized path status request: %s\n",
		       cp);
	}
	return status;
}

static int
set_path_ratio(struct mpl_path *path, char **line)
{
	char *cp;
	unsigned int ratio;

	if ((cp = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "Missing traffic ratio\n");
		return -EINVAL;
	}
	ratio = simple_strtol(cp, NULL, 0);
	mpl_set_path_ratio(path, ratio);
	return 0;
}

static int
do_path_configure_device(struct mpl_path *path)
{
	int  status;

	status = mpl_set_device_configured(path->device, CONFIG_CONFIGURED);
	if (!status) {
		status = mpl_set_path_configured(path, CONFIG_CONFIGURED);
	}
	if (!status) {
		struct mpl_host *new_host;
		struct mpl_device *device;

		new_host = path->host;
		device = path->device;
		if (!mpl_is_presentation_path(path)) {
			list_del(&path->path_list);
			list_add(&path->path_list, &path->device->paths);
			if (device->host != new_host) {
				list_del(&device->devices);
				list_add_tail(&device->devices,
					      &new_host->devices);
				device->host = new_host;
			}
		}

		path->device->id.host_no = path->host->host_no;
		path->device->id.tgt_no = path->target;
		path->device->id.lun_no = path->lun;
	}
	return status;
}

static int
do_path_deconfigure(struct mpl_path *path)
{
	int status;

	if (mpl_is_presentation_path(path)) {
		if (path->configured != CONFIG_UNAVAILABLE)
			return -EBUSY;
	}
	if (path->configured == CONFIG_UNAVAILABLE) {
		status = mpl_set_path_configured(path, CONFIG_UNCONFIGURED);
		if (!status) {
			/* For this case we are Importing the path back */
			path->StaticConfig = 0;
			mpl_num_exclusion_devices--;
		}
	}
	else {
		status = mpl_set_path_configured(path, CONFIG_UNCONFIGURED);
	}
	return status;
}

static int
do_path_command(char **line)
{
	struct  mpl_path   *path;
	struct  mpl_device *device;
	int     status = -EINVAL;
	char    *cp;

	cp = *line;
	path = lookup_path(line);
	if (!path) {
		mpl_print(MPL_IOCTL, MPL_INFO "Cannot find path: %s\n", cp);
		cp = strsep(line, " \t");
		return -EINVAL;
	}
	device = path->device;
	if ((cp = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "Do what with device at %s ?\n", cp);
		return -EINVAL;
	}
	if (!strcmp(cp, "status")) {
		status = do_path_status(path, line);
	} else if (!strcmp(cp, "configure")) {
		if (path->configured == CONFIG_UNAVAILABLE) {
			status =
			    mpl_set_path_configured(path, CONFIG_CONFIGURED);
			if (!status)
				mpl_num_exclusion_devices--;
		}
		else {
			status =
			    mpl_set_path_configured(path, CONFIG_CONFIGURED);
		}
		if (!mpl_is_presentation_path(path) && path->sdev) {
			path->sdev->type = TYPE_PROCESSOR;
		}
	} else if (!strcmp(cp, "deconfigure")) {
		status = do_path_deconfigure(path);
	} else if (!strcmp(cp, "export")) {
		/* Enforce the limit for max number of exclusion devices */
		if (mpl_num_exclusion_devices >= MPL_MAX_EXCLUSION_DEVICES) {
			mpl_print(0, MPL_ERR "Attempt to exclude > %d paths\n",
				MPL_MAX_EXCLUSION_DEVICES);
			return -EBUSY;
		}
		status = mpl_set_path_configured(path, CONFIG_UNAVAILABLE);
		if (!status)
			mpl_num_exclusion_devices++;
	} else if (!strcmp(cp, "ratio")) {
		status = set_path_ratio(path, line);
	} else if (!strcmp(cp, "configure-device")) {
		if (mpl_is_path_configured(path) ||
		    mpl_is_device_configured(device)) {
			status = -EBUSY;
		} else {
			status = do_path_configure_device(path);
			device->deconfig_api = 0;
		}
	} else if (!strcmp(cp, "deconfigure-device")) {
		if (!mpl_is_device_configured(device)) {
			status = -ENODEV;
		} else if (mpl_quiesce_device(device, 2)) {
 				/* We're running on insmod or user app thread.
				 * Make sure the path quiesces before
				 * returning because the app may tear down the
				 * mid-layer device. */

			status = -EBUSY;
		} else {
			status = configure_device(device,
						  CONFIG_UNCONFIGURED);
		}
		if (!status) {
			status = deconfigure_paths(device);
		}
		device->deconfig_api = 1;
	} else {
		mpl_print(0, MPL_ERR "Unrecognized action for path %s\n", cp);
	}
	return status;
}

static int
mpl_set_hbeat(unsigned long value, int fail_flag)
{
	if ((value == 0) || (value > MPL_MAX_HBEAT_INTERVAL)) {
		mpl_print(0, MPL_ERR "hbeat value (%ld) out of range [1-%d]\n",
		    value, MPL_MAX_HBEAT_INTERVAL);
		return -EINVAL;
	}
	switch (fail_flag) {
	case 0: /* Heartbeat validation path interval */
		if(value < mpl_hbeat_fail_interval) {
			mpl_print(0, MPL_ERR
			"hbeat value (%ld) must be >= hbeat fail (%d)\n",
			 value, mpl_hbeat_fail_interval);
			return -EINVAL;
		}
		mpl_hbeat_interval = value;
		break;
	case 1: /* Heartbeat failed path interval */
		if(value > mpl_hbeat_interval) {
			mpl_print(0, MPL_ERR
			"hbeat fail value (%ld) must be <= hbeat (%d)\n",
			 value, mpl_hbeat_interval);
			return -EINVAL;
		}
		mpl_hbeat_fail_interval = value;
		break;
	case 2: /* Failed heartbeat cmd retry counter */
		mpl_max_hbeat_retry = value;
		break;
	}
	return 0;
}

static int
do_qfull_command(char **line)
{
	char *cp;
	unsigned long value;

	if ((cp = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "Missing qfull parameter.\n");
		return -EINVAL;
	}
	value = simple_strtoul(cp, NULL, 0);
	if (value > 1) {
		mpl_print(0, MPL_ERR "Invalid qfull parameter (boolean).\n");
		return -EINVAL;
	}
	mpl_handle_qfull = value;
	return 0;
}

static int
do_tqdepth_command(char **line)
{
	char *cp;
	unsigned long value;

	if ((cp = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "Missing tqdepth parameter.\n");
		return -EINVAL;
	}
	value = simple_strtoul(cp, NULL, 0);
	if (value > 1) {
		mpl_print(0, MPL_ERR "Invalid tqdepth parameter (boolean).\n");
		return -EINVAL;
	}
	mpl_enable_tgt_qdepth = value;
	return 0;
}

static int
do_hbeat_command(char **line)
{
	char *cp;
	int rv;

	if ((cp = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "Missing hbeat parameter.\n");
		return -EINVAL;
	}
	if (!strcmp(cp, "retry")) { /* optional parameter */
		if ((cp = strsep(line, " \t")) == NULL) {
			mpl_print(0, MPL_ERR "Missing hbeat parameter.\n");
			return -EINVAL;
		}
		rv = mpl_set_hbeat(simple_strtoul(cp, NULL, 0), 2);
	}
	else if (!strcmp(cp, "fail")) { /* optional parameter */
		if ((cp = strsep(line, " \t")) == NULL) {
			mpl_print(0, MPL_ERR "Missing hbeat parameter.\n");
			return -EINVAL;
		}
		rv = mpl_set_hbeat(simple_strtoul(cp, NULL, 0), 1);
	}
	else {
		rv = mpl_set_hbeat(simple_strtoul(cp, NULL, 0), 0);
	}
	return rv;
}

static void
create_proc_debug(void)
{
	struct proc_dir_entry *proc_dev;
	static struct file_operations debug_fops = {
		.open    = debug_open,
		.release = common_close,
		.read    = seq_read,
		.llseek  = seq_lseek,
	};

	mpl_module_unlock();
	proc_dev = create_proc_entry(DEBUG_NODE, PROC_ENTRY_FLAGS, mpl_root);
	if (proc_dev == NULL) {
		mpl_print(0, MPL_ERR "Could not create " DEBUG_NODE "\n");
	} else {
		proc_dev->owner = THIS_MODULE;
		proc_dev->proc_fops = &debug_fops;
	}
	mpl_module_lock();
}

static int
do_debug_command(char **line)
{
	char *cp;

	if ((cp = strsep(line, " \t")) == NULL) {
		mpl_print(0, MPL_ERR "Missing debug parameter.\n");
		return -EINVAL;
	}
	if (!strcmp(cp, "level")) {
		if ((cp = strsep(line, " \t")) == NULL) {
			mpl_print(0, MPL_ERR "Missing debug level.\n");
			return -EINVAL;
		}
		mpl_log_set_verbose(simple_strtoul(cp, NULL, 0));
	} else if (!strcmp(cp, "proc")) {
		create_proc_debug();
	}
	return 0;
}

static int
config_execute_line(char *line)
{
	char *cp;
	int  status = -EINVAL;

	mpl_print(MPL_IOCTL, MPL_INFO "write config: %s\n", line);
	while ((cp = strsep(&line, " \t")) != NULL && *cp != 0) {
		mpl_module_lock();
		if (!strcmp(cp, "device")) {
			status = do_device_command(&line);
		} else if (!strcmp(cp, "path")) {
			status = do_path_command(&line);
		} else if (!strcmp(cp, "debug")) {
			status = do_debug_command(&line);
		} else if (!strcmp(cp, "hbeat")) {
			status = do_hbeat_command(&line);
		} else if (!strcmp(cp, "qfull")) {
			status = do_qfull_command(&line);
		} else if (!strcmp(cp, "tqdepth")) {
			status = do_tqdepth_command(&line);
		} else {
			mpl_print(0, MPL_ERR "execute_line. token: %s.\n", cp);
		}
		mpl_module_unlock();
	}
	mpl_print(MPL_IOCTL, MPL_INFO "write config: status %d\n", status);
	return status;
}

static ssize_t
control_write(struct file *file, const char * __user buffer, size_t count, loff_t *off)
{
	char *data;
	char *save_data;
	char *cp;
	int  rv = 0;
	int  status = 0;

	data = kmalloc(count + 1, GFP_KERNEL);
	if (data == NULL) {
		rv = -EFAULT;
		goto out;
	}
	if (copy_from_user(data, buffer, count)) {
		rv = -EFAULT;
		goto out_kfree;
	}
	data[count] = 0;
	rv = count;

	/* Need to save data since strsep will eventually zero it */
	save_data = data;
	while (status == 0 && (cp = strsep(&data, "\n")) != NULL && *cp != 0) {
		status = config_execute_line(cp);
	}
	data = save_data;

	file->private_data = ERR_PTR(status);

out_kfree:
	kfree(data);
out:
	return rv;
}

static ssize_t
control_read(struct file *file, char * __user buffer, size_t count,
	     loff_t *off)
{
	char tmp[32];
	int  n;

	if (*off) {
		return 0;
	}
	n = snprintf(tmp, sizeof(tmp), "%ld\n", PTR_ERR(file->private_data));
	if (n < sizeof(tmp)) {
		tmp[n] = 0;
		++n;
	}
	if (n > count) {
		n = count;
	}
	if (copy_to_user(buffer, tmp, n)) {
		return -EFAULT;
	}
	mpl_print(MPL_IOCTL, MPL_INFO "read config: status %d\n",
		(unsigned int)(unsigned long)(file->private_data));
	*off += n;
	return n;
}

static int
control_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int
control_close(struct inode *inode, struct file *file)
{
	return 0;
}

/* The version_start, version_next, version_stop, version_release are derived
 * from single_start, single_next, single_stop, and single_release in
 * newer versions of the kernel.  Largely cloned here to support SLES 8.
 */

static void *
version_start(struct seq_file *p, loff_t *pos)
{
	return NULL + (*pos == 0);
}

static void *
version_next(struct seq_file *p, void *v, loff_t *pos)
{
	++*pos;
	return NULL;
}

static void 
version_stop(struct seq_file *p, void *v)
{
}

static int
version_show(struct seq_file *file, void *arg)
{
	seq_printf(file, MPL_MODULE_NAME ": Version " MPL_MODULE_VERSION "\n");
	return 0;
}

struct seq_operations version_seq_ops = {
	.start = version_start,
	.next  = version_next,
	.stop  = version_stop,
	.show  = version_show
};

static int 
version_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &version_seq_ops);
}

int version_release(struct inode *inode, struct file *file)
{
	return seq_release(inode, file);
}

static struct file_operations config_fops = {
	.open    = config_open,
	.release = common_close,
	.read    = seq_read,
	.llseek  = seq_lseek,
};

static struct file_operations exclusion_fops = {
	.open    = exclusion_open,
	.release = common_close,
	.read    = seq_read,
	.llseek  = seq_lseek,
};

static struct file_operations control_fops = {
	.open    = control_open,
	.release = control_close,
	.read    = control_read,
	.write   = control_write
};

static struct file_operations stats_fops = {
	.open    = stats_open,
	.release = common_close,
	.read    = seq_read,
	.llseek  = seq_lseek,
};

static struct file_operations stats_tgt_fops = {
	.open    = stats_tgt_open,
	.release = common_close,
	.read    = seq_read,
	.llseek  = seq_lseek,
};

#if 1				/* replacement of single_open,
				 * et.al. for SLES 8 */
static struct file_operations version_fops = {
	.open    = version_open,
	.release = version_release,
	.read    = seq_read,
	.llseek  = seq_lseek,
};

#else
static struct file_operations version_fops = {
	.open    = version_open,
	.release = single_release,
	.read    = seq_read,
};
#endif

int mpl_create_proc(void)
{
	struct proc_dir_entry *proc_dev;
	int rv = 0;

	mpl_root = proc_mkdir("scsi/" MPL_MODULE_NAME, 0);
	if (mpl_root == NULL) {
		mpl_print(0, MPL_ERR "Could not create " MPL_MODULE_NAME "\n");
		rv = -ENOMEM;
		goto out;
	}
	mpl_root->owner = THIS_MODULE;

	proc_dev = create_proc_entry(CONFIG_NODE, PROC_ENTRY_FLAGS, mpl_root);
	if (proc_dev == NULL) {
		mpl_print(0, MPL_ERR "Could not create " CONFIG_NODE "\n");
		rv = -ENOMEM;
		goto out_rm_root;
	}
	proc_dev->owner = THIS_MODULE;
	proc_dev->proc_fops = &config_fops;

	proc_dev = create_proc_entry(CONTROL_NODE, PROC_ENTRY_FLAGS, mpl_root);
	if (proc_dev == NULL) {
		mpl_print(0, MPL_ERR "Could not create " CONTROL_NODE "\n");
		rv = -ENOMEM;
		goto out_rm_config;
	}
	proc_dev->owner = THIS_MODULE;
	proc_dev->proc_fops = &control_fops;

	proc_dev = create_proc_entry(STATS_NODE, PROC_ENTRY_FLAGS, mpl_root);
	if (proc_dev == NULL) {
		mpl_print(0, MPL_ERR "Could not create " STATS_NODE "\n");
		rv = -ENOMEM;
		goto out_rm_control;
	}
	proc_dev->owner = THIS_MODULE;
	proc_dev->proc_fops = &stats_fops;

	proc_dev = create_proc_entry(STATS_TGT_NODE, PROC_ENTRY_FLAGS, mpl_root);
	if (proc_dev == NULL) {
		mpl_print(0, MPL_ERR "Could not create " STATS_TGT_NODE "\n");
		rv = -ENOMEM;
		goto out_rm_stats;
	}
	proc_dev->owner = THIS_MODULE;
	proc_dev->proc_fops = &stats_tgt_fops;

	proc_dev = create_proc_entry(VERSION_NODE, PROC_ENTRY_FLAGS,
					mpl_root);
	if (proc_dev == NULL) {
		mpl_print(0, MPL_ERR "Could not create " VERSION_NODE "\n");
		rv = -ENOMEM;
		goto out_rm_stats_tgt;
	}
	proc_dev->owner = THIS_MODULE;
	proc_dev->proc_fops = &version_fops;

	proc_dev = create_proc_entry(EXCLUSION_NODE, PROC_ENTRY_FLAGS, mpl_root);
	if (proc_dev == NULL) {
		mpl_print(0, MPL_ERR "Could not create " EXCLUSION_NODE "\n");
		rv = -ENOMEM;
		goto out_rm_version;
	}
	proc_dev->owner = THIS_MODULE;
	proc_dev->proc_fops = &exclusion_fops;

	goto out;

out_rm_version:
	remove_proc_entry(VERSION_NODE, mpl_root);
out_rm_stats_tgt:
	remove_proc_entry(STATS_TGT_NODE, mpl_root);
out_rm_stats:
	remove_proc_entry(STATS_NODE, mpl_root);
out_rm_control:
	remove_proc_entry(CONTROL_NODE, mpl_root);
out_rm_config:
	remove_proc_entry(CONFIG_NODE, mpl_root);
out_rm_root:
	remove_proc_entry("scsi/" MPL_MODULE_NAME, 0);
out:
	return rv;
}

extern void mpl_remove_proc(void)
{
	remove_proc_entry(EXCLUSION_NODE, mpl_root);
	remove_proc_entry(VERSION_NODE, mpl_root);
	remove_proc_entry(DEBUG_NODE, mpl_root);
	remove_proc_entry(STATS_TGT_NODE, mpl_root);
	remove_proc_entry(STATS_NODE, mpl_root);
	remove_proc_entry(CONTROL_NODE, mpl_root);
	remove_proc_entry(CONFIG_NODE, mpl_root);
	remove_proc_entry("scsi/" MPL_MODULE_NAME, 0);
}
