/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2008 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_stub.c 1081 2006-07-31 17:14:23Z jinfante $
 */

#include <linux/version.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>
#include <linux/smp_lock.h>

#include "mpl.h"
#include "mpl_mem.h"
#include "mpl_log.h"
#include "mpl_crtn.h"
#include "mpl_version.h"

extern struct list_head mpl_hosts;
extern struct list_head deferral_queue;
extern struct list_head eh_deferral_queue;
extern struct list_head int_scsi_tmo_queue;

#ifdef MPL_SCSICMD_SANITY
/* Global queue for any new incoming scsi command */
extern struct list_head scsiglobal_queue;
#endif

/*
 * This file contains the following routines for IO stub handling.
 * 
 * For mpl_host lpfc_queue handling:
 * mpl_find_stub_in_lpfcq(struct io_stub *item)
 * mpl_find_scsi_cmd_in_lpfcq(struct scsi_cmnd *cmd, int flag)
 * mpl_add_stub_to_lpfcq(struct io_stub *stub)
 * mpl_remove_stub_from_lpfcq(struct io_stub *stub)
 * mpl_get_first_lpfcq(struct mpl_host *host)
 * 
 * For mpl_device pending_ios handling:
 * mpl_find_stub_in_pendq(struct io_stub *item)
 * mpl_find_scsi_cmd_in_pendq(struct scsi_cmnd *cmd, int flag)
 * mpl_add_stub_to_pendq(struct io_stub *stub)
 * mpl_remove_stub_from_pendq(struct io_stub *stub)
 * mpl_get_first_pendq(struct mpl_device *device)
 * 
 * For mpl module deferral list handling (completions):
 * mpl_find_stub_in_deferral(struct io_stub *item)
 * mpl_find_scsi_cmd_in_deferral(struct scsi_cmnd *cmd, int flag)
 * mpl_add_stub_to_deferral(struct io_stub *stub)
 * mpl_remove_stub_from_deferral(struct io_stub *stub)
 * mpl_get_first_deferral(void)
 * 
 * For mpl module eh_deferral list handling (timeouts):
 * mpl_find_stub_in_eh_deferral(struct io_stub *item)
 * mpl_find_scsi_cmd_in_eh_deferral(struct scsi_cmnd *cmd, int flag)
 * mpl_add_stub_to_eh_deferral(struct io_stub *stub)
 * mpl_remove_stub_from_eh_deferral(struct io_stub *stub)
 * mpl_get_first_eh_deferral(void)
 *
 * Misc routines:
 * mpl_find_stub(struct scsi_cmnd *cmd)
 * mpl_match_cmnd_remove_stub(struct scsi_cmnd *cmd, int flag)
 * mpl_find_scsi_cmd_in_anyq(struct scsi_cmnd *cmd, int flag)
 */

struct io_stub *
mpl_find_stub_in_lpfcq(struct io_stub *item)
{
	struct io_stub *stub;
	struct mpl_host *host;

	list_for_each_entry(host, &mpl_hosts, host_list) {
		list_for_each_entry(stub, &host->lpfc_queue, stub_queue) {
			if (stub->state != STUB_LPFCQ) {
				mpl_print(0, MPL_ERR "Bad State: stub_in_lpfcq:"
				" %p, %s %s, cmd = %02x %02x %02x %02x\n",
		       		stub,
		       		mpl_print_device(stub->path->device),
		       		mpl_print_path(stub->path),
		       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
			}
			if (stub == item) {
				return stub;
			}
		}
	}
	return NULL;
}

struct io_stub *
mpl_find_scsi_cmd_in_lpfcq(struct scsi_cmnd *cmd, int flag)
{
	struct io_stub *stub;
	struct mpl_host *host;

	list_for_each_entry(host, &mpl_hosts, host_list) {
		list_for_each_entry(stub, &host->lpfc_queue, stub_queue) {
			if (stub->state != STUB_LPFCQ) {
				mpl_print(0, MPL_ERR "Bad State: cmnd_in_lpfcq:"
				" %p, %s %s, cmd = %02x %02x %02x %02x\n",
		       		stub,
		       		mpl_print_device(stub->path->device),
		       		mpl_print_path(stub->path),
		       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
			}
			if ((flag == MPL_SCSI_CMND) && (stub->cmdp == cmd)) {
				return stub;
			}
			if ((flag == MPL_INTERNAL_CMND) && (&stub->cmd == cmd)) {
				return stub;
			}
		}
	}
	return NULL;
}

void
mpl_add_stub_to_lpfcq(struct io_stub *stub)
{

#ifdef MPL_STUB_SANITY
	if (mpl_find_stub_in_lpfcq(stub)) {
		mpl_print(0, MPL_ERR "Duplicate: add_to_lpfcq: %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub,
		       mpl_print_device(stub->path->device),
		       mpl_print_path(stub->path),
		       stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	if (stub->state != STUB_DQ) {
		mpl_print(0, MPL_ERR "Bad State: add_to_lpfcq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
#endif
	stub->state = STUB_LPFCQ;
#ifdef MPL_SCSICMD_SANITY
	if (stub->flag & MPL_SCSI_CMD_STUB) {
		stub->history[stub->his_ind] = stub->reason;
		stub->his_ind++ ;
		stub->history[stub->his_ind] = stub->state;
		stub->his_ind++ ;
			
	}	
#endif
	list_add_tail(&stub->stub_queue, &stub->path->host->lpfc_queue);
	return;
}

void
mpl_remove_stub_from_lpfcq(struct io_stub *stub)
{
	if (stub->state != STUB_LPFCQ) {
		mpl_print(0, MPL_ERR "Bad State: rm_from_lpfcq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	stub->state = STUB_DQ;
	list_del(&stub->stub_queue);
	return;
}

struct io_stub *
mpl_get_first_lpfcq(struct mpl_host *host)
{
        struct list_head *first;
	struct mpl_device *device;
        struct io_stub *stub = NULL;

        first = list_first(&host->lpfc_queue);
        if (first) {
                stub = list_entry(first, struct io_stub, stub_queue);
		device = stub->path->device;
		if (stub->state != STUB_LPFCQ) {
			mpl_print(0, MPL_ERR "Bad State: stub_first_lpfcq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
			stub = NULL;
		}
		else {
			stub->state = STUB_DQ;
                	list_del(&stub->stub_queue);
		}
        }
        return stub;
}

struct io_stub *
mpl_find_stub_in_pendq(struct io_stub *item)
{
	struct io_stub *stub;
	struct mpl_host *host;
	struct mpl_device *device;

	list_for_each_entry(host, &mpl_hosts, host_list) {
		list_for_each_entry(device, &host->devices, devices) {
			list_for_each_entry(stub, &device->pending_ios, stub_queue) {
				if (stub->state != STUB_PENDQ) {
					mpl_print(0, MPL_ERR "Bad State: stub_in_pendq:"
					" %p, %s %s, cmd = %02x %02x %02x %02x\n",
		       			stub,
		       			mpl_print_device(stub->path->device),
		       			mpl_print_path(stub->path),
		       			stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       			stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
				}
				if (stub == item) {
					return stub;
				}
			}
		}
	}
	return NULL;
}

struct io_stub *
mpl_find_scsi_cmd_in_pendq(struct scsi_cmnd *cmd, int flag)
{
	struct io_stub *stub;
	struct mpl_host *host;
	struct mpl_device *device;

	list_for_each_entry(host, &mpl_hosts, host_list) {
		list_for_each_entry(device, &host->devices, devices) {
			list_for_each_entry(stub, &device->pending_ios, stub_queue) {
				if (stub->state != STUB_PENDQ) {
					mpl_print(0, MPL_ERR "Bad State: cmnd_in_pendq:"
					" %p, %s %s, cmd = %02x %02x %02x %02x\n",
		       			stub,
		       			mpl_print_device(stub->path->device),
		       			mpl_print_path(stub->path),
		       			stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       			stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
				}
				if ((flag == MPL_SCSI_CMND) && (stub->cmdp == cmd)) {
					return stub;
				}
				if ((flag == MPL_INTERNAL_CMND) && (&stub->cmd == cmd)) {
					return stub;
				}
			}
		}
	}
	return NULL;
}

void
mpl_add_stub_to_pendq(struct io_stub *stub)
{
	struct mpl_device *device;

#ifdef MPL_STUB_SANITY
	if (mpl_find_stub_in_pendq(stub)) {
		mpl_print(0, MPL_ERR "Duplicate: add_to_pendq: %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub,
		       mpl_print_device(stub->path->device),
		       mpl_print_path(stub->path),
		       stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	if (stub->state != STUB_DQ) {
		mpl_print(0, MPL_ERR "Bad State: add_to_pendq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
#endif
	device = stub->path->device;
	stub->state = STUB_PENDQ;
#ifdef MPL_SCSICMD_SANITY
	if (stub->flag & MPL_SCSI_CMD_STUB) {
		stub->history[stub->his_ind] = stub->reason;
		stub->his_ind++ ;
		stub->history[stub->his_ind] = stub->state;
		stub->his_ind++ ;
			
	}
#endif	
	list_add_tail(&stub->stub_queue, &device->pending_ios);
	atomic_inc(&device->pending_count);
	return;
}

void
mpl_remove_stub_from_pendq(struct io_stub *stub)
{
	struct mpl_device *device;

	device = stub->path->device;
	if (stub->state != STUB_PENDQ) {
		mpl_print(0, MPL_ERR "Bad State: rm_from_pendq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	atomic_dec(&device->pending_count);
	stub->state = STUB_DQ;
	list_del(&stub->stub_queue);
	return;
}

struct io_stub *
mpl_get_first_pendq(struct mpl_device *device)
{
        struct list_head *first;
        struct io_stub *stub = NULL;

        first = list_first(&device->pending_ios);
        if (first) {
                stub = list_entry(first, struct io_stub, stub_queue);
		if (stub->state != STUB_PENDQ) {
			mpl_print(0, MPL_ERR "Bad State: stub_first_pendq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
			stub = NULL;
		}
		else {
			atomic_dec(&device->pending_count);
			stub->state = STUB_DQ;
                	list_del(&stub->stub_queue);
		}
        }
        return stub;
}

#ifdef MPL_SCSICMD_SANITY
struct io_stub *
mpl_find_stub_in_scsiglobal(struct io_stub *item)
{
	struct io_stub *stub;

	list_for_each_entry(stub, &scsiglobal_queue, global_queue) {
		if (stub == item) {
			return stub;
		}
	}
	return NULL;
}

void
mpl_add_stub_to_scsiglobal(struct io_stub *stub)
{
#ifdef MPL_STUB_SANITY
	if (mpl_find_stub_in_scsiglobal(stub)) {
		mpl_print(0, MPL_ERR "Duplicate: add_to_Global: %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub,
		       mpl_print_device(stub->path->device),
		       mpl_print_path(stub->path),
		       stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
#endif
	list_add_tail(&stub->global_queue, &scsiglobal_queue);
	return;
}

void
mpl_remove_stub_from_scsiglobal(struct io_stub *stub)
{
	list_del(&stub->global_queue);
	return;
}

#endif

struct io_stub *
mpl_find_stub_in_deferral(struct io_stub *item)
{
	struct io_stub *stub;

	list_for_each_entry(stub, &deferral_queue, stub_queue) {
		if (stub->state != STUB_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: stub_in_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
		if (stub == item) {
			return stub;
		}
	}
	return NULL;
}

struct io_stub *
mpl_find_scsi_cmd_in_deferral(struct scsi_cmnd *cmd, int flag)
{
	struct io_stub *stub;

	list_for_each_entry(stub, &deferral_queue, stub_queue) {
		if (stub->state != STUB_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: cmnd_in_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
		if ((flag == MPL_SCSI_CMND) && (stub->cmdp == cmd)) {
			return stub;
		}
		if ((flag == MPL_INTERNAL_CMND) && (&stub->cmd == cmd)) {
			return stub;
		}
	}
	return NULL;
}

void
mpl_add_stub_to_deferral(struct io_stub *stub)
{
	struct mpl_device *device;

#ifdef MPL_STUB_SANITY
	if (mpl_find_stub_in_deferral(stub)) {
		mpl_print(0, MPL_ERR "Duplicate: add_to_deferq: %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub,
		       mpl_print_device(stub->path->device),
		       mpl_print_path(stub->path),
		       stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	if (stub->state != STUB_DQ) {
		mpl_print(0, MPL_ERR "Bad State: add_to_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
#endif
	device = stub->path->device;
	stub->state = STUB_DEFERQ;
#ifdef MPL_SCSICMD_SANITY
	if (stub->flag & MPL_SCSI_CMD_STUB) {
		stub->history[stub->his_ind] = stub->reason;
		stub->his_ind++ ;
		stub->history[stub->his_ind] = stub->state;
		stub->his_ind++ ;
			
	}	
#endif
	atomic_inc(&device->deferral_count);
	list_add_tail(&stub->stub_queue, &deferral_queue);
	return;
}

void
mpl_remove_stub_from_deferral(struct io_stub *stub)
{
	struct mpl_device *device;

	if (stub->state != STUB_DEFERQ) {
		mpl_print(0, MPL_ERR "Bad State: rm_from_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	device = stub->path->device;
	atomic_dec(&device->deferral_count);
	stub->state = STUB_DQ;
	list_del(&stub->stub_queue);
	return;
}

struct io_stub *
mpl_get_first_deferral(void)
{
        struct list_head *first;
	struct mpl_device *device;
        struct io_stub *stub = NULL;

        first = list_first(&deferral_queue);
        if (first) {
                stub = list_entry(first, struct io_stub, stub_queue);
		device = stub->path->device;
		if (stub->state != STUB_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: stub_first_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
			stub = NULL;
		}
		else {
			/* decrement deferral_count in caller */
			stub->state = STUB_DQ;
			atomic_dec(&device->deferral_count);
                	list_del(&stub->stub_queue);
		}
        }
        return stub;
}


struct io_stub *
mpl_find_stub_in_hbeat_deferral(struct io_stub *item)
{
	struct io_stub *stub;
	struct mpl_host *host;

	list_for_each_entry(host, &mpl_hosts, host_list) {
		list_for_each_entry(stub, &host->hbeat_deferral_queue, stub_queue) {
			if (stub->state != STUB_HBEAT_DEFERQ) {
				mpl_print(0, MPL_ERR "Bad State: stub_in_DEFERQ:"
				" %p, %s %s, cmd = %02x %02x %02x %02x\n",
		       		stub,
		       		mpl_print_device(stub->path->device),
		       		mpl_print_path(stub->path),
		       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
			}
			if (stub == item) {
				return stub;
			}
		}
	}
	return NULL;
}


void
mpl_add_stub_to_hbeat_deferral(struct io_stub *stub)
{

#ifdef MPL__STUB_SANITY
	if (mpl_find_stub_in_hbeat_deferral(stub)) {
		mpl_print(0, MPL_ERR "Duplicate: add_to_hbeat_deferral: %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub,
		       mpl_print_device(stub->path->device),
		       mpl_print_path(stub->path),
		       stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	if (stub->state != STUB_DQ) {
		mpl_print(0, MPL_ERR "Bad State: add_to_hbeat_deferral:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
#endif
	stub->state = STUB_HBEAT_DEFERQ;
	list_add_tail(&stub->stub_queue, &stub->path->host->hbeat_deferral_queue);
	return;
}

void
mpl_remove_stub_from_hbeat_deferral(struct io_stub *stub)
{
	if (stub->state != STUB_HBEAT_DEFERQ) {
		mpl_print(0, MPL_ERR "Bad State: rm_from_hbeat_deferral:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	stub->state = STUB_DQ;
	list_del(&stub->stub_queue);
	return;
}

struct io_stub *
mpl_get_first_hbeat_deferral(struct mpl_host *host)
{
        struct list_head *first;
	struct mpl_device *device;
        struct io_stub *stub = NULL;

        first = list_first(&host->hbeat_deferral_queue);
        if (first) {
                stub = list_entry(first, struct io_stub, stub_queue);
		device = stub->path->device;
		if (stub->state != STUB_HBEAT_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: stub_first_hbeat_deferral:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
			stub = NULL;
		}
		else {
			stub->state = STUB_DQ;
                	list_del(&stub->stub_queue);
		}
        }
        return stub;
}

struct io_stub *
mpl_find_stub_in_eh_deferral(struct io_stub *item)
{
	struct io_stub *stub;

	list_for_each_entry(stub, &eh_deferral_queue, stub_queue) {
		if (stub->state != STUB_EH_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: stub_in_eh_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
		if (stub == item) {
			return stub;
		}
	}
	return NULL;
}

struct io_stub *
mpl_find_scsi_cmd_in_eh_deferral(struct scsi_cmnd *cmd, int flag)
{
	struct io_stub *stub;

	list_for_each_entry(stub, &eh_deferral_queue, stub_queue) {
		if (stub->state != STUB_EH_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: cmnd_in_eh_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
		if ((flag == MPL_SCSI_CMND) && (stub->cmdp == cmd)) {
			return stub;
		}
		if ((flag == MPL_INTERNAL_CMND) && (&stub->cmd == cmd)) {
			return stub;
		}
	}
	return NULL;
}

struct io_stub *
mpl_find_devloss_tmo_in_eh_deferral(uint64_t tgt_wwpn)
{
	struct io_stub *stub;

	list_for_each_entry(stub, &eh_deferral_queue, stub_queue) {
		if (stub->state != STUB_EH_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: cmnd_in_eh_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
			stub,
			mpl_print_device(stub->path->device),
			mpl_print_path(stub->path),
			stub->cmd.cmnd[0], stub->cmd.cmnd[1],
			stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
		if (( stub->reason == MPL_DO_DEVLOSS_TMO) && 
			(wwn_to_u64(stub->path->tgt->wwpn)== tgt_wwpn )) {
			return stub;
		}
	}
	return NULL;
}

void
mpl_add_stub_to_eh_deferral(struct io_stub *stub)
{
	struct mpl_device *device;

#ifdef MPL_STUB_SANITY
	if (mpl_find_stub_in_eh_deferral(stub)) {
		mpl_print(0, MPL_ERR "Duplicate: add_to_eh_deferq: %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub,
		       mpl_print_device(stub->path->device),
		       mpl_print_path(stub->path),
		       stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	if (stub->state != STUB_DQ) {
		mpl_print(0, MPL_ERR "Bad State: add_to_eh_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
#endif
	device = stub->path->device;
	stub->state = STUB_EH_DEFERQ;
#ifdef MPL_SCSICMD_SANITY
	if (stub->flag & MPL_SCSI_CMD_STUB) {
		stub->history[stub->his_ind] = stub->reason;
		stub->his_ind++ ;
		stub->history[stub->his_ind] = stub->state;
		stub->his_ind++ ;
			
	}	
#endif
	atomic_inc(&device->eh_deferral_count);
	list_add_tail(&stub->stub_queue, &eh_deferral_queue);
	return;
}

void
mpl_remove_stub_from_eh_deferral(struct io_stub *stub)
{
	struct mpl_device *device;

	if (stub->state != STUB_EH_DEFERQ) {
		mpl_print(0, MPL_ERR "Bad State: rm_from_eh_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	device = stub->path->device;
	atomic_dec(&device->eh_deferral_count);
	stub->state = STUB_DQ;
	list_del(&stub->stub_queue);
	return;
}

struct io_stub *
mpl_get_first_eh_deferral(void)
{
        struct list_head *first;
	struct mpl_device *device;
        struct io_stub *stub = NULL;

        first = list_first(&eh_deferral_queue);
        if (first) {
                stub = list_entry(first, struct io_stub, stub_queue);
		device = stub->path->device;
		if (stub->state != STUB_EH_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: stub_first_eh_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
			stub = NULL;
		}
		else {
			/* decrement eh_deferral_count in caller */
			atomic_dec(&device->eh_deferral_count);	
			stub->state = STUB_DQ;
                	list_del(&stub->stub_queue);
		}
        }
        return stub;
}

struct io_stub *
mpl_find_stub_in_scsitmo_deferral(struct io_stub *item)
{
	struct io_stub *stub;

	list_for_each_entry(stub, &int_scsi_tmo_queue, stub_queue) {
		if (stub->state != STUB_SCSITMO_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: stub_in_eh_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
		if (stub == item) {
			return stub;
		}
	}
	return NULL;
}

struct io_stub *
mpl_get_first_scsitmo_deferral(void)
{
	struct list_head *first;
	struct io_stub *stub = NULL;

	first = list_first(&int_scsi_tmo_queue);
	if (first)
		stub = list_entry(first, struct io_stub, stub_queue);

	return stub;
}

void
mpl_add_stub_to_scsitmo_deferral(struct io_stub *stub)
{
	struct mpl_device *device;

#ifdef MPL_STUB_SANITY
	if (mpl_find_stub_in_scsitmo_deferral(stub)) {
		mpl_print(0, MPL_ERR "Duplicate: add_to_eh_deferq: %p, "
				  "%s %s, cmd = %02x %02x %02x %02x\n",
		       stub,
		       mpl_print_device(stub->path->device),
		       mpl_print_path(stub->path),
		       stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
	if (stub->state != STUB_DQ) {
		mpl_print(0, MPL_ERR "Bad State: add_to_eh_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		return;
	}
#endif
	device = stub->path->device;
	stub->state = STUB_SCSITMO_DEFERQ;
#ifdef MPL_SCSICMD_SANITY
	if (stub->flag & MPL_SCSI_CMD_STUB) {
		stub->history[stub->his_ind] = stub->reason;
		stub->his_ind++ ;
		stub->history[stub->his_ind] = stub->state;
		stub->his_ind++ ;
			
	}	
#endif
	/* atomic_inc(&device->eh_deferral_count); */
	list_add_tail(&stub->stub_queue, &int_scsi_tmo_queue);
	return;
}

struct io_stub *
mpl_match_cmnd_remove_from_scsitmo_deferral(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;

	list_for_each_entry(stub, &int_scsi_tmo_queue, stub_queue) {
		if (stub->state != STUB_SCSITMO_DEFERQ) {
			mpl_print(0, MPL_ERR "Bad State: cmnd_in_eh_deferq:"
			" %p, %s %s, cmd = %02x %02x %02x %02x\n",
	       		stub,
	       		mpl_print_device(stub->path->device),
	       		mpl_print_path(stub->path),
	       		stub->cmd.cmnd[0], stub->cmd.cmnd[1],
	       		stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
		if (stub->cmdp == cmd) {
			stub->state = STUB_DQ;
			list_del(&stub->stub_queue);
			return stub;
		}
	}
	return NULL;
}

/* Given an mpl module cmnd, calculate the
 * corresponding stub pointer.
 */
struct io_stub *
mpl_find_stub(struct scsi_cmnd *cmd)
{
	struct io_stub *stub;

	stub = mpl_locate_stub(cmd);
	if (stub->magic != MPL_STUB_MAGIC) {
		mpl_print(0, MPL_ERR "No stub found: %p from  %p, %d/%d/%d "
		  "cmd = %02x %02x %02x %02x\n",
		  cmd, (&cmd)[-1],
		  mpl_host_no_from_cmd(cmd),
		  mpl_target_no_from_cmd(cmd),
		  mpl_lun_no_from_cmd(cmd),
		  cmd->cmnd[0], cmd->cmnd[1], cmd->cmnd[2], cmd->cmnd[3]);
		return 0;
	}
	return stub;
}

struct io_stub *
mpl_match_cmnd_remove_stub(struct scsi_cmnd *cmd, int flag)
{
	struct mpl_host *host;
	struct mpl_device *device;
	struct io_stub *stub, *next_stub;
	int  i;
	int  rm = 0;
	int ignore_aborted_cmd = flag & MPL_ABORT_IGNORE_ABORTED;

	/* NOTE: cmd must match the original cmd, NOT the copy since this routine
	 * is called from the abort, device reset and bus reset handlers.
	 */
	flag &= MPL_ABORT_TYPE_MASK;

	/* First check pending_ios queue for ALL devices */
	for (i = 0; ; i++) {
		host = mpl_find_host(i);
		if (host == NULL) {
			break;
		}

		if (flag == MPL_ABORT_HOST &&
		    mpl_scsi_host_from_cmd(cmd) == host->scsi_host &&
		    cmd->device->host == host->scsi_host) {
			rm = 1; /* remove everything on this host */
		}

		list_for_each_entry(device, &host->devices, devices) {
			list_for_each_entry_safe(stub, next_stub, &device->pending_ios, stub_queue) {
#ifdef MPL_STUB_SANITY
				if ((stub->flag & MPL_INTERNAL_CMD_STUB)  ||
					(!(stub->cmdp))) {
					mpl_print(0, MPL_ERR
					"Internal Cmd in pending_ios %p"
				  	"%s %s, cmd = %02x %02x %02x %02x\n",
		       			stub, 
		       			mpl_print_device(stub->path->device),
		       			mpl_print_path(stub->path),
		       			stub->cmd.cmnd[0], stub->cmd.cmnd[1],
		       			stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
					continue;
				}
#endif
				if ((flag == MPL_ABORT_DEVICE) &&
		    		   (cmd->device == stub->cmdp->device)) {
					rm = 1; /* remove everything on this device */
				}
				if ((flag == MPL_ABORT_EXACT) &&
		    			(cmd == stub->cmdp)) {
					rm = 1; /* remove this specific IO */
				}
				if(rm) {
					if (stub->state != STUB_PENDQ) {
						mpl_print(0, MPL_ERR
						"Bad State: stub_match_pendq:"
						" %p, %s %s, "
						"cmd = %02x %02x %02x %02x\n",
		       				stub,
		       				mpl_print_device(device),
		       				mpl_print_path(stub->path),
		       				stub->cmd.cmnd[0],
						stub->cmd.cmnd[1],
		       				stub->cmd.cmnd[2],
						stub->cmd.cmnd[3]);
						stub = NULL;
					}
					else {
						atomic_dec(&device->pending_count);
						stub->state = STUB_DQ;
						list_del(&stub->stub_queue);
					}
					return stub;
				}
			}
		}
	}

	/* Next lets check if its on the deferral list */
	list_for_each_entry_safe(stub, next_stub, &deferral_queue, stub_queue) {
		if (stub->flag & MPL_INTERNAL_CMD_STUB) {
			continue;  /* skip internal commands */
		}
#ifdef MPL_STUB_SANITY
		if (!(stub->cmdp)) {
			mpl_print(0, MPL_ERR "Bad stub in pending_ios:2: %p from  %p, "
		  			"%s %s, cmd = %02x %02x %02x %02x\n",
       			stub, (&stub)[-1],
       			mpl_print_device(stub->path->device),
       			mpl_print_path(stub->path),
       			stub->cmd.cmnd[0], stub->cmd.cmnd[1],
       			stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
		}
#endif
		/* skip already aborted commands */
		if ((ignore_aborted_cmd) && 
		    (stub->reason == MPL_ABORTED_CMD)) { 
			continue;
		}
		if ((flag == MPL_ABORT_DEVICE) &&
    		   (cmd->device == stub->cmdp->device)) {
			rm = 1; /* remove everything on this device */
		}
		if ((flag == MPL_ABORT_EXACT) &&
    			(cmd == stub->cmdp)) {
			rm = 1; /* remove this specific IO */
		}
		if(rm) {
			if (stub->state != STUB_DEFERQ) {
				mpl_print(0, MPL_ERR
				"Bad State: stub_match_deferq:"
				" %p, %s %s, cmd = %02x %02x %02x %02x\n",
       				stub,
       				mpl_print_device(stub->path->device),
       				mpl_print_path(stub->path),
       				stub->cmd.cmnd[0], stub->cmd.cmnd[1],
       				stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
				stub = NULL;
			}
			else {
				atomic_dec(&stub->path->device->deferral_count);
				stub->state = STUB_DQ;
				list_del(&stub->stub_queue);
			}
			return stub;
		}
	}

	/* Next lets check if its on the eh_deferral list */
	list_for_each_entry_safe(stub, next_stub, &eh_deferral_queue, stub_queue) {
		if (stub->flag & MPL_INTERNAL_CMD_STUB) {
			continue;  /* skip internal commands */
		}
		if ((stub->reason == MPL_DO_INT_SCSI_TMO ) &&
			((flag == MPL_ABORT_DEVICE ) || 
			(flag == MPL_ABORT_HOST))) {
			continue; /* skip command which timeout internally */
		}  
#ifdef MPL_STUB_SANITY
		if (!(stub->cmdp)) {
			mpl_print(0, MPL_ERR "Bad stub in pending_ios:3: %p from  %p, "
		  			"%s %s, cmd = %02x %02x %02x %02x\n",
       			stub, (&stub)[-1],
       			mpl_print_device(stub->path->device),
       			mpl_print_path(stub->path),
       			stub->cmd.cmnd[0], stub->cmd.cmnd[1],
       			stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
			continue;
		}
#endif
		if ((flag == MPL_ABORT_DEVICE) &&
    		   (cmd->device == stub->cmdp->device)) {
				rm = 1; /* remove everything on this device */
		}
		if ((flag == MPL_ABORT_EXACT) &&
    			(cmd == stub->cmdp)) {
			rm = 1; /* remove this specific IO */
		}
		if (rm) {
			if (stub->state != STUB_EH_DEFERQ) {
				mpl_print(0, MPL_ERR
				"Bad State: stub_match_eh_deferq:"
				" %p, %s %s, cmd = %02x %02x %02x %02x\n",
       				stub,
       				mpl_print_device(stub->path->device),
       				mpl_print_path(stub->path),
       				stub->cmd.cmnd[0], stub->cmd.cmnd[1],
       				stub->cmd.cmnd[2], stub->cmd.cmnd[3]);
				stub = NULL;
			}
			else {
				atomic_dec(&stub->path->device->eh_deferral_count);
				stub->state = STUB_DQ;
				list_del(&stub->stub_queue);
			}
			return stub;
		}
	}
	return NULL;
}

struct io_stub *
mpl_find_scsi_cmd_in_anyq(struct scsi_cmnd *cmd, int flag)
{
	struct io_stub *stub;

	stub = mpl_find_scsi_cmd_in_lpfcq(cmd, flag);
	if (stub)
		return stub;
	stub = mpl_find_scsi_cmd_in_pendq(cmd, flag);
	if (stub)
		return stub;
	stub = mpl_find_scsi_cmd_in_deferral(cmd, flag);
	if (stub)
		return stub;
	stub = mpl_find_scsi_cmd_in_eh_deferral(cmd, flag);
	if (stub)
		return stub;
	return NULL;
}
