
/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/
/* 
 * Note:
 * The following have adapted from udev (file scsi_serial.c) 
 * The driver needs to reproduce exactly the serial number of the devices
 * reported by udev for the application to match them later on
*/

#include <linux/version.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>
#include <linux/smp_lock.h>

#include "mpl.h"
#include "mpl_log.h"
#include "mpl_crtn.h"

/*
 * A priority based list of id, naa, and binary/ascii for the identifier
 * descriptor in VPD page 0x83.
 *
 * Brute force search for a match starting with the first value in the
 * following id_search_list. This is not a performance issue, since there
 * is normally one or some small number of descriptors.
 */
struct scsi_id_search_values mpl_id_search_list[] = {
	{ SCSI_ID_NAA,	SCSI_ID_NAA_IEEE_REG_EXTENDED,	SCSI_ID_BINARY },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_IEEE_REG_EXTENDED,	SCSI_ID_ASCII },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_IEEE_REG,	SCSI_ID_BINARY },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_IEEE_REG,	SCSI_ID_ASCII },
	/*
	 * Devices already exist using NAA values that are now marked
	 * reserved. These should not conflict with other values, or it is
	 * a bug in the device. As long as we find the IEEE extended one
	 * first, we really don't care what other ones are used. Using
	 * don't care here means that a device that returns multiple
	 * non-IEEE descriptors in a random order will get different
	 * names.
	 */
	{ SCSI_ID_NAA,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_BINARY },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_ASCII },
	{ SCSI_ID_EUI_64,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_BINARY },
	{ SCSI_ID_EUI_64,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_ASCII },
	{ SCSI_ID_T10_VENDOR,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_BINARY },
	{ SCSI_ID_T10_VENDOR,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_ASCII },
	{ SCSI_ID_VENDOR_SPECIFIC,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_BINARY },
	{ SCSI_ID_VENDOR_SPECIFIC,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_ASCII },
};

int mpl_check_fill_0x83_id(char *page_83, 
			const struct scsi_id_search_values *id_search, 
			char *serial, int *serial_size)
{
	int i, j, len;

	/*
	 * ASSOCIATION must be with the device (value 0)
	 */
	if ((page_83[1] & 0x30) != 0)
		return 1;

	if ((page_83[1] & 0x0f) != id_search->id_type)
		return 2;

	/*
	 * Possibly check NAA sub-type.
	 */
	if ((id_search->naa_type != SCSI_ID_NAA_DONT_CARE) &&
	    (id_search->naa_type != (page_83[4] & 0xf0) >> 4))
		return 3;

	/*
	 * Check for matching code set - ASCII or BINARY.
	 */
	if ((page_83[0] & 0x0f) != id_search->code_set)
		return 4;
	/*
	 * page_83[3]: identifier length
	 */
	len = page_83[3];
	i = 4; /* offset to the start of the identifier */
	j = 0;

	/* Internally store validation id as BINARY */
	while (i < (4 + page_83[3])){
		serial[j++] = page_83[i++];
	}
	*serial_size = len;
	return 0;
}

/* Return 0 if we find a valid validation id, else return 1 */
int mpl_get_validation_id(char *page_83, char *serial, int *serial_size)
{
	return mpl_get_scsi_page83_serial(page_83, mpl_id_search_list,
		sizeof(mpl_id_search_list)/sizeof(mpl_id_search_list[0]),
		serial, serial_size);
}

/* Get device identification VPD page */
int mpl_get_scsi_page83_serial(char *page_83,
	struct scsi_id_search_values *list, int list_size,
	char *serial, int *serial_size)
{
	int retval = 0;
	unsigned int id_ind, j, count = 0;

	/*
	 * Search for a match in the prioritized id_search_list.
	 */
	for (id_ind = 0;
	     id_ind < list_size;
	     id_ind++) {
		/*
		 * Examine each descriptor returned. There is normally only
		 * one or a small number of descriptors.
		 */
		for (j = 4; j <= (unsigned int)page_83[3] + 3; j += page_83[j + 3] + 4) {
			retval = mpl_check_fill_0x83_id(&page_83[j],
						    &list[id_ind],
						    serial, serial_size);
			count ++;
			if (!retval) {
				return retval;
			}
		}
	}
	return 1;
}
