/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_util.c 1342 2007-05-10 21:31:12Z bsebastian $
 */

#include <linux/version.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>

#include <linux/ctype.h>
#include <linux/string.h>

#include "mpl.h"
#include "mpl_mem.h"
#include "mpl_log.h"
#include "mpl_crtn.h"

extern int   (* mpl_base_slave_alloc)(struct scsi_device *sdev);
extern int (* mpl_base_abort_handler)(struct scsi_cmnd *);

extern spinlock_t stub_queue_lock;
extern atomic_t total_internal_count;
extern volatile int driver_unloading;

extern int mpl_data_int_cmd;

int
mpl_vendor_hp_check(struct scsi_cmnd *cmd)
{
	uint8_t hp_str[] = "HP";
	uint8_t compaq_str[] = "COMPAQ";
	uint8_t *request;
	uint8_t *vendor;
	int len = 0;

	request = mpl_find_cmnd_data(cmd, &len);
	if ((request == NULL) || (len < 36)) {
		mpl_print(0, MPL_ERR "No data in vendor check: len %d\n", len);
		return 1;
	}
	vendor = request + 8;

	if ((memcmp(vendor, hp_str, strlen(hp_str)-1) == 0) ||
	    (memcmp(vendor, compaq_str, strlen(compaq_str)-1) == 0)) {
		mpl_release_cmnd_data(cmd);
		return 1;
	}
	mpl_release_cmnd_data(cmd);
	return 0;
}

void
mpl_restart_sched_cmd(struct mpl_path *path, unsigned long expires)
{
	if (!driver_unloading && path->state != PATH_DISABLED) {
		mpl_module_unlock();
		mod_timer(&path->sched_cmd, expires);
		mpl_module_lock();
	}
}

void
mpl_set_cmd_path(struct scsi_cmnd *cmd, struct mpl_path *path)
{
	cmd->device = path->sdev;
	return;
}

void
mpl_cmd_timeout(unsigned long data)
{
	struct io_stub *stub = (struct io_stub *)data;
	struct io_stub *s;
	unsigned long flags;

	mpl_module_lock();

	spin_lock_irqsave(&stub_queue_lock, flags);
	s = mpl_find_stub_in_lpfcq(stub);
	if (s) {
		mpl_remove_stub_from_lpfcq(s);
	}
	else {
		s = mpl_find_stub_in_deferral(stub);
		if (s) {
			mpl_remove_stub_from_deferral(s);
		}
	}
	spin_unlock_irqrestore(&stub_queue_lock, flags);

	if (s != stub) {
		mpl_print(0, MPL_ERR "Timed out command not found, "
			  "stub %p:%p, state: %d cmd: %02x, reason: %x, %s\n",
			  stub, s, stub->state, stub->cmd.cmnd[0], stub->reason,
			  mpl_print_path(stub->path));
		mpl_module_unlock();
		return;
	}

	spin_lock_irqsave(&stub_queue_lock, flags);
	switch (stub->reason) {
	case MPL_HEARTBEAT_CMD:
		stub->reason = MPL_DO_HBEAT_TMO;
		mpl_add_stub_to_eh_deferral(stub);
		break;
	case MPL_INTERNAL_CMD:
		stub->reason = MPL_DO_INT_TMO;
		mpl_add_stub_to_eh_deferral(stub);
		break;
	default:
		mpl_print(0, MPL_ERR "Timed out command bad reason, "
			  "stub = %p, state = %d cmd = %02x, reason = %x, %s\n",
			  stub, stub->state, stub->cmd.cmnd[0], stub->reason,
			  mpl_print_path(stub->path));
	}
	spin_unlock_irqrestore(&stub_queue_lock, flags);

	mpl_module_unlock();
	mpl_wake_eh_thread();
	return;
}

void
mpl_set_cmd_timer(struct io_stub *stub, int tmo)
{
	struct scsi_cmnd *cmd = &stub->cmd;

	if (tmo) {
		cmd->timeout_per_command = jiffies + (HZ * tmo);
		cmd->eh_timeout.data = (unsigned long) stub;
		mpl_module_unlock();
		mod_timer(&cmd->eh_timeout, jiffies + HZ * tmo);
		mpl_module_lock();
	}
	return;
}

/* 
 *
 */
struct io_stub *
mpl_setup_inquiry(struct mpl_path *path, unsigned char page)
{
	struct scsi_cmnd *cmd;
	struct io_stub *stub;
	uint8_t *databuf;

	databuf = mpl_get_scsi_buf(); 
	if (!databuf)
		goto buf_alloc_err;
	mpl_data_int_cmd++;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(0, path, 0);
	if (!stub) 
		goto stub_alloc_err;
	stub->flag |= MPL_INTERNAL_CMD_STUB;
	stub->state = STUB_DQ;

	cmd = &stub->cmd;

	mpl_set_cmd_path(cmd, path);
	if (path->failures && mpl_is_path_configured(path)) {
		/* Use a shorter timeout */
		mpl_set_cmd_timer(stub, MPL_RETRY_CMD_TMO); 
	}
	else
		mpl_set_cmd_timer(stub, MPL_INQUIRY_CMD_TMO); 

	cmd->cmd_len = 6;               /* 6 byte CDB */
	cmd->cmnd[0] = INQUIRY; 
	cmd->cmnd[1] = 1;               /* evpd */
	cmd->cmnd[2] = page;            /* serial # page */
	cmd->cmnd[4] = 0xff;            /* allocation len */

	/* with memset above, use_sg and underflow are all properly setup */
	cmd->sc_data_direction = DMA_FROM_DEVICE;
	cmd->request_bufflen = 0xff;
	cmd->request_buffer = databuf;
	memset(cmd->request_buffer, 0, cmd->request_bufflen);

	return stub;

stub_alloc_err:
	mpl_data_int_cmd--;
	mpl_put_scsi_buf(databuf);
buf_alloc_err:
	return NULL;
}


struct io_stub *
mpl_setup_report_tgt_port(struct mpl_path *path)
{
	struct scsi_cmnd *cmd;
	struct io_stub *stub;
	uint8_t *databuf;

	databuf = mpl_get_scsi_buf(); 
	if (!databuf)
		goto buf_alloc_err;
	mpl_data_int_cmd++;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(0, path, 0);
	if (!stub) 
		goto stub_alloc_err;
	stub->flag |= MPL_INTERNAL_CMD_STUB;
	stub->state = STUB_DQ;

	cmd = &stub->cmd;

	mpl_set_cmd_path(cmd, path);
	if (path->failures && mpl_is_path_configured(path)) {
		/* Use a shorter timeout */
		mpl_set_cmd_timer(stub, MPL_RETRY_CMD_TMO); 
	}
	else
		mpl_set_cmd_timer(stub, MPL_INTERNAL_CMD_TMO); 

	cmd->cmd_len = 12;              /* 12 byte CDB */
	cmd->cmnd[0] = REPORT_TGT_PORT_GROUP;
	cmd->cmnd[1] = 0xa;             /* service action */
	cmd->cmnd[9] = 0xff;            /* allocation len */

	/* with memset above, use_sg and underflow are all properly setup */
	cmd->sc_data_direction = DMA_FROM_DEVICE;
	cmd->request_bufflen = 0xff;
	cmd->request_buffer = databuf;
	memset(cmd->request_buffer, 0, cmd->request_bufflen);

	return stub;

stub_alloc_err:
	mpl_data_int_cmd--;
	mpl_put_scsi_buf(databuf);
buf_alloc_err:
	return NULL;
}

struct io_stub *
mpl_setup_set_tgt_port(struct mpl_path *path,
	int port_group, int state)
{
	struct mpl_device *device = path->device;
	struct scsi_cmnd *cmd;
	struct io_stub *stub;
	uint8_t *databuf;

	databuf = mpl_get_scsi_buf(); 
	if (!databuf)
		goto buf_alloc_err;
	mpl_data_int_cmd++;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(0, path, 0);
	if (!stub) 
		goto stub_alloc_err;
	stub->flag |= MPL_INTERNAL_CMD_STUB;
	stub->state = STUB_DQ;

	cmd = &stub->cmd;

	mpl_set_cmd_path(cmd, path);
	mpl_set_cmd_timer(stub, MPL_INTERNAL_CMD_TMO); 

	cmd->cmd_len = 12;              /* 12 byte CDB */
	cmd->cmnd[0] = SET_TGT_PORT_GROUP;
	cmd->cmnd[1] = 0xa;             /* service action */
	cmd->cmnd[9] = 0x8;             /* allocation len */

	/* with memset above, use_sg and underflow are all properly setup */
	cmd->sc_data_direction = DMA_TO_DEVICE;
	cmd->request_bufflen = 0x8;
	cmd->request_buffer = databuf;
	memset(cmd->request_buffer, 0, cmd->request_bufflen);
	*(databuf+4) = (uint8_t)state;
	*(databuf+7) = (uint8_t)port_group;

	/* This is implemented in a non-standard way for DS400 */
	if (device->plugin && (memcmp(device->plugin->name, "ds400", 5) == 0)) {
		cmd->cmnd[9] = 0;             /* allocation len */
		cmd->request_bufflen = 0;
	}
	return stub;

stub_alloc_err:
	mpl_data_int_cmd--;
	mpl_put_scsi_buf(databuf);
buf_alloc_err:
	return NULL;
}

struct io_stub *
mpl_setup_startstop(struct mpl_path *path)
{
	struct scsi_cmnd *cmd;
	struct io_stub *stub;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(0, path, 0);
	if (!stub) 
		goto stub_alloc_err;
	stub->flag |= MPL_INTERNAL_CMD_STUB;
	stub->state = STUB_DQ;

	cmd = &stub->cmd;

	mpl_set_cmd_path(cmd, path);
	mpl_set_cmd_timer(stub, MPL_INTERNAL_CMD_TMO); 

	cmd->cmd_len = 6;               /* 6 byte CDB */
	cmd->cmnd[0] = START_STOP; 
	cmd->cmnd[4] = 1;               /* start */

	/* with memset above, request_bufflen, use_sg and underflow
	 * are all properly setup */
	cmd->sc_data_direction = DMA_NONE;

	return stub;

stub_alloc_err:
	return NULL;
}


struct io_stub *
mpl_setup_tur(struct mpl_path *path)
{
	struct scsi_cmnd *cmd;
	struct io_stub *stub;

	wmb();

	/* internal cmd so no done route (param 3) */
	stub = mpl_stub_alloc(0, path, 0);
	if (!stub) 
		goto stub_alloc_err;
	stub->flag |= MPL_INTERNAL_CMD_STUB;
	stub->state = STUB_DQ;

	cmd = &stub->cmd;

	mpl_set_cmd_path(cmd, path);
	if (path->failures && mpl_is_path_configured(path)) {
		/* Use a shorter timeout */
		mpl_set_cmd_timer(stub, MPL_RETRY_CMD_TMO); 
	}
	else
		mpl_set_cmd_timer(stub, MPL_TUR_CMD_TMO); 

	cmd->cmd_len = 6;               /* 6 byte CDB */
	cmd->cmnd[0] = TEST_UNIT_READY; 

	/* with memset above, request_bufflen, use_sg and underflow
	 * are all properly setup */
	cmd->sc_data_direction = DMA_NONE;
	return stub;

stub_alloc_err:
	return NULL;
}

void
mpl_send_internal_completion(struct mpl_path *path, struct io_stub *stub,
			     void (*done)(struct scsi_cmnd *))
{
	struct scsi_cmnd *cmd = &stub->cmd;
	int  rv;
	unsigned long sflag;
	unsigned long flags;
        struct Scsi_Host *shost;
	spinlock_t    *host_lock;

	if (driver_unloading || path->state == PATH_DISABLED ||
		!cmd->device) {
		goto int_cmd_not_sent;
	}

	stub->path = path;
	shost =  path->host->scsi_host;
	if (stub->reason == MPL_HEARTBEAT_CMD) {
		if (path->hbstate == HB_STOPPED) {
			goto int_cmd_not_sent;
		}
		if ( (shost->host_busy > shost->can_queue) &&
			(atomic_read(&total_internal_count) > 10)) {
			mpl_module_unlock();
			del_timer_sync(&cmd->eh_timeout);
			mpl_module_lock();
			mpl_print(MPL_INTERNAL, MPL_INFO
				"Defer cmd 0x%x for %s %s: IOs %d > %d\n",
				cmd->cmnd[0],
				mpl_print_device(path->device),
				mpl_print_path(path),
				shost->host_busy, shost->can_queue);

			if (cmd->cmnd[0] == REPORT_TGT_PORT_GROUP)  {
				path->device->tpg_cmd_inp = 0;
			}
			mpl_stub_free(stub);
			mpl_hbeat_state_machine(path, HB_BUSY);
			return;
		}
	}

	mpl_print(MPL_INTERNAL, MPL_INFO "Issue cmd 0x%x for %s %s blk %d %d\n",
		  cmd->cmnd[0],
		  mpl_print_device(path->device),
		  mpl_print_path(path),
		  (path->host->blocked || path->lld_blocked),
		  path->failures);

	spin_lock_irqsave(&stub_queue_lock, flags);
	mpl_add_stub_to_lpfcq(stub);
	spin_unlock_irqrestore(&stub_queue_lock, flags);

	if (path->host->blocked || path->lld_blocked ||
		path->device->reset_pending) {
		cmd->result = ScsiResult(DID_BUS_BUSY, SAM_STAT_BUSY);
		cmd->retries = 0;
		mpl_module_unlock();
		done(&stub->cmd);
		mpl_module_lock();
		return;
	}

	mpl_module_unlock();
	host_lock = cmd->device->host->host_lock;
	spin_lock_irqsave(host_lock, sflag);
	rv = mpl_base_queuecommand(cmd, done);
	spin_unlock_irqrestore(host_lock, sflag);
	/* If 8.xx driver return SCSI_MLQUEUE_HOST_BUSY retry the command */
	if ( rv == SCSI_MLQUEUE_HOST_BUSY) {
		cmd->result = ScsiResult(0,SAM_STAT_TASK_SET_FULL);
		done(cmd);
		rv = 0;
	}
	if (rv) {
		mpl_module_lock();
		mpl_module_unlock();
		del_timer_sync(&cmd->eh_timeout);
		mpl_module_lock();
		spin_lock_irqsave(&stub_queue_lock, flags);
		mpl_remove_stub_from_lpfcq(stub);
		spin_unlock_irqrestore(&stub_queue_lock, flags);

		if ((cmd->cmnd[0] == INQUIRY) && (cmd->cmnd[2] == 0x83)) {
			path->flag |= PATH_NEEDS_INQUIRYP0;
		} else if (cmd->cmnd[0] == TEST_UNIT_READY) {
			path->flag |= PATH_NEEDS_TUR;
		} else if (cmd->cmnd[0] == START_STOP) {
			path->flag |= PATH_NEEDS_STARTSTOP;
		}
		path->retries = cmd->retries + 1;
		mpl_restart_sched_cmd(path, jiffies + HZ);

int_cmd_not_sent:
		mpl_inc_can_queue(stub);
		mpl_stub_free(stub);
		return;
	}
	mpl_module_lock();
	return;
}

void
mpl_path_schedule_cmd(unsigned long data)
{
	struct mpl_path *path = (struct mpl_path *)data;
	struct mpl_device *device = path->device;
	struct io_stub *stub;
	unsigned long expires = jiffies + 1;
	unsigned long flags;
	int wake_it = 0;

	mpl_module_lock();
	mpl_print(MPL_TMO, MPL_INFO "Schedule (%d) Cmd flg:x%x io:%d dev:%d on %s %s\n",
		path->retries, path->flag, 
		atomic_read(&device->io_count), device->state,
		mpl_print_device(device), mpl_print_path(path));

	if ((path->flag & PATH_DEV_QUIESCE) && atomic_read(&device->io_count)) {
		/* If we need to quiesce, and there are outstanding IOs,
		 * reschedule this command for later.
		 */
		mpl_restart_sched_cmd(path, jiffies + HZ);
		mpl_module_unlock();
		return;
	}
	path->flag &= ~PATH_DEV_QUIESCE;

	if (path->flag & PATH_NEEDS_TUR) {
		stub = mpl_setup_tur(path);
		if (stub) {
			path->flag &= ~PATH_NEEDS_TUR;
			expires = 0;
			stub->cmd.retries = path->retries;
			path->retries = 0;
			spin_lock_irqsave(&stub_queue_lock, flags);
			stub->reason = MPL_INTERNAL_CMD;
			mpl_add_stub_to_deferral(stub);
			spin_unlock_irqrestore(&stub_queue_lock, flags);
			wake_it = 1;
		}
	} else if (path->flag & PATH_NEEDS_INQUIRYP0) {
		stub = mpl_setup_inquiry(path, 0);
		if (stub) {
			path->flag &= ~PATH_NEEDS_INQUIRYP0;
			expires = 0;
			stub->cmd.retries = path->retries;
			path->retries = 0;
			spin_lock_irqsave(&stub_queue_lock, flags);
			stub->reason = MPL_INTERNAL_CMD;
			mpl_add_stub_to_deferral(stub);
			spin_unlock_irqrestore(&stub_queue_lock, flags);
			wake_it = 1;
		}
	} else if (path->flag & PATH_NEEDS_STARTSTOP) {
		stub = mpl_setup_startstop(path);
		if (stub) {
			path->flag &= ~PATH_NEEDS_STARTSTOP;
			expires = 0;
			stub->cmd.retries = path->retries;
			path->retries = 0;
			spin_lock_irqsave(&stub_queue_lock, flags);
			stub->reason = MPL_INTERNAL_CMD;
			mpl_add_stub_to_deferral(stub);
			spin_unlock_irqrestore(&stub_queue_lock, flags);
			wake_it = 1;
		}
	} else if (path->flag & PATH_NEEDS_SET_TGT_PORT) {
		stub = mpl_setup_set_tgt_port(path, path->tgt_port_group, MPL_ALUA_ACTIVE_OPT);
		if (stub) {
			path->flag &= ~PATH_NEEDS_SET_TGT_PORT;
			expires = 0;
			stub->cmd.retries = path->retries;
			path->retries = 0;
			spin_lock_irqsave(&stub_queue_lock, flags);
			stub->reason = MPL_INTERNAL_CMD;
			mpl_add_stub_to_deferral(stub);
			spin_unlock_irqrestore(&stub_queue_lock, flags);
			wake_it = 1;
		}
	} else {
		/* @TLE. this should not happen */
		expires = 0;
	}

	if (expires) {
		path->retries++;
		mpl_restart_sched_cmd(path, expires);
	}
	mpl_module_unlock();
	if (wake_it)
		mpl_wake_thread();
	return;
}

int
mpl_do_base_queuecommand(struct mpl_device *device, struct mpl_path *path,
			 struct io_stub *stub)
{
	int rv;
	unsigned long sflag;
	unsigned long flags;
	spinlock_t    *prev_host_lock, *new_host_lock ;

				/* resubmit stub -- need to re-add to hash */
	stub->path = path;

	if (path->sdev) {
		spin_lock_irqsave(&stub_queue_lock, flags);
		mpl_add_stub_to_lpfcq(stub);
		spin_unlock_irqrestore(&stub_queue_lock, flags);
		stub->cmd.device = path->sdev;
	} else {
		return SCSI_MLQUEUE_HOST_BUSY;
	}

	if (path->host->blocked || path->lld_blocked) {
		stub->cmd.result = ScsiResult(DID_BUS_BUSY, SAM_STAT_BUSY);
		stub->cmd.retries = 0;
		mpl_module_unlock();
		mpl_done(&stub->cmd);
		mpl_module_lock();
		return 0;
	}

	mpl_module_unlock();
	prev_host_lock = stub->sdev->host->host_lock;
	new_host_lock  = stub->cmd.device->host->host_lock;

	spin_unlock_irq(prev_host_lock);
	spin_lock_irqsave(new_host_lock, sflag);
	stub->start_time = jiffies;
	rv = mpl_base_queuecommand(&stub->cmd, mpl_done);
	spin_unlock_irqrestore(new_host_lock, sflag);
	spin_lock_irq(prev_host_lock);

	mpl_module_lock();

	if (rv) {
		spin_lock_irqsave(&stub_queue_lock, flags);
		mpl_remove_stub_from_lpfcq(stub);
		spin_unlock_irqrestore(&stub_queue_lock, flags);
	}
	return rv;
}

int
mpl_queuecommand_alternate_path(struct mpl_device *device,
				struct mpl_path *path,
				struct io_stub *stub)
{
	struct scsi_device *sdev;
	int rc;

	mpl_set_cmd_path(&stub->cmd, path);

	mpl_stub_switch_path(stub, path);

	mpl_module_unlock();
	sdev = stub->sdev;
	spin_lock_irq(sdev->host->host_lock);
	mpl_module_lock();
	rc =  mpl_do_base_queuecommand(device, path, stub);
	mpl_module_unlock();
	spin_unlock_irq(sdev->host->host_lock);
	mpl_module_lock();
	return rc;
}

void
mpl_path_set_active(struct mpl_path *path, int boolean)
{
	struct mpl_host *host = path->device->host;
	struct mpl_device *device;
	struct mpl_path *p;
	int  target = path->target;
	int  port_group = path->tgt_port_group;

	if (path->device->scsi_type) {
		path->active_controller = 0;
		return;
	}

	if (path->active_controller == boolean) {
		return;
	}

	list_for_each_entry(device, &host->devices, devices) {
		list_for_each_entry(p, &device->paths, path_list) {
			if (p->host->host_no == path->host->host_no &&
			    p->target == target &&
			    p->tgt_port_group == port_group) {
				/* LUN 0 on MSA1000 is never available */
				if (p->lun == 0)
					continue;
				p->active_controller = boolean;
			}
		}
	}
	return;
}

void
mpl_path_flip_all_active(struct mpl_path *path)
{
	struct mpl_host *host = path->device->host;
	struct mpl_device *device;
	struct mpl_device *flip_device;
	struct mpl_path *p;
	struct mpl_path *flip_p;
	int  target = path->target;
	int  port_group = path->tgt_port_group;

	if (path->active_controller) {
		return;
	}

	list_for_each_entry(device, &host->devices, devices) {
		list_for_each_entry(p, &device->paths, path_list) {
			if (p->host->host_no == path->host->host_no &&
			    p->target == target &&
			    p->tgt_port_group == port_group) {
				/* LUN 0 on MSA1000 is never available */
				if (p->lun == 0)
					continue;

				if(p->active_controller < 2) {
					/* Now mark all paths in this device
					 * to be flipped
					 */
					flip_device = p->device;
					list_for_each_entry(flip_p, &flip_device->paths, path_list) {
						if(flip_p->active_controller < 2) {
							flip_p->active_controller += 2; 
						}
					}
				}
			}
		}
	}

	list_for_each_entry(device, &host->devices, devices) {
		list_for_each_entry(p, &device->paths, path_list) {
			/* Flip all marked paths */
			if(p->active_controller == 2) {
				p->active_controller = 1; 
			}
			else if(p->active_controller == 3) {
				p->active_controller = 0; 
			}
		}
	}
	return;
}

void
mpl_device_process_pendq(struct mpl_device *device)
{
	struct io_stub *stub;
	unsigned long flags;

	spin_lock_irqsave(&stub_queue_lock, flags);
	while ((stub = mpl_get_first_pendq(device)) != NULL) {
		stub->reason = MPL_RETRY_CMD;
		mpl_add_stub_to_deferral(stub);
	}
	spin_unlock_irqrestore(&stub_queue_lock, flags);
	mpl_module_unlock();
	mpl_wake_thread();
	mpl_module_lock();
}

#include "lpfc_hw.h"
#include "lpfc_sli.h"
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
#include "lpfc_mem.h"
#endif
#include "lpfc_scsi.h"
#include "lpfc_disc.h"
#include "lpfc.h"


#ifdef LPFC_SLI3_NPIV_ENABLED
/* Search for a nodelist entry on a specific list */
int
mpl_get_tgt_wwinfo(void *p, uint32_t tgt_id,
	struct mpl_target *tgtp)
{
	struct lpfc_vport *vport = (struct lpfc_vport *)p;
	struct lpfc_nodelist *ndlp;
	struct lpfc_nodelist *next_ndlp;

	list_for_each_entry_safe(ndlp, next_ndlp, &vport->fc_nodes, nlp_listp) {
		if ((ndlp->nlp_state == NLP_STE_MAPPED_NODE) &&
			(tgt_id == ndlp->nlp_sid)) {
			goto found;
		}
	}

	/* no match found */
	return 0;
found:
	memcpy(tgtp->wwpn, (uint8_t *) &ndlp->nlp_portname, 8);
	memcpy(tgtp->wwnn, (uint8_t *) &ndlp->nlp_nodename, 8);
	tgtp->rport = ndlp->rport;
	return 1;
}

void
mpl_get_hba_wwpn(void *p, uint8_t *ptr)
{
	struct lpfc_vport *vport = (struct lpfc_vport *)p;
	if (vport) {
		memcpy(ptr, (uint8_t *) &vport->fc_portname, 8);
	}
	return;
}

int
mpl_get_hba_nodev_tmo(void *p)
{
	struct lpfc_vport *vport = (struct lpfc_vport *)p;
	return vport->phba->cfg_nodev_tmo;
}

int
mpl_host_is_offline(void *p)
{
	struct lpfc_vport *vport = (struct lpfc_vport *)p;

	if (vport->fc_flag & FC_OFFLINE_MODE) {
		return 1;
	}
	return 0;
}

unsigned short
mpl_get_lun_q_depth( struct mpl_host *host)
{
	unsigned short depth = 0;
	struct lpfc_vport *vport = (struct lpfc_vport *)host->hba;

	if (vport)
		depth = (unsigned short)(vport->phba->cfg_lun_queue_depth);
	return depth;
}

void *
mpl_get_ndlp(void *data)
{
	struct lpfc_rport_data *rdata = (struct lpfc_rport_data *)data;
	if (rdata)
		return ((void *)rdata->pnode);
	else
		return NULL;
}

#else

/* Search for a nodelist entry on a specific list */
int
mpl_get_tgt_wwinfo(void *p, uint32_t tgt_id,
	struct mpl_target *tgtp)
{
	struct lpfc_hba *phba = (struct lpfc_hba *)p;
	struct lpfc_nodelist *ndlp;
	struct list_head *pos;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	struct lpfc_target *targetp;

	targetp = phba->device_queue_hash[tgt_id];
	/* First see if the SCSI ID has an allocated LPFCSCSITARGET_t */

	if (targetp) {
		ndlp =  targetp->pnode;
		if (ndlp) {
			goto found;
		}
	}
#endif

	/* Now try the hard way */
	list_for_each(pos, &phba->fc_nlpmap_list) {
		ndlp = list_entry(pos, struct lpfc_nodelist, nlp_listp);
		if (tgt_id == ndlp->nlp_sid) {
			goto found;
		}
	}

	/* no match found */
	return 0;
found:
	memcpy(tgtp->wwpn, (uint8_t *) &ndlp->nlp_portname, 8);
	memcpy(tgtp->wwnn, (uint8_t *) &ndlp->nlp_nodename, 8);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
	tgtp->rport = ndlp->rport;
#endif
	return 1;
}

void
mpl_get_hba_wwpn(void *p, uint8_t *ptr)
{
	struct lpfc_hba *phba = (struct lpfc_hba *)p;

	if (phba) {
		memcpy(ptr, (uint8_t *) &phba->fc_portname, 8);
	}
	return;
}

int
mpl_get_hba_nodev_tmo(void *p)
{
	struct lpfc_hba *phba = (struct lpfc_hba *)p;
	return phba->cfg_nodev_tmo;
}

int
mpl_host_is_offline(void *p)
{
	struct lpfc_hba *phba = (struct lpfc_hba *)p;

	if (phba->fc_flag & FC_OFFLINE_MODE) {
		return 1;
	}
	return 0;
}

unsigned short
mpl_get_lun_q_depth( struct mpl_host *host)
{
	unsigned short depth = 0;
	struct lpfc_hba *phba = (struct lpfc_hba *)host->hba;

	if (phba)
		depth = (unsigned short)(phba->cfg_lun_queue_depth);
	return depth;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
void *
mpl_get_ndlp(void *data)
{
	struct lpfc_rport_data *rdata = (struct lpfc_rport_data *)data;
	if (rdata)
		return ((void *)rdata->pnode);
	else
		return NULL;
}
#endif
#endif
