/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Fibre Channel Host Bus Adapters.                                *
 * Copyright (C) 2003-2006 Emulex.  All rights reserved.           *
 * EMULEX and SLI are trademarks of Emulex.                        *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of version 2 of the GNU General       *
 * Public License as published by the Free Software Foundation.    *
 * This program is distributed in the hope that it will be useful. *
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND          *
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE      *
 * DISCLAIMED, EXCEPT TO THE EXTENT THAT SUCH DISCLAIMERS ARE HELD *
 * TO BE LEGALLY INVALID.  See the GNU General Public License for  *
 * more details, a copy of which can be found in the file COPYING  *
 * included with this package.                                     *
 *******************************************************************/

/* 
 * $Id: mpl_xp.c 1296 2006-12-29 03:24:24Z jinfante $
 */

#include <linux/version.h>
#include <linux/pci.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_transport_fc.h>

#include <linux/string.h>

#include "mpl.h"
#include "mpl_crtn.h"
#include "mpl_log.h"

static uint8_t xp_str[] = "OPEN";

static int
mpl_xp_probe(struct scsi_cmnd *cmd, struct mpl_path *path)
{
	uint8_t vendor[8 + 1];
	uint8_t product[16 + 1];
	uint8_t rev[4 + 1];
	uint8_t *request;
	uint8_t *prod;
	int len = 0;

	request = mpl_find_cmnd_data(cmd, &len);
	if ((request == NULL) || (len < 36)) {
		mpl_print(0, MPL_ERR "No data in EVA probe len %d: %s %s\n",
      			len, mpl_print_device(path->device),
      			mpl_print_path(path))
		return 0;
	}
	prod = request + 16;

	if (!mpl_vendor_hp_check(cmd)) {
		mpl_release_cmnd_data(cmd);
		return 0;
	}

	if (memcmp(prod, xp_str, sizeof(xp_str) - 1) == 0) {
		trim_to(vendor, request + 8, sizeof(vendor));
		trim_to(product, request + 16, sizeof(product));
		trim_to(rev, request + 32, sizeof(rev));
		mpl_release_cmnd_data(cmd);
		mpl_print(MPL_INIT, MPL_INFO " xp: plugin support for: "
		  	"tgt %d lun %d: %s, %s, %s\n",
		  	cmd->device->id, cmd->device->lun,
			vendor, product, rev);
		return 1;
	} else {
		mpl_release_cmnd_data(cmd);
		return 0;
	}
}

static void
mpl_xp_path_down(struct mpl_path *path)
{
	mpl_print(MPL_RECOVERY, MPL_INFO "Path Down Plugin %s %s\n",
		mpl_print_device(path->device), mpl_print_path(path));
	return;
}

static int
mpl_xp_path_up(struct mpl_path *path)
{
	mpl_print(MPL_RECOVERY, MPL_INFO "Path Up Plugin %s %s\n",
		mpl_print_device(path->device), mpl_print_path(path));
	return 1;
}

static int
mpl_xp_diagnose(struct io_stub *stub, int flag)
{
	struct scsi_cmnd *cmd = &stub->cmd;
	struct mpl_path *path = stub->path;
	uint8_t sk, asc, ascq;
	int rv = 0;

	mpl_print(MPL_RECOVERY, MPL_INFO "Diagnose Plugin "
		"%s %s%scmd %x result %x\n",
		mpl_print_device(path->device), mpl_print_path(path),
		(stub->done) ? " " : " internal ",
		(unsigned int)cmd->cmnd[0], cmd->result);

	sk = cmd->sense_buffer[2] & 0xF;
	asc = cmd->sense_buffer[12];
	ascq = cmd->sense_buffer[13];

	/* Setup alua / active-passive on first command */
	/* Make it think we support active-active ALUA, but don't
	 * set the alua support flag in the device structure.
	 */
	path->tpgs = MPL_ALUA_ACTIVE_ACTIVE;
	path->sas  = MPL_ALUA_ACTIVE_OPT;
	path->device->active_passive = 0;  /* active - active */

	switch (flag) {
	case PLUG_INTERNAL_CMD:
		/* mpl driver ignores the return value for this case */
		break;

	case PLUG_HBEAT_CMD:
		/* mpl driver assumes a return value of 0 means the hbeat
		 * command succeeded.
		 */
		switch (cmd->cmnd[0]) {
		case INQUIRY:
			/* Setup tgt qdepth after first INQUIRY */
			/* Assume 70% utilization of max (1024) */
			path->device->tgt_queue_depth = 717;
		case TEST_UNIT_READY:
			if(cmd->result)
				rv = 1; /* error */
		}
		break;

	case PLUG_SCSI_CMD:
		/* mpl driver assumes a return value of 1 means to initiate
		 * a failover and retry the command.
		 */
		break;
	}
#if 0
	if (sk == 0x02) {
		if ((asc == 0x04) && 
		    ((ascq == 0x00) || (ascq = 0x02) || (ascq == 0x03))) {
			return 1;
		}
		   
	} else if (sk == 0x06) {
		if (((asc == 0x0A) && (ascq == 0x08)) ||
		    ((asc == 0x04) && (ascq == 0x02))) {
			return 1;
		}
	}
#endif
	return rv;
}

struct scsi_id_search_values mpl_xp_id_search_list[] = {
	{ SCSI_ID_VENDOR_SPECIFIC,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_ASCII },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_IEEE_REG_EXTENDED,	SCSI_ID_BINARY },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_IEEE_REG_EXTENDED,	SCSI_ID_ASCII },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_IEEE_REG,	SCSI_ID_BINARY },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_IEEE_REG,	SCSI_ID_ASCII },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_BINARY },
	{ SCSI_ID_NAA,	SCSI_ID_NAA_DONT_CARE,	SCSI_ID_ASCII },
};

/* Return 0 if we find a valid validation id, else return 1 */
int mpl_xp_validation_id(
	struct mpl_path *path,
	uint8_t *page_83,
	uint8_t *serial, int *serial_size)
{
	return mpl_get_scsi_page83_serial(page_83, mpl_xp_id_search_list,
		sizeof(mpl_xp_id_search_list)/sizeof(mpl_xp_id_search_list[0]),
		serial, serial_size);
}

/* DEFINE_PLUGIN(id, name, probe, prequeue, diagnose, down, up, valid)
 */

DEFINE_PLUGIN(xp, "xp", mpl_xp_probe, NULL, mpl_xp_diagnose, 
		mpl_xp_path_down, mpl_xp_path_up, mpl_xp_validation_id)



